#include "deftypes.h"
 
#define	QLEN		   5	
#define	BUFSIZE		20480

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif  
extern int  errno;
extern char *sys_errlist[];
   
 int errexit(char *format, ...)

{
    va_list args;
 
    va_start(args,format);
    _doprnt(format, args, stderr);
    va_end(args);
    exit(1);
}   

int recv_raw_data(int fd, char *buf, int nbytes)
{
    int nleft, nread;
    char *ptr = buf;
    sigset_t oset;
    
    /* Mask three signals */
    /* mask_three_signals(&oset); */

    nleft = nbytes;
    while (nleft > 0) {
      /* nread = read(fd, ptr, nleft);*/
        nread = recv(fd, ptr, nleft,0);
        if (nread < 0)
            return(nread);      /* error, return < 0 */
        else if (nread == 0)
            break;          /* EOF */

        nleft -= nread;
        ptr   += nread;
    }
    printf("%d bytes read for var =%d original = %d left =%d\n",ptr-buf,
	       fd,nbytes,nleft);
    
    /* Unmask three signals */
    /* unmask_three_signals(&oset); */
    return(nbytes - nleft);     /* return >= 0 */

}

void send_raw_data(int sockfd,char *buf,int len)
{
  char *ptr;
  int nleft,nwritten;
  sigset_t oset;
 
  /* Mask three signals */
  /* mask_three_signals(&oset); */

  ptr   = buf;
  nleft = len;
  while (nleft > 0) {
    /*nwritten = write(sockfd, ptr, nleft); */
    nwritten = send(sockfd, ptr, nleft,0);
    if (nwritten<=0)
      errexit("send_buffer: nwritten <= 0",sys_errlist[errno]);
    nleft -= nwritten;
    ptr += nwritten;
  }

  /* Unmask three signals */
  /* unmask_three_signals(&oset); */
  
  printf("I send raw data sent len =%d\n",len);
}

void recv_int(int sockfd,int *command)
{
  char cmnd[4];
  XDR xdrs;
 
  recv_raw_data(sockfd,cmnd,sizeof(int));
  xdrmem_create(&xdrs,cmnd,sizeof(int),XDR_DECODE); 
  xdr_int(&xdrs,command);

  xdr_destroy(&xdrs);
  printf("I received integer : %d\n",*command);
 
}

void send_int(int sockfd,int command)
{
  char cmnd[4];
  XDR xdrs;
 
  xdrmem_create(&xdrs,cmnd,sizeof(int),XDR_ENCODE); 
  xdr_int(&xdrs,&command);
  send_raw_data(sockfd,cmnd,sizeof(int));

  xdr_destroy(&xdrs);
  printf("After send INTEGER : %d\n",command);
}

void recv_String (int sockfd,char *variableName)
{
   bool_t ret;

   char buffer[REGBUFSIZE],intbuf[4];
   int strLen,totalLen;
   XDR xdrs;   
   int i;

   recv_raw_data(sockfd,buffer,sizeof(int));
   strncpy(intbuf,buffer,4);
   xdrmem_create(&xdrs,buffer,sizeof(int),XDR_DECODE); 
   ret=xdr_int(&xdrs,&strLen);
   xdr_destroy(&xdrs);

   totalLen = strLen;
   if ((strLen%4) != 0)
     totalLen = strLen+4-strLen%4;

   recv_raw_data(sockfd,&buffer[sizeof(int)],totalLen);
   /* strncpy(buffer,intbuf,4); */
   xdrmem_create(&xdrs,buffer,REGBUFSIZE,XDR_DECODE);
   ret = xdr_string(&xdrs,&variableName,totalLen);
   variableName[strLen] = 0; /* Append terminating character  */ 
   printf("I received string : %s strlen = %d totlen=%d\n",
         variableName,strLen,totalLen);
      
   xdr_destroy(&xdrs);
}


void send_String (int sockfd,char *variableName)
{
   char buffer[REGBUFSIZE];
   int totalLen,ret;
   XDR xdrs;
       
   xdrmem_create(&xdrs,buffer,REGBUFSIZE,XDR_ENCODE); 
   ret = xdr_string(&xdrs,&variableName,strlen(variableName)); 
   
   totalLen = xdr_getpos(&xdrs);
   send_raw_data(sockfd,buffer,totalLen);
   
   xdr_destroy(&xdrs);
   free(buffer);
   printf("I  send string = %s strlen =%d\n",
          variableName,strlen(variableName));
}

void recv_double(int sockfd,double *command)
{
  char cmnd[8];
  XDR xdrs;
 
  recv_raw_data(sockfd,cmnd,sizeof(double));
  xdrmem_create(&xdrs,cmnd,sizeof(double),XDR_DECODE); 
  xdr_double(&xdrs,command);

  xdr_destroy(&xdrs);
  if (DEBUG)
    printf("I received double : %g\n",*command);
 
}
  

int recv_reg_data(int sockfd)
{
   int tag,type,ndims,nelems,i;
   char variableName[256];
   double *data;
   
   
   /* recv registration info. */
   recv_int(sockfd,&tag);
   recv_String(sockfd,variableName);
   recv_int(sockfd,&type);
   recv_int(sockfd,&ndims);
   recv_int(sockfd,&nelems);
   printf(
    "Registration properties for %s : tag=%d type=%d ndims=%d nelems=%d \n",
     variableName,tag,type,ndims,nelems);
   
   data = (double *)malloc(sizeof(double)*nelems);
   for(i=0;i<nelems;i++)
     recv_double(sockfd,&data[i]);
   printf("I received data \n");
   
   free(data);
   return 4;
   
     
}

int send_config(int sockfd)
{
  send_int(sockfd,WEBFLOW);
  send_String(sockfd,"gauss.config");
  
  send_String(sockfd,"osprey1.npac.syr.edu");
  send_String(sockfd,"4001");
  
  send_String(sockfd,"osprey1.npac.syr.edu");
  send_String(sockfd,"4002");

  send_String(sockfd,"osprey1.npac.syr.edu");
  send_String(sockfd,"4003");
  return 5;
  
    
}




/*------------------------------------------------------------------------
 * main - Concurrent TCP server for ECHO service
 *------------------------------------------------------------------------
 */

int main(int argc, char *argv[])
{
	char	*service = "3334";	/* service name or port number	*/
	struct sockaddr_in fsin;	/* the from address of a client	*/
	int	msock;			/* master server socket		*/
	fd_set	rfds;			/* read file descriptor set	*/
	fd_set	afds;			/* active file descriptor set	*/
	int	alen;			/* from-address length		*/
	int	fd, nfds;
	int ret,i;
	struct timeval tval={0,0};
        int flag=0;
			
	switch (argc) {
	case	1:
		break;
	case	2:
		service = argv[1];
		break;
	}

        if (argc > 2){
	  char stemp[1024];
	     
	  msock = connectsock("osprey2","3338","tcp");
	  send_config(msock);
	  
	  for(;;){
	    scanf("%d",&flag);
	    if (flag){
	      
	      send_int(msock,STEPOVER);
	      recv_String(msock,stemp);
	    }
	    else{
	      
	      send_int(msock,STOP);
	      return;
	    }
	  }
	}
	
	          	  
	
        printf("Begin enf SageServer \n");
	msock = passivesock(service, "tcp", QLEN);
	  

	nfds = getdtablesize();
	FD_ZERO(&afds);
	FD_SET(msock, &afds);

	while (1) {
	    bcopy((char *)&afds, (char *)&rfds, sizeof(rfds));

	    if (select(nfds, &rfds, (fd_set *)0, (fd_set *)0,&tval) < 0)
		errexit("select: %s\n", sys_errlist[errno]);
	    if (FD_ISSET(msock, &rfds)) {
		int	ssock;

		alen = sizeof(fsin);
		ssock = accept(msock, (struct sockaddr *)&fsin,	&alen);
		if (ssock < 0)
		   errexit("accept: %s\n", sys_errlist[errno]);
		FD_SET(ssock, &afds);

		printf("I am connected to port : %d\n",ssock);
		
	     }
   	     for (fd=0; fd<nfds; ++fd)
	       if (fd != msock && FD_ISSET(fd, &rfds)){
	           ret = recv_reg_data(fd);

		   /* Close connection */				   
		   if (ret == 1) {
		      (void) close(fd);
		      FD_CLR(fd, &afds);
		   }
		   /* Stop server */
		   else if (ret == 2){
		     for (fd=0; fd<nfds; ++fd)
		       if (fd != msock && FD_ISSET(fd, &rfds)){
			 (void) close(fd);
			 FD_CLR(fd, &afds);
			 printf("I am closing fd : %d\n",fd);
			 
		       }
		     (void) close(msock);
		     FD_CLR(msock, &afds);
		     printf("I am closing msock : %d\n",msock);
		     exit(0);
		   } /*els if (ret == 2) */
		   
	       } /* if (fd != msock.. */

	}/* while */
}/* main */



/*------------------------------------------------------------------------
 * passivesock - allocate & bind a server socket using TCP or UDP
 *------------------------------------------------------------------------
 */

int passivesock(char *service, char *protocol, int qlen )
/* char    *service;  */     /* service associated with the desired port     */
/* char    *protocol;  */    /* name of protocol to use ("tcp" or "udp")     */
/* int     qlen;      */     /* maximum length of the server request queue   */
{
        struct servent  *pse;   /* pointer to service information entry */
        struct protoent *ppe;   /* pointer to protocol information entry*/
        struct sockaddr_in sin; /* an Internet endpoint address         */
        int     s, type;        /* socket descriptor and socket type    */
	int ret;
	u_short  portbase = 0;  

        bzero((char *)&sin, sizeof(sin));
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = INADDR_ANY;

    /* Map service name to port number */
        if ( pse = getservbyname(service, protocol) )
                sin.sin_port = htons(ntohs((u_short)pse->s_port)
                        + portbase);
        else if ( (sin.sin_port = htons((u_short)atoi(service))) == 0 )
                errexit("can't get \"%s\" service entry\n", service);

    /* Map protocol name to protocol number */
        if ( (ppe = getprotobyname(protocol)) == 0)
                errexit("can't get \"%s\" protocol entry\n", protocol);

    /* Use protocol to choose a socket type */
        if (strcmp(protocol, "udp") == 0)
                type = SOCK_DGRAM;
        else
                type = SOCK_STREAM;

    /* Allocate a socket */
        s = socket(PF_INET, type, ppe->p_proto);
        if (s < 0)
                errexit("can't create socket: %s\n", sys_errlist[errno]);

    /* Bind the socket */
        if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
                errexit("can't bind to %s port: %s\n", service,
                        sys_errlist[errno]);
        if (type == SOCK_STREAM && listen(s, qlen) < 0)
                errexit("can't listen on %s port: %s\n", service,
                        sys_errlist[errno]);
	
        return s;
}

int
connectsock( host, service, protocol )
char    *host;      /* name of host to which connection is desired  */
char    *service;   /* service associated with the desired port */
char    *protocol;  /* name of protocol to use ("tcp" or "udp") */
{
    struct hostent  *phe;   /* pointer to host information entry    */
    struct servent  *pse;   /* pointer to service information entry */
    struct protoent *ppe;   /* pointer to protocol information entry*/
    struct sockaddr_in sin; /* an Internet endpoint address     */
    int s, type;    /* socket descriptor and socket type    */
 
 
    bzero((char *)&sin, sizeof(sin));
    sin.sin_family = AF_INET;
 
    /* Map service name to port number */
    if ( pse = getservbyname(service, protocol) )
        sin.sin_port = pse->s_port;
    else if ( (sin.sin_port = htons((u_short)atoi(service))) == 0 )
        errexit("can't get \"%s\" service entry\n", service);
 
    /* Map host name to IP address, allowing for dotted decimal */
    if ( phe = gethostbyname(host) )
 
        bcopy(phe->h_addr, (char *)&sin.sin_addr, phe->h_length);
    else if ( (sin.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE )
        errexit("can't get \"%s\" host entry\n", host);
 
    /* Map protocol name to protocol number */
    if ( (ppe = getprotobyname(protocol)) == 0)
        errexit("can't get \"%s\" protocol entry\n", protocol);
 
    /* Use protocol to choose a socket type */
    if (strcmp(protocol, "udp") == 0)
        type = SOCK_DGRAM;
    else
        type = SOCK_STREAM;
 
    /* Allocate a socket */
    s = socket(PF_INET, type, ppe->p_proto);
    if (s < 0)
        errexit("can't create socket: %s\n", sys_errlist[errno]);
 
    /* Connect the socket */
    if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
        errexit("can't connect to %s.%s: %s\n", host, service,
            sys_errlist[errno]); 
    return s;
}


