program test2d
integer, parameter         :: rsize=25
integer, parameter         :: asize=25
integer, parameter         :: bsize=25
integer, parameter         :: csize=100
real*8, dimension(rsize,rsize,rsize) :: r
real*8, dimension(asize,asize) :: a
real*8, dimension(bsize,bsize) :: b
real*8, dimension(rsize,rsize) :: p1x,p1y,p1z
real*8, dimension(asize)     :: xa,ya
real*8, dimension(bsize)     :: xb,yb
real*8, dimension(rsize)     :: xr,yr
real*8, dimension(csize)     :: c,xc
real*8 sigma, ampl, timei,offset
integer ndelay
real*8 result

!HPF$ PROCESSORS P(NUMBER_OF_PROCESSORS())
!HPF$ DISTRIBUTE (*,BLOCK) ONTO P :: b,p1x,p1y,p1z
!HPF$ DISTRIBUTE ya(BLOCK) ONTO P
!HPF$ DISTRIBUTE yb(BLOCK) ONTO P
!HPF$ DISTRIBUTE yr(BLOCK) ONTO P
!HPF$ DISTRIBUTE r(*,BLOCK,*) ONTO P

interface
EXTRINSIC (HPF_LOCAL) subroutine SciViz2d(iflag,timei,x,y,fx,np)
real*8, dimension(:)    ::  x,y
real*8, dimension(:,:)  ::  fx
real*8                  ::  timei
integer                 ::  np,iflag
!HPF$ DISTRIBUTE *(*,BLOCK) :: fx
!HPF$ DISTRIBUTE *(BLOCK) :: y
end subroutine

EXTRINSIC (HPF_LOCAL) subroutine SciViz1d(iflag,timei,x,f,np)
real*8, dimension(:)    ::  x
real*8, dimension(:)    ::  f
real*8                  ::  timei
integer                 ::  np,iflag
end subroutine

subroutine compute_r(ampl,sigma,r)

real*8 ampl,sigma
real*8, dimension(:,:,:), intent(out) :: r
!HPF$ PROCESSORS P(NUMBER_OF_PROCESSORS())
!HPF$ DISTRIBUTE *(*,BLOCK,*) ONTO *P :: r

end subroutine



end interface  

integer i,j,k,l,m

np=NUMBER_OF_PROCESSORS()
ampl=1.0d-02
sigma=20.d0
offset=asize/2.0d0
ndelay=6000000

forall(i=1:asize) xa(i)=i*1.d0
forall(i=1:asize) ya(i)=i*1.d0
forall(i=1:bsize) xb(i)=i*1.d0
forall(i=1:bsize) yb(i)=i*1.d0
forall(i=1:rsize) xr(i)=i*1.d0
forall(i=1:rsize) yr(i)=i*1.d0
forall(i=1:csize) xc(i)=i*1.d0

w=3.0
do k=1,20

call delayit(ndelay,result)
write(6,*) result
sigma=sigma*1.25

call compute_r(ampl,sigma,r)

forall(i=1:asize,j=1:asize) &
  a(i,j)=ampl* exp(-((xa(i)-offset)**2/sigma))*   &
               exp(-((ya(j)-offset)**2/sigma))

forall(i=1:bsize,j=1:bsize) &
  b(i,j)=ampl*sin(0.1d0*(xb(i)+k))*cos(0.1d0*(yb(j)-k))

forall(i=1:csize) c(i)=ampl*sin(0.314d0*i+k)
p1x=r(10,:,:)
p1y=r(:,10,:)
p1z=r(:,:,10)

timei=1.0d0*k

call SciViz2d(2,timei,xb,yb,b,np)
call SciViz1d(3,timei,xc,c,np)
call SciViz2d(1,timei,xr,yr,p1z,np)
call SciViz2d(4,timei,xr,yr,p1x,np)
call SciViz2d(5,timei,xr,yr,p1y,np)

enddo

do l=1,19

call delayit(ndelay,result)
write(6,*) result
sigma=sigma/1.25

call compute_r(ampl,sigma,r)

forall(i=1:asize,j=1:asize) &
  a(i,j)=ampl* exp(-((xa(i)-offset)**2/sigma))*   &
               exp(-((ya(j)-offset)**2/sigma))

forall(i=1:bsize,j=1:bsize) &
  b(i,j)=ampl*sin(0.1d0*(xb(i)+k))*cos(0.1d0*(yb(j)-k))

forall(i=1:csize) c(i)=ampl*sin(0.314d0*i+k)

p1x=r(10,:,:)
p1y=r(:,10,:)
p1z=r(:,:,10)
k=k+1
timei=1.0d0*k

call SciViz2d(1,timei,xr,yr,p1z,np)
call SciViz2d(2,timei,xb,yb,b,np)
call SciViz1d(3,timei,xc,c,np)
call SciViz2d(4,timei,xr,yr,p1x,np)
call SciViz2d(5,timei,xr,yr,p1y,np)

enddo


do l=k,k+100
call delayit(ndelay,result)
write(6,*) result
forall(i=1:csize) c(i)=ampl*sin(0.314d0*i+k)
timei=1.0d0*l
call SciViz1d(3,timei,xc,c,np)
enddo

end

subroutine compute_r(ampl,sigma,r)

real*8 ampl,sigma
real*8, dimension(:,:,:), intent(out) :: r

real*8, dimension(size(r,dim=1)):: x,y,z

!HPF$ PROCESSORS P(NUMBER_OF_PROCESSORS())
!HPF$ DISTRIBUTE *(*,BLOCK,*) ONTO *P :: r

!HPF$ DISTRIBUTE (BLOCK) :: y

integer asize,offset
integer i,j,k

asize=size(r,dim=1)
offset=asize/2

forall(i=1:asize) x(i)=i*1.d0
forall(i=1:asize) y(i)=i*1.d0
forall(i=1:asize) z(i)=i*1.d0

forall(i=1:asize,j=1:asize,k=1:asize) &
  r(i,j,k)=ampl*sin(0.1d0*(x(i)+k))*   &
               exp(-((y(j)-offset)**2/sigma))*   &
               exp(-((z(k)-offset)**2/sigma))

end

subroutine delayit(n,result)
integer n
real*8 result
integer i
real*8 c1,c2,c3

c1=1.34d0
c2=2.34d0
c3=3.45d0

result=1.0d0
do i=1,n
  result=sqrt(i*c1+c2)/sqrt(i*c3-c1)+result
enddo
return
end

EXTRINSIC (HPF_LOCAL) subroutine SciViz2d(iflag,timei,x,y,fx,np)


USE HPF_LOCAL_LIBRARY

real*8, dimension(:)    ::  x,y
real*8, dimension(:,:)  ::  fx
real*8                  ::  timei
integer                 ::  np,iflag
!HPF$ DISTRIBUTE *(*,BLOCK) :: fx
!HPF$ DISTRIBUTE *(BLOCK)   :: y

integer isrv 
real*8, dimension(4)    ::  bb
integer                 ::  nx,ny,me
integer,dimension(2)    ::  ylocal,yglobal
integer, dimension(1)   :: tt

nx=size(x)
ny=size(y)

me=MY_PROCESSOR()

!ylocal(1)=1
!ylocal(2)=ny
!
!call local_to_global(y,ylocal,yglobal)


bb(1)=1
bb(2)=nx
bb(3)=me*ny+1
bb(4)=(me+1)*ny

write(6,100) timei,np,nx,ny,bb(3),bb(4),me

100 Format(G17.3,1x,i3,1x,i3,1x,i3,2(1x,g8.3),1x,i2)

!isrv=java_pbbser2d("2d Gauss",timei,bb,nx,ny,fx,np)

np=NUMBER_OF_PROCESSORS()
select case (iflag)


case(1)
     isrv=java_bbser2d("a",timei,bb,nx,ny,fx,np)
case(2) 
     isrv=java_bbser2d("b",timei,bb,nx,ny,fx,np)   
case(4) 
     isrv=java_bbser2d("px",timei,bb,nx,ny,fx,np)   
case(5) 
     isrv=java_bbser2d("py",timei,bb,nx,ny,fx,np)   

end select
return

end




EXTRINSIC (HPF_LOCAL) subroutine SciViz1d(iflag,timei,x,f,np)

USE HPF_LOCAL_LIBRARY

real*8, dimension(:)    ::  x
real*8, dimension(:)    ::  f
real*8                  ::  timei
integer                 ::  np,iflag

integer isrv 
real*8, dimension(2)    ::  bb
integer                 ::  nx,me
integer,dimension(2)    ::  ylocal,yglobal
integer, dimension(1)   :: tt

nx=size(x)

me=MY_PROCESSOR()

bb(1)=1
bb(2)=nx

!if(iflag==1) then
!     isrv=java_bbser2d("Gauss",timei,bb,nx,ny,fx)
!else if(iflag==2) then 
!     isrv=java_bbser2d("SinCos",timei,bb,nx,ny,fx)   
!else
     isrv=java_bbser("c",timei,bb,nx,f)        
!endif
return

end






