module decls
interface
extrinsic (F77_LOCAL) subroutine Configure(var_id,var_port)
  integer  var_id
  integer   var_port
end subroutine configure
  
extrinsic (F77_LOCAL) subroutine Send_data(var_id)
  integer var_id
end subroutine send_data

extrinsic (F77_LOCAL) subroutine Recv_data(var_id)
  integer var_id
end subroutine Recv_data

extrinsic (F77_LOCAL) subroutine Unregister(var_id)
  integer var_id
end subroutine unregister

subroutine send_filtered_data(buf,surface,bounds,s_id1,s_id2,s_id3)
      double precision  surface(:,:) ,buf(:)
      integer bounds(:),s_id1,s_id2,s_id3
      !HPF$ DISTRIBUTE *(BLOCK,BLOCK) ::  surface
end subroutine

subroutine setup_temp(temp,bounds,ndims,dim_info,s_id1,s_id2,s_id3,port)

      double precision ,dimension(:) :: temp
      integer bounds(:),ndims,dim_info(2)
      integer s_id1,s_id2,s_id3,port
end subroutine 

end interface

integer T_CHAR,T_INT ,T_FLOAT, T_LONG,T_DOUBLE,T_BOOL,T_BYTE,T_SHORT
parameter (T_CHAR=0,T_INT=1 ,T_FLOAT=2, T_LONG=3 )
parameter (T_DOUBLE=4,T_BOOL=5,T_BYTE=6,T_SHORT=7)
end module decls



      Program Laplace

      USE HPF_LIBRARY
      USE HPF_LOCAL_LIBRARY
      USE decls
 
  
!************************************************************
!*
!* This HPF program implements a finite difference method for
!* solving the Laplace Heat Equation via iteration.  Iteration
!* is continued until a steady state is reached to within some
!* epsilon.
!*
!************************************************************

      INTEGER*4 maxX$,maxY$
      PARAMETER (maxX$=16)
      PARAMETER (maxY$=16) 
      double precision  surface(maxX$,maxY$)
      double precision  temp(maxX$,maxY$)
      double precision  buf(size(surface))

!HPF$ DISTRIBUTE (BLOCK,BLOCK) :: surface

      INTEGER port,count,S_ID
      DOUBLE PRECISION epsilon,maxErr
      integer ndims,s_id1,s_id2,s_id3,dim_info(2)
      integer step,bounds(6)

      ndims = 2
      port = 3336
 
      call get_parameters(epsilon,maxerr,step,bounds,port)
      print *,'after get_parameters'

      call setup_temp(buf,bounds,ndims,dim_info,s_id1,s_id2,s_id3,port)
      CALL Init(surface,maxX$,2,maxX$-1,maxY$,2,maxY$-1,0.0)
      
      count = 0
      print *, 'maxErr & epsilon', maxErr, ' ',epsilon
      do while (maxErr > epsilon)
        print *,"ITERATION: ",count
        if (MOD(count,step) .EQ. 0) then 
          call send_filtered_data(buf,surface,bounds,s_id1,s_id2,s_id3) 
          print *,'after send_filtered_data & count ',count
        endif

	temp = surface             
        surface(2:maxX$-1,2:maxY$-1) = (temp(1:maxX$-2,2:maxY$-1) + & 
                                        temp(3:maxX$,2:maxY$-1) + &
                                        temp(2:maxX$-1,1:maxY$-2) + &
                                        temp(2:maxX$-1,3:maxY$))/4.0 

       	 maxErr = MAXVAL(ABS(surface-temp))
         
         if (MOD(count,100) .EQ. 0) then
           print *,count,maxErr
         end if
         count = count + 1
      end do
      print *, "I am done in loop"
      STOP
      End program Laplace


!******************************************************************
!******************************************************************

     subroutine setup_temp(temp,bounds,ndims,dim_info,s_id1,s_id2,s_id3,port)
      USE DECLS
      double precision ,dimension(:) :: temp
      integer bounds(:),ndims,dim_info(2)
      integer s_id1,s_id2,s_id3,port

      integer ubns(2)
      integer dims(2),dimlen
   
      ubns(1) = ceiling(real(bounds(2)- bounds(1)+1)/real(bounds(3)))
      ubns(2) = ceiling(real(bounds(5)- bounds(4)+1)/real(bounds(6)))
      dim_info = ubns  
      print *,'ubns ',ubns       

      dimlen = 0
      call Register_var(T_INT,S_ID1,ndims,'S1',dimlen,dims)
      dimlen  = 1
      dims(1) = ndims
      call Register_var(T_INT,S_ID2,dim_info,'S2',dimlen,dims) 
      dimlen  = 1
      ubns(1) = ubns(1)*ubns(2)
      call Register_var(T_DOUBLE,S_ID3,temp,'S3',dimlen,ubns)
     
      call configure(S_ID1, port)
      call configure(S_ID2, port)
      call configure(S_ID3, port)

      call send_data(s_id1) 
      call send_data(s_id2)
     
    end subroutine


!******************************************************************
!******************************************************************

      subroutine get_parameters(epsilon,maxerr,step,bounds,port)
      USE decls  
      double precision epsilon,maxerr
      integer step,bounds(:),port
      integer dims(1),S_ID1,S_ID2,S_ID3,S_ID4

      double precision epsilon_ar(number_of_processors())
      double precision maxerr_ar(number_of_processors())
      integer  step_ar(number_of_processors())
      integer bounds_ar(number_of_processors(),6)
      !HPF$ DISTRIBUTE (BLOCK)   :: epsilon_ar, maxerr_ar ,step_ar
      !HPF$ DISTRIBUTE (BLOCK,*) :: bounds_ar

      ndims = 0
      call Register_var(T_DOUBLE,S_ID1,epsilon,'S1',ndims,dims)
      call Register_var(T_DOUBLE,S_ID2,maxErr,'S2',ndims,dims)
      call Register_var(T_INT,S_ID3,step,'S3',ndims,dims)
      ndims = 1
      dims(1) = 6
      call Register_var(T_INT,S_ID4,bounds,'S4',ndims,dims) 
         
      call configure(S_ID1, port)
      call configure(S_ID2, port)
      call configure(S_ID3, port)
      call configure(S_ID4, port)

      call recv_data(S_ID1)
      call recv_data(S_ID2)
      call recv_data(S_ID3)
      call recv_data(S_ID4)

      !processors_0 copies  locally assigned scalar values to distributed
      ! HPF arrays 
      epsilon_ar(1)    = epsilon
      maxerr_ar(1)     = maxerr
      step_ar(1)       = step
      bounds_ar(1,:)   = bounds

      ! Each processor gets the scalar value from processor_0
      epsilon = epsilon_ar(1)
      maxerr  = maxerr_ar(1)
      step    = step_ar(1)
      bounds  =  bounds_ar(1,:) 
 		
      end subroutine  get_parameters

!******************************************************************
!******************************************************************

      subroutine send_filtered_data(buf,surface,bounds,s_id1,s_id2,s_id3)
      USE decls 
      double precision  surface(:,:) ,buf(:)
      integer bounds(:),s_id1,s_id2,s_id3
      !HPF$ DISTRIBUTE *(BLOCK,BLOCK) ::  surface

      double precision,allocatable :: temp(:,:)
      integer ubns(2),lbns(2)
      integer dimlen,dims(6),ub1,ub2,lb1,lb2,s1,s2  
           
      lb1 = bounds(1)
      ub1 = bounds(2)
      s1  = bounds(3)
      lb2 = bounds(4)
      ub2 = bounds(5)
      s2  = bounds(6)
  
      buf  = reshape(surface(lb1:ub1:s1,lb2:ub2:s2),&
            SHAPE=(/ size(surface(lb1:ub1:s1,lb2:ub2:s2)) /))
      print *,'temp ',surface(lb1:ub1:s1,lb2:ub2:s2)
       
      call send_data(S_ID1)
      call send_data(S_ID2)
      call send_data(S_ID3) 

      end subroutine


!************************************************************
!*
!* Subroutine Init initializes a 2D surface with a random 
!* distribution of temperatures in the interior and with 
!* temperature, temp, at the edges.
!*
!************************************************************
      Subroutine Init(surface,sizeX,x0,xn,sizeY,y0,yn,temp)

      INTEGER*4 sizeX,sizeY
      DOUBLE PRECISION surface(sizeX,sizeY)
      INTEGER*4 x0,xn,y0,yn
      DOUBLE PRECISION temp
      DOUBLE PRECISION rand
      
      INTEGER*4 i,j,seed

      seed = time()
      Call srand(seed)
	
      do i = x0,xn
	do j = y0,yn
      	  surface(i,j) = rand() 
	end do
      end do

      print *, "Init - rand OK"
      surface(1,1:sizeY) = temp
      surface(sizeX,1:sizeY) = temp
      surface(1:sizeX,1) = temp
      surface(1:sizeX,sizeY) = temp

      print *, "init done"
      return
      end



