#include <stdio.h>
#include "compilerApi.h"
#include <malloc.h>
#include <string.h>

// Find the enclosing function/subroutine unit

// decl for lexical analyser
#define VARID   509
#define MAXVARS 256
char *input_file;
FILE *inf_ptr;
#undef YYLMAX
#define YYLMAX 8192
extern char yytext[YYLMAX];
extern int yylineno;

SgFile *curSgFile;
extern "C" int generateNewFunction(char *hpfStmtList,char *calledfunc,
           char *funcName,char *usedVars[],int *numOfusedVars,char *genFunc);
extern "C" void openNewProject(char *depfileName);
extern "C" int find_var(char *origProgName,char *usedVars);

struct Internalfunctable{
  SgStatement *startAddr;
  char *fname;
}curFuncTable[256];
static int    numOfFuncs;

insertList(char *list[],int *len,char *key);
findUsedVars(char *inputFile,char *usedVars[],int *nVars);
int find_var(char *origProgName,char *usedVars);
      
void openNewProject(char *depFileName)
{
   int j;
   char *myargs[3]={NULL,NULL,(char *)0};
   myargs[1] =  (char *)malloc(256);

   strcpy(myargs[1],depFileName);
   SgProject *project = new SgProject("",2,myargs);
   curSgFile = &(project->file(0));

   for(j=0;j<curSgFile->numberOfFunctions();j++){
     curFuncTable[j].startAddr = curSgFile->functions(j);
     SgSymbol *subsym  =  (curFuncTable[j].startAddr)->symbol();
     curFuncTable[j].fname     = subsym->identifier();
   }
   numOfFuncs = curSgFile->numberOfFunctions();

} 


int findex(char *name)
{
   int found = -1;  

   for(int j=0;j<numOfFuncs;j++){
     if (!strcmp(name,curFuncTable[j].fname )){
         found = j;
	 break;
     }
   }  
  
   if ((found<0)||(found==numOfFuncs)){
      printf("I could not find file : %s\n",name);
      exit(0);
   }
   return found;

}   

int generateNewFunction(char *hpfStmtList,char *calledFunc,
  char *funcName,char *usedVars[],int *numOfusedVars,char *generatedFunction)
{
  SgFunctionSymb *fsymb;
  SgStatement    *prev,*s,*ptfunc;
  int nVars;
  SgStatement *curStmt;
  FILE *fTemp;
  char *notDefinedVars[MAXVARS];
  int  numOfNDVars=0;
 

  /* Store the user's HPF stsmts into temporary File */
  char *ts = (char *)malloc(strlen(hpfStmtList));
  strcpy(ts,hpfStmtList);
  ts[strlen(ts)] = 0;
  fTemp = fopen("userStmts","w");
  fprintf(fTemp,"%s",ts);
  fclose(fTemp);
  free(ts);
      
  //Find the variables used in the HPF stmts user gave in intercative mode
  findUsedVars("userStmts",usedVars,&nVars);

  // get the scope and name of enclosing program unit
  int index = findex(calledFunc);
  SgStatement *origProgUnit = curFuncTable[index].startAddr;
  char *origProgName = (((SgProgHedrStmt *)origProgUnit)->name()).identifier();

  // Create new function
  char *newSubName          = (char *)malloc(256);
  sprintf(newSubName, "newgen_%s", origProgName);
  strcpy(funcName,newSubName);
  fsymb   = new SgFunctionSymb(PROCEDURE_NAME,newSubName,
			       TYNULL,*(origProgUnit->controlParent()));
  
  ptfunc  = new SgProcHedrStmt(*fsymb);

  // add all the declaration stmts of original function
  prev = ptfunc;

  for(int i=0;i<origProgUnit->numberOfChildrenList1()-1;i++){
    SgStatement *chld = origProgUnit->childList1(i);
    if ((chld->variant() == VAR_DECL)||
        (chld->variant() == COMBINED_DIRECTIVE_DECL)||
	(chld->variant() == DISTRIBUTE_DECL)||
        (chld->variant() == ALIGN_DECL)||
        (chld->variant() == PARAM_DECL)||
        (chld->variant() == PROCESSORS_STMT)||
        (chld->variant() == TEMPLATE_STMT)  ) 
    {
       s = &origProgUnit->childList1(i)->copy();
       prev->insertStmtAfter(*s,*ptfunc);
       prev = s;
   }
  }  
     
  *numOfusedVars = nVars;    
  // Add used variables as parameters to new function
  int cnt=0;
  for(int i=0;i<nVars;i++){
    int index = find_var(origProgName,usedVars[i]);
    /*printf("I am adding %s as parameter in %s index=%d\n",
	  usedVars[i],origProgName,index);  */    
    if (index >= 0){    
      SgVariableSymb *ns = new SgVariableSymb(usedVars[i]);
      fsymb->insertParameter(cnt++,*ns);
    }
    else{
      notDefinedVars[numOfNDVars] = (char *)malloc(256);
      strcpy(notDefinedVars[numOfNDVars],usedVars[i]);
      numOfNDVars++; 
    }
  }
      
  // add the user's HPF stmts as comments before "subroutine End" stmt
  SgStatement *endStmt = ptfunc->childList1(ptfunc->numberOfChildrenList1()-1);
  endStmt ->addComment(hpfStmtList);

  // Unparse new created function to standard output
  char *tempBuf = ptfunc->unparseToBuf();
  fTemp = fopen("userStmts.hpf","w");
  fprintf(fTemp,"%s",tempBuf);
  fclose(fTemp);
  
  int retval = system(pghpf -Mfree -c userStmts.hpf > errout 2>&1");
  /* There is syntax error, leave it as it is and report error to user */
  if (retval != 0)
    return retval;

  /* transform to dep file */
  system("hpf2dep userStmts.hpf");

  /* open the project file */
  char *myargs[3]={NULL,NULL,(char *)0};
  myargs[1] =  (char *)malloc(256);
  strcpy(myargs[1],"userStmts.dep");
  SgProject *project = new SgProject("",2,myargs);
  SgFile *tempSgFile = &(project->file(0));
  
  /* add registration for variables user defined now */
  SgStatement *firstStmt = tempSgFile->firstStatement();
  insert_decls(firstStmt->lexNext());
  insert_register_defined_vars(firstStmt->lexNext(),
		 numOfNDVars, notDefinedVars);
  tempBuf = firstStmt->lexNext()->unparseToBuf();  
  fTemp = fopen("foo.hpf","w");
  fprintf(fTemp,"%s",tempBuf);
  fclose(fTemp);
  generatedFunction = tempBuf; 
  
  return 0;
  
}

findUsedVars(char *inputFile,char *usedVars[],int *nVars)
{

  int cnt=0,token;
  
  input_file = (char *)malloc(strlen(inputFile)+1);
  strcpy(input_file,inputFile);
 
  if (!(inf_ptr = fopen(input_file,"r"))){
    printf("I can not open input file : %s\n",input_file);
    exit(0);
  }
  
  while((token=yylex())){
    if (token == VARID)
      insertList(usedVars,&cnt,yytext);
  }

  fclose(inf_ptr);  
  *nVars = cnt;

}

insertList(char *list[],int *len,char *key)
{
  int found = 0;
  
  for(int i=0;i<(*len);i++)
    if (!strcmp(list[i],key)){
      found = 1;
      break;
    }
  if (!found){
    list[(*len)] = (char *)malloc(strlen(key)+1);
    strcpy(list[(*len)],key);
    (*len)++;
  }
}

