#include <stdio.h>
#include "compilerApi.h"
#include "debugger.h"
#include <malloc.h>
#include <string.h>

#define TIMER_CLEAR    0
#define TIMER_START    1
#define TIMER_STOP     2
#define TIMER_ELAPSED  3
#define REGISTER       4
#define HPF_SERVER     5
#define INIT           6
#define SLOCAL       1
#define DISTRIBUTED 2

#define BEFORE  0
#define AFTER   1
#define MAX_FUNC_NUM 300
#define MAX_PAIR     300

enum FLAG {off, on};
FLAG in_main_program;
FLAG in_procedure;
int tabnumber=0;
void doFullType(SgType *t);
void doSymbAttribs(SgSymbol *s);
void doRoutineHeader(SgStatement *s);
void doSymb(SgSymbol *s);
void register_var();
void timer_start(),timer_clear(),timer_stop(),timer_elapsed();
void configure();
void insert_register_and_config_params(SgStatement *);
int give_type_symbol();
SgStatement *find_first_exec_stmt(SgStatement *);
void hpf_server();
void find_exec_next_stmt(SgStatement *,int *,int *);
SgStatement *find_outermost_scope(SgStatement *);
void create_one_func(SgStatement *);
SgVariableSymb *DeclareTemporaryLike(SgSymbol * ,int );
void insert_decls(SgStatement *,int numOfNDVars,char *notDefinedVars[]);
void build_function(int ,SgStatement *);
int isInLoop(SgStatement *);
int isTemplate(SgSymbol *symb);

create_functionTable_file(SgProject * project, int *nfiles);
int isInLoop(SgStatement *s);
SgStatement *find_outermost_scope(SgStatement *s);
void action(SgStatement * s,int find);
void transform_registration(SgStatement * s,int find) ;
write_summary(char *outFile,int nfiles);
SgStatement *find_first_exec_stmt(SgStatement *s);
void insert_register_defined_vars(SgStatement *s,
		int numOfNDVars,  char *notDefinedVars[]);
extern "C" void readParseTree(char *projectFileName ,int *nfiles);

//char projectName[256];
typedef struct festmts
{
  char *fname;
  int  feline;
  SgStatement *s;
}FUNCS;

SgStatement *first;
int nextp;
FUNCS funcs[MAX_FUNC_NUM];
int nextf ;
FILES *parsefiles;
int   *lengthOfFiles;
FTAB  *parfunctab;
int fileIndex;
SgVariableSymb *hpf_server_symb,*jmp_symb;
char *parcurFileName,*parcurFuncName;



int parse_find_var(char *funcName,char *name)
{
  int i,found=-1;
  
  for(i=0;i< max_push_id;i++){
     if ((!strcmp(stack[i].funcName,funcName))&&
	(!strcmp(stack[i].name,name))){
       found = i;
       break;
    }
  } 
  return found;
}



// Insert a C function call statement
void 
InsertFCallNode(SgStatement *s,int wfunction,int how)
{
  int myline = s->lineNumber();
  SgSymbol *symb;
  char *fsymb ;
  SgStatement *f;

  switch (wfunction) {
    case TIMER_CLEAR: symb = new SgVariableSymb("timer_clear");
                        break;
    case TIMER_START: symb = new SgVariableSymb("timer_start");
                        break;
    case TIMER_STOP:  symb = new SgVariableSymb("timer_stop");
                        break;
    case TIMER_ELAPSED:symb = new SgVariableSymb("timer_elapsed"); 
                         break;
    case REGISTER : symb = new SgVariableSymb("REGISTER");
      break;
    case HPF_SERVER: symb = hpf_server_symb;
      break;
    case INIT: symb = new SgVariableSymb("init_hpf_server");
      break;
    
 }

 

  // you can put break only at points before which HPF_Server 
  // call is inserted   
  parsefiles[fileIndex].lineInfo[myline].isExecutable = 1;
  parsefiles[fileIndex].lineInfo[myline].isInLoop = isInLoop(s);
  SgCallStmt *func = new SgCallStmt(*symb);

  /* add myline parameter */  
  SgValueExp *ev = new SgValueExp(myline);
  func->addArg(*ev);

  f = find_outermost_scope(s);
  SgProgHedrStmt *fph = (SgProgHedrStmt *)f;
  fsymb = (fph->name()).identifier();

  // add current function name  
  char *ts  = (char *)malloc(strlen(fsymb)+5);
  strcpy(&ts[1], fsymb );
  ts[0] = '\"';
  ts[strlen(fsymb)+1] = '\\';
  ts[strlen(fsymb)+2] = '0';
  ts[strlen(fsymb)+3] = '\"';
  ts[strlen(fsymb)+4] = 0;    

  SgVariableSymb *dummy = new SgVariableSymb(ts,*f);
  SgValueExp *varName = new SgValueExp(dummy->identifier());
  func->addArg(*varName);

  // add current file name
  ts  = (char *)malloc(strlen(parcurFileName)+5);
  strcpy(&ts[1], parcurFileName);
  ts[0] = '\"';
  ts[strlen(parcurFileName)+1] = '\\';
  ts[strlen(parcurFileName)+2] = '0';
  ts[strlen(parcurFileName)+3] = '\"';
  ts[strlen(parcurFileName)+4] = 0;    

  dummy = new SgVariableSymb(ts,*f);
  varName = new SgValueExp(dummy->identifier());
  func->addArg(*varName);

  switch (how) {
    case BEFORE:
         s->insertStmtBefore(*func,*(s->controlParent()));
         break;
    case AFTER:
         s->insertStmtAfter(*func,*(s->controlParent()));
         break;
  }
 
  /***********************************************************/
  /* Add setjump function call to store stack information */
   SgCallStmt *jmp_func = new SgCallStmt(*jmp_symb);
   SgValueExp *jmp_ev = new SgValueExp(myline);
   jmp_func->addArg(*jmp_ev);
   func->insertStmtBefore(*jmp_func,*(s->controlParent()));
  /***********************************************************/

}

void action(SgStatement * s)
{
  SgDeclarationStatement *decl;
  SgProgHedrStmt *spg;
  SgCallStmt *cs;
  SgForStmt *fs;
  int flag = 0;
  SgAssignStmt *ass ;
  SgForStmt *sfo ; 
  SgCallStmt *call;
  
  if (call = isSgCallStmt(s))
     parcurFuncName = call->name()->identifier();
      
  if (cs=isSgCallStmt(s)){
      if ((!strcmp(cs->name()->identifier(),"register_var"))||
          (!strcmp(cs->name()->identifier(),"setjmp"))||
          (!strcmp(cs->name()->identifier(),"hpf_server")))
	   
        return;
  }
  
  SgStatement *cp = s->controlParent();
  if ((isSgExecutableStatement(s))&&
      (s->variant() != CONTROL_END)&&
      (s->variant() != MODULE_STMT)&&
      (s->variant() != USE_STMT)&&
      (s->variant() != INCLUDE_STAT)&&
      ((!cp)||(!(cp->lineNumber()==s->lineNumber()))) )
          InsertFCallNode(s , HPF_SERVER,BEFORE);
 

  // put server function call before end of function unit
    
}

void transform_registration(SgStatement * s) 
{
  int i=0;

  int org_numberOfChildren1=s->numberOfChildrenList1();
   int org_numberOfChildren2= s->numberOfChildrenList2();
  int org_numberOfChildren = org_numberOfChildren1 + org_numberOfChildren2;
   SgStatement * * children = new SgStatement * [org_numberOfChildren];
   SgStatement *st;
   SgStatement *line;
   
  for (i=0; i<org_numberOfChildren1; i++)
	children[i]=s->childList1(i);

  for (; i<org_numberOfChildren; i++)
     children[i]=s->childList2(i-org_numberOfChildren1);

   for (i=0; i<org_numberOfChildren; i++){

     st =  children[i];
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
         
     {
         action(st);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st);
	   i++;
	   continue;
     }
     else{
         action(st);
         insert_register_and_config_params(st);
         transform_registration(st);
       }
     
   }
   delete [] children;
   return;
   
}


void transform_registrationA(SgStatement * s) 
{
  int i=0;

  
   SgStatement *st;
   SgStatement *line;
   

   for (line=s; line;line=line->lexNext()){
     st =  line;
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
         
     {
         action(st);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st);
	   i++;
	   continue;
     }
     else{
         action(st);
         insert_register_and_config_params(st);
       }
     
   }
   return;
}



 SgStatement *find_outermost_scope(SgStatement *s)
{
   SgProgHedrStmt *    prog,*curs;    
   SgProcHedrStmt *    subr;  
   SgProsHedrStmt *    pros;
   SgFuncHedrStmt *    func;  
   SgStatement *cp;


   cp = s;
   curs = NULL;
   do{
      if(subr = isSgProcHedrStmt (cp))
        curs = (SgProgHedrStmt *)subr;
      else if (pros = isSgProsHedrStmt(cp))
        curs = (SgProgHedrStmt *)pros;
      else if(func = isSgFuncHedrStmt(cp))
        curs = (SgProgHedrStmt *)func;
      else if(prog = isSgProgHedrStmt(cp))
        curs = prog;
      
      if (curs == NULL){
      
	cp = cp->controlParent();
      
      }
   }while(!curs);
   return (curs);
   
}

//find whether or not current stmt s is in one loop ,
// if so return line number of the header of innermost loop  
// otherwise return 0
int isInLoop(SgStatement *s)
{

   SgStatement *cp,*curs;
   SgForStmt      *fstmt;
   SgDoWhileStmt  *dwstmt;

   cp   = s;
   curs = NULL;
   do{
     if ((fstmt = isSgForStmt(cp))||(dwstmt = isSgDoWhileStmt(cp)))
       curs =  cp;
     else if(cp != NULL)
   	cp = cp->controlParent();
   }while((cp)&&(!curs));
   if (curs )
     return (curs->lineNumber());
   else 
     return 0;
         
}



void find_exec_next_stmt(SgStatement *s,int *lexnext,int *callnext)
{
  SgStatement *cp;
  SgCallStmt  *cs;
  int i;
  
  if (s->lexNext()->variant() != CONTROL_END){
    if (s->lineNumber() == s->lexNext()->lineNumber()){
       cp = s;
       while((cp->nextInChildList())&& 
          (cp->nextInChildList()->variant() == CONTROL_END))
       cp = cp->controlParent();
       if (cp->nextInChildList())
         *lexnext  = cp->nextInChildList()->lineNumber();
       else
         *lexnext  = -3;     
    }
    
    else
      *lexnext  = s->lexNext()->lineNumber();
  }
      
  else{
    cp = s;
    while((cp->nextInChildList())&& 
         (cp->nextInChildList()->variant() == CONTROL_END))
      cp = cp->controlParent();
    if (cp->nextInChildList())
      *lexnext  = cp->nextInChildList()->lineNumber();
    else
      *lexnext  = -2;
    
  }
  
  if (cs = isSgCallStmt(s)){
    for(i=0;i<nextf;i++)
      if(!strcmp(funcs[i].fname,cs->name()->identifier()))
        break;
    if (i == nextf) 
      *callnext = -1;
    else
      *callnext = funcs[i].feline;
  }
  else
    *callnext  = -2;
  
}


SgStatement *find_first_exec_stmt(SgStatement *s)
{
  int i;
  
   int nlchilds = s->numberOfChildrenList1();
   
   for(i=0;i<nlchilds;i++)
     if ((isSgExecutableStatement(s->childList1(i)))&&
        (s->childList1(i)->variant() != CONTROL_END)&&
        (s->childList1(i)->variant() != USE_STMT)&&
        (s->childList1(i)->variant() != INCLUDE_STAT)&&
	(s->childList1(i)->variant() != PROCESSORS_STMT)&&
        (s->childList1(i)->variant() != TEMPLATE_STMT)  )
	 return (s->childList1(i));
   
   return NULL;
   
}

int give_type_symbol(SgType *vtype)
{
  int variant_type = vtype->variant();
  int var_type;

  switch(variant_type){
           
     case T_CHAR: var_type = T_CHAR; break;
     case T_INT: var_type = T_INT; break;
     case T_FLOAT: var_type = T_FLOAT; break;
     case T_DOUBLE : var_type = T_DOUBLE; break;
     case T_BOOL: var_type = T_BOOL; break;
     case T_ARRAY: var_type = give_type_symbol(vtype->baseType());break;
     case T_KIND:
	int base_type = 
            vtype->thetype->entry.kind_decl.base_type->variant;
        int kind_val = (vtype->thetype->entry.kind_decl.kindexp)->entry.                   list_s.item->entry.literal.value->entry.ival; 
	
        switch(base_type){
	case T_INT:
	  switch(kind_val){
	  case 1:var_type = T_BYTE;break; /* Byte */
	  case 2:var_type = T_SHORT;break; /* Short */
	  case 4:var_type = T_INT;break; /* Integer */
	  case 8:var_type = T_LONG;break; /* Long */
	  }
	  	  
	  break;
	case T_FLOAT:  
	  switch(kind_val){
	  case 4:var_type = T_FLOAT;break; /* Float */
	  case 8:var_type = T_DOUBLE;break; /* Double */
	  }
	  
	  break;
	}
		
	break;
  }

  return (var_type);
 
}


int symb_ndims(SgSymbol *symb)

{
  int vtype =  symb->type()->variant();
  SgArrayType *atype;
  int ndims;
    
  switch(vtype){
  case T_ARRAY: if (atype = isSgArrayType(symb->type()))
                   ndims = atype->dimension();
  break;
  
  default:ndims = 0;break;
  }
  return ndims;
  
}

  
void insert_register_and_config_params(SgStatement *s)
{

    SgProgHedrStmt *    prog,*curs;    
    SgProcHedrStmt *    subr;  
    SgProsHedrStmt *    pros;
    SgFuncHedrStmt *    func;       
    SgSymbol *symb;
    SgVariableSymb *fsymb,*shape_symb;
    SgStatement    *    festmt;
    int var_type;
    int ndims,scope,var_id;
    char *name,indexName[256] ;
    SgValueExp *varType,*varName,*varNdims,*varShape,*varScope;
    SgCallStmt *scstmt;
    SgVarRefExp *e1;
    SgVariableSymb *ptsymb;
    SgFunctionCallExp *fce;
    int distFlag;
    SgStatement *decls;
    int nextTemp=0;
    char *ts ;
     
    curs = NULL;
    if(subr = isSgProcHedrStmt (s))
       curs = (SgProgHedrStmt *)subr;
    else if (pros = isSgProsHedrStmt(s))
      curs = (SgProgHedrStmt *)pros;
    else if(func = isSgFuncHedrStmt(s))
      curs = (SgProgHedrStmt *)func;
    else if(prog = isSgProgHedrStmt(s))
      curs = prog;
    
    if (curs == NULL) return;
    //printf("symbol:%s", (curs->name()).identifier());
    //    printf(" which is a subroutine.\n");
    
        /* find first executable stmt in current scope */
        festmt = find_first_exec_stmt(curs);
        
	
        if (festmt == NULL)
          return;
	
 
         /* add the first executable stmt of this func./subr. to funcs
           record array */
        funcs[nextf].feline = festmt->lineNumber();
	funcs[nextf].fname   = (curs->name()).identifier();
        funcs[nextf++].s = curs;
	
        SgVariableSymb *register_var_symb = new SgVariableSymb("register_var",*curs);
        hpf_server_symb = new SgVariableSymb("hpf_server",*curs);
        jmp_symb =  new SgVariableSymb("setjmp",*curs);        

        // create SgSymbol for function name     
	char *funcName = (curs->name()).identifier();
        char *funcNameSymb  = (char *)malloc(strlen(funcName)+5);
        strcpy(&funcNameSymb[1], funcName);
        funcNameSymb[0] = '\"';
        funcNameSymb[strlen(funcName)+1] = '\\';
        funcNameSymb[strlen(funcName)+2] = '0';
        funcNameSymb[strlen(funcName)+3] = '\"';
        funcNameSymb[strlen(funcName)+4] = 0;    
        SgVariableSymb *fSymb = new SgVariableSymb(funcNameSymb,*curs);
        SgValueExp *funcVarName = new SgValueExp(fSymb->identifier());

        // new addition
        // get the expression list of variables delcared
        for(decls=curs->lexNext();decls!=festmt;decls = decls->lexNext())
	{
	  
           if (!isSgVarDeclStmt(decls))
             continue;
           
	   SgVarDeclStmt *svar = (SgVarDeclStmt *)decls;
           int ns = svar->numberOfSymbols();
	   
           for(int i=0;i<ns;i++)
	   {
              if (svar->symbol(i)->variant() != VARIABLE_NAME) 
                continue;
                 
	      symb = svar->symbol(i);
              //printf("I am registring : %s variant = %d in fun=%s\n",
              //       symb->identifier(),symb->variant(),
              //       (curs->name()).identifier() );
            	          	   
              /* Dont register variables other than variable names */
              if (symb->variant() != VARIABLE_NAME)
                continue;
                          
              fsymb = register_var_symb;
	      scstmt = new SgCallStmt(*fsymb);
	      
              /* add var_type parameters to register_var subroutine */
	      var_type = give_type_symbol(symb->type());
              varType = new SgValueExp(var_type);
	      scstmt->addArg(*varType);

              /* add variable id = parameter var_id */	
              sprintf(indexName, "ind_%d", nextTemp++);
              ptsymb = new SgVariableSymb(indexName,*curs);
	      e1 = new SgVarRefExp(*ptsymb);
	      scstmt->addArg(*e1);
              // make declaration this new created temporary variable
	      // in this function/subroutine. Fotget if for now
              // ptsymb->declareTheSymbol(*curs);
	        

              /* add variable itself = var parameter */
              e1 = new SgVarRefExp(*symb);
	      scstmt->addArg(*e1);
                 
              /* add variable literal name = name parameter */
              ts  = (char *)malloc(strlen(symb->identifier())+5);
              strcpy(&ts[1], symb->identifier() );
              ts[0] = '\"';
              ts[strlen(symb->identifier())+1] = '\\';
              ts[strlen(symb->identifier())+2] = '0';
              ts[strlen(symb->identifier())+3] = '\"';
              ts[strlen(symb->identifier())+4] = 0;                   
   


              SgVariableSymb *dummy = new SgVariableSymb(ts,*curs);

	      varName = new SgValueExp(dummy->identifier());
              //SgExpression *varName = new SgExpression(STRING_VAL,ts);
	      scstmt->addArg(*varName);
	      //printf("ts = %s arg = %s\n",ts,varName->stringValue());
 
              /* add var_ndims parameter */
	      ndims = symb_ndims(symb);
              varNdims = new SgValueExp(ndims);
	      scstmt->addArg(*varNdims);


              // add current function name  
              scstmt->addArg(*funcVarName);

              /* Add distributed flag:is distributed or not */      
              if ((symb->attributes() & ALIGN_BIT)||
                    (symb->attributes() & DISTRIBUTE_BIT ))
		  distFlag = 1;
	      else
		  distFlag = 0;
              SgValueExp *dflag = new SgValueExp(distFlag);       
              scstmt->addArg(*dflag);
   
              /****************************************************/
              /* if its is array ,adds its ubounds and lbounds as argument */
              if ((symb->attributes()& DIMENSION_BIT)||
		  strstr(symb->identifier(),"LOCAL_")){
                 /*prepare LBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the LBOUND of var */
                 shape_symb = new SgVariableSymb("lbound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgValueExp(symb->identifier());
                 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);


                 /*prepare UBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the UBOUND of var */
                 shape_symb = new SgVariableSymb("ubound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgValueExp(symb->identifier());
                 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);
	      } 
              /****************************************************/

              /* finally add register call stmt */
              festmt->insertStmtBefore(*scstmt,*(festmt->controlParent()));

              //scstmt->unparsestdout();

	   }
	}
       	
        // put server function call before end of function unit
        SgStatement *lastStmt = curs->lastExecutable()->lexNext();
        if (curs->lastExecutable()->variant() != RETURN_STAT)
	  InsertFCallNode(lastStmt , HPF_SERVER,BEFORE);
	free(ts);
}


/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
/* Here we are rewriting to insert the registration of only the variables*/
/* we are interedted inone variable I know this is extra but later I will*/
/* reorganise  my coding to make it more structured                      */
/*************************************************************************/
void insert_register_defined_vars(SgStatement *s,
		int numOfNDVars,  char *notDefinedVars[])
{

    SgProgHedrStmt *    prog,*curs;    
    SgProcHedrStmt *    subr;  
    SgProsHedrStmt *    pros;
    SgFuncHedrStmt *    func;       
    SgSymbol *symb;
    SgVariableSymb *fsymb,*shape_symb;
    SgStatement    *    festmt;
    int var_type;
    int ndims,scope,var_id;
    char *name,indexName[256] ;
    SgValueExp *varType,*varName,*varNdims,*varShape,*varScope;
    SgCallStmt *scstmt;
    SgVarRefExp *e1;
    SgVariableSymb *ptsymb;
    SgFunctionCallExp *fce;
    int distFlag;
    SgStatement *decls;
    int nextTemp=0,found,k;
    char *ts ;
     
    curs = (SgProgHedrStmt *)s;
    /* find first executable stmt in current scope */
    festmt = find_first_exec_stmt(curs);
    if (festmt == NULL)
       return;
    SgVariableSymb *register_var_symb = new SgVariableSymb("register_var",*curs);

    printf("symbol:%s", (curs->name()).identifier());
    printf(" which is a subroutine.\n");
  
    //create SgSymbol for function name     
    char *funcName = (curs->name()).identifier();
    char *funcNameSymb  = (char *)malloc(strlen(funcName)+5);
    strcpy(&funcNameSymb[1], funcName);
    funcNameSymb[0] = '\"';
    funcNameSymb[strlen(funcName)+1] = '\\';
    funcNameSymb[strlen(funcName)+2] = '0';
    funcNameSymb[strlen(funcName)+3] = '\"';
    funcNameSymb[strlen(funcName)+4] = 0;    
    SgVariableSymb *fSymb = new SgVariableSymb(funcNameSymb,*curs);
    SgVarRefExp *funcVarName = new SgVarRefExp(fSymb);

    for(decls=curs->lexNext();decls!=festmt;decls = decls->lexNext())
    {
 	  
       if (!isSgVarDeclStmt(decls))
          continue;
           
       SgVarDeclStmt *svar = (SgVarDeclStmt *)decls;
       int ns = svar->numberOfSymbols();
	   
       for(int i=0;i<ns;i++)
       {
              if (svar->symbol(i)->variant() != VARIABLE_NAME) 
                continue;
	      symb = svar->symbol(i);
              /* Dont register variables other than variable names */
              if (symb->variant() != VARIABLE_NAME)
                continue;
         
              for(found=0,k=0;k<numOfNDVars;k++){
		if (!strcmp(notDefinedVars[k],symb->identifier())){
                   found = 1;
                   break;
		}
	      }
              /* If the variable is defined in the program unit we
		 are requesting the interpretation HPF Stmts from */
	      /* Register only the variabels user define at first time */
              printf("Variable : %s result :%d\n",symb->identifier(),found);
              if (!found )
		continue;

              /* set SAVE attribute */ 
             /*(symb->thesymb)->attr = (symb->thesymb)->attr |SAVE_BIT; */     
                           
              fsymb = register_var_symb;
	      scstmt = new SgCallStmt(*fsymb);
	      
              /* add var_type parameters to register_var subroutine */
	      var_type = give_type_symbol(symb->type());
              varType = new SgValueExp(var_type);
	      scstmt->addArg(*varType);

              /* add variable id = parameter var_id */	
              sprintf(indexName, "ind_%d", nextTemp++);
              ptsymb = new SgVariableSymb(indexName,*curs);
	      e1 = new SgVarRefExp(*ptsymb);
	      scstmt->addArg(*e1);
              // make declaration this new created temporary variable
	      // in this function/subroutine. Fotget if for now
              // ptsymb->declareTheSymbol(*curs);
	        

              /* add variable itself = var parameter */
              e1 = new SgVarRefExp(*symb);
	      scstmt->addArg(*e1);
                 
              /* add variable literal name = name parameter */
              ts  = (char *)malloc(strlen(symb->identifier())+5);
              strcpy(&ts[1], symb->identifier() );
              ts[0] = '\"';
              ts[strlen(symb->identifier())+1] = '\\';
              ts[strlen(symb->identifier())+2] = '0';
              ts[strlen(symb->identifier())+3] = '\"';
              ts[strlen(symb->identifier())+4] = 0;                   
   


              SgVariableSymb *dummy = new SgVariableSymb(ts,*curs);
 
              SgVarRefExp *tvarName = new SgVarRefExp(dummy);
 	      scstmt->addArg(*tvarName);
	 
              /* add var_ndims parameter */
	      ndims = symb_ndims(symb);
              varNdims = new SgValueExp(ndims);
	      scstmt->addArg(*varNdims);


              // add current function name  
              scstmt->addArg(*funcVarName);

              /* Add distributed flag:is distributed or not */      
              if ((symb->attributes() & ALIGN_BIT)||
                    (symb->attributes() & DISTRIBUTE_BIT ))
		  distFlag = 1;
	      else
		  distFlag = 0;
              SgValueExp *dflag = new SgValueExp(distFlag);       
              scstmt->addArg(*dflag);
   
              /****************************************************/
              /* if its is array ,adds its ubounds and lbounds as argument */
              if ((symb->attributes()& DIMENSION_BIT)||
		  strstr(symb->identifier(),"LOCAL_")){
                 /*prepare LBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the LBOUND of var */
                 shape_symb = new SgVariableSymb("lbound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgValueExp(symb->identifier());
                 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);


                 /*prepare UBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the UBOUND of var */
                 shape_symb = new SgVariableSymb("ubound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgValueExp(symb->identifier());
                 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);
	      } 
              /****************************************************/

              /* finally add register call stmt */
              festmt->insertStmtBefore(*scstmt,*(festmt->controlParent()));

              //scstmt->unparsestdout();

       }
    }
    free(ts);
}



SgVariableSymb *DeclareTemporaryLike(SgSymbol * x,int sd)
{
       SgStatement  *old_decl_stmt;
       SgVariableSymb *tmp;
       SgVarDeclStmt *decl_stmt;
       SgExpression *exp;
       char *name;
       SgExprListExp *elist;
       SgType *old_type;
       SgVarDeclStmt *decl;
       SgProgHedrStmt *ispg;
              
       // now find the declaration statement for x and get
       // the correct initializing expression.
       
       old_decl_stmt = x ->declaredInStmt();
       // get the expression list of variables delcared
       exp = old_decl_stmt->expr(0);
       // look for the reference to our symbol x.
       while(elist = isSgExprListExp(exp)){
       	 if((elist->value())->IsSymbolInExpression(*x) )
               break;       
            exp = elist->next();
            }    
       if(exp == NULL) return NULL;

       // now create a new expression list and 
       // and a new declaration statement
       elist = new SgExprListExp(exp->copy());
       if (decl = isSgVarDeclStmt(old_decl_stmt))
         old_type = decl->type();
       decl_stmt = new SgVarDeclStmt(*elist,*old_type);

       // create a new variable of the same class
       // generate the name as "_Ti" where i is counter.
       name = (char *)malloc(256);
       if (sd == SLOCAL)     
	 sprintf(name, "LOCAL_%s", x->identifier());
       else
	 sprintf(name, "DISTR_%s", x->identifier());
       tmp = new SgVariableSymb(name, (x->type())->copy(), *(x->scope()));
    
       //Test if new declaration can be added
       SgArrayRefExp *arefexp = new  SgArrayRefExp(*(&tmp->copy())); 
       //prepare size(A) call fucntion for dim_info parameter
       //Here we are trying to calculate the shape of var 
       for(int i=1;i<=symb_ndims(x);i++){
            SgVariableSymb *shape_symb = new SgVariableSymb("size");
	    SgFunctionCallExp *fce   = new SgFunctionCallExp(*shape_symb );
	    //SgValueExp *sve = new SgValueExp(x->identifier());
            SgValueExp *sve;
            SgVarRefExp   *sv = new SgVarRefExp(x);
            printf("LOCAL : %s\n",x->identifier());
	    //fce -> addArg(*sve);
            fce -> addArg(*sv);
            sve = new SgValueExp(i);
            fce -> addArg(*sve); 
            arefexp->addSubscript(*fce);
	}
        
	SgVarDeclStmt *ndecl = new SgVarDeclStmt(*arefexp,
               *(x->type()->baseType()));
	
	old_decl_stmt->insertStmtAfter(*ndecl,
                *(old_decl_stmt->controlParent()));

        return tmp;

    }

void insert_decls(SgStatement *first,int numOfNDVars,char *notDefinedVars[])
{
   SgStatement *prev,*s1;
   SgSymbol *symb;
   SgVarDeclStmt *decl;
   SgProgHedrStmt *p;
   int found ,k;
         
   for(prev = first;prev;prev = prev->lexNext()){
     SgStatement *s = prev->controlParent();
     if (s&&((s->variant() == INTERFACE_STMT)||
         (s->variant() == MODULE_STMT)))
        continue;
    
     if (s&&(s1=s->controlParent())&&((s1->variant() == INTERFACE_STMT)||
         (s1->variant() == MODULE_STMT)))
       continue;
    
      if  ((prev->variant() == ALIGN_DECL)||   
	  (prev->variant() == DISTRIBUTE_DECL)||
          (prev->variant() == COMBINED_DIRECTIVE_DECL)||
	  (prev->variant() == ALIGN_DECL)||
          (prev->variant() == TEMPLATE_STMT) )
       {
        
         decl = (SgVarDeclStmt *)prev;
       	 for(int i=0;i<decl->numberOfSymbols();i++){
            if (isTemplate(decl->symbol(i)))
              continue;

            /* If the variable is declared in function ,that is
            it is not declared in function inside which we are sending
            HPF Stmts to interpret */

            for(found=0,k=0;k<numOfNDVars;k++){
              if (!strcmp(notDefinedVars[k],decl->symbol(i)->identifier())){
                  found = 1;
                  break;
	      }	
	    }
            if (!found)
                continue;
            DeclareTemporaryLike(decl->symbol(i),SLOCAL);
	 }

	 
     }
   }
  
}

int isTemplate(SgSymbol *symb)
{
   SgStatement *prev = symb->declaredInStmt();

   if  ((prev->variant() == ALIGN_DECL)||   
	(prev->variant() == DISTRIBUTE_DECL)||
        (prev->variant() == COMBINED_DIRECTIVE_DECL)||
	(prev->variant() == ALIGN_DECL)||
        (prev->variant() == TEMPLATE_STMT) )
     return 1;
   else
     return 0;
}
   


int fileLength(SgFile *f,int num_of_routines)
{
  SgStatement *s = f->functions(num_of_routines-1);
  s = s->childList1(s->numberOfChildrenList1()-1);
  return (s->lineNumber()+1);

}


void create_functionTable_file(SgProject * project,int *nfuncs, int *nfiles)
{
  int numOfFiles = project->numberOfFiles();
  int num_of_routines;
  int filelen,nroutines=0,nf=0;
  SgStatement *child;
  int rem,k,ninterfaces=0;

  
  for(int i=0 ; i< numOfFiles;i++){
    SgFile *f = &(project->file(i));
    nroutines += (f->numberOfFunctions());
  }

  
  parfunctab = (FTAB *)calloc(nroutines,sizeof(FTAB));
  parsefiles = (FILES *)calloc(numOfFiles,sizeof(FILES));
  lengthOfFiles = (int *)malloc(numOfFiles*sizeof(int));
  *nfiles = numOfFiles;

  for(int i=0;i< numOfFiles;i++){
    SgFile *f;
    SgStatement *mp;

    f = &(project->file(i));
    if (f -> mainProgram())
      parsefiles[i].isIncludeMain = 1;
    else
      parsefiles[i].isIncludeMain = 0;   


    num_of_routines = f->numberOfFunctions();

    // calculate filelen and structures    
    filelen = fileLength(f, num_of_routines);
    parsefiles[i].lineInfo = (SINFO *)calloc(filelen,sizeof(SINFO));
    strcpy(parsefiles[i].fileName,project->fileName(i));
    lengthOfFiles[i] = filelen;

    for(int j=0;j < num_of_routines;j++){
      SgStatement *sub;
      SgSymbol    *subsymb;

      sub     = f->functions(j);
      subsymb = sub->symbol();

      if (sub->controlParent()->variant() !=  INTERFACE_STMT){
         parfunctab[nf].lineNum = sub->lineNumber();
         strcpy(parfunctab[nf].funcName,subsymb->identifier());
         strcpy(parfunctab[nf++].fileName,project->fileName(i));
      }
      else 
        ninterfaces++;

      SgStatement *ls = sub->childList1(sub->numberOfChildrenList1()-1);
      ls = ls->lexNext();

      SgStatement *ts = sub;
      for (; ts != ls; ts=ts->lexNext()){
         child = ts;
         k = child->lineNumber();
        
         parsefiles[i].lineInfo[k].lineNum = child->lineNumber();
         parsefiles[i].lineInfo[k].stmtID  = child->variant();
         parsefiles[i].lineInfo[k].parent  = child->controlParent()->lineNumber();
         parsefiles[i].lineInfo[k].isExecutable = 0;
       }
    }
  }
   *nfuncs = nroutines-ninterfaces;
}

void write_summary(char *outFile,int nfuncs,int nfiles)
{

    FILE *fp = fopen(outFile,"a+");

    fprintf(fp,"//File name \n");
    fprintf(fp,"%s\n",parsefiles[0].fileName);
    fprintf(fp,"//Length of File \n");
    fprintf(fp,"%d  %d\n",lengthOfFiles[0]-1,parsefiles[0].isIncludeMain);

    fprintf(fp,"// number of functions \n");
    fprintf(fp,"%d\n",nfuncs);
    fprintf(fp,"//LineNumber  FunctionName \n");
    for(int i=0;i<nfuncs;i++)
      fprintf(fp,"%d   %s  \n", parfunctab[i].lineNum,parfunctab[i].funcName);

    fprintf(fp,"// LineNumber  stmtID parent isExecutable \n");
    for(int i=0;i<nfiles;i++)
      for(int k=1;k<lengthOfFiles[i];k++)
        fprintf(fp,"%d   %d   %d   %d   %d\n",
          parsefiles[i].lineInfo[k].lineNum, 
          parsefiles[i].lineInfo[k].stmtID  ,
          parsefiles[i].lineInfo[k].parent  ,
          parsefiles[i].lineInfo[k].isExecutable ,
          parsefiles[i].lineInfo[k].isInLoop );
    

    fclose(fp);
}

void insertInitStmt(SgFile *file)
{
   SgStatement *s,*fs;
   SgVariableSymb *funcname,*symbGetarg;
   SgCallStmt *getargFunc,*func;

   if (s = file -> mainProgram()){
      // find the first executable statement 
       fs = find_first_exec_stmt(s);

       // make function call init_hpf_server(all the command line args) 
       funcname = new SgVariableSymb("init_hpf_server",
                                   *(file->firstStatement()));
       func = new SgCallStmt(*funcname);
       SgVarRefExp *portarg = new SgVarRefExp(
              new SgVariableSymb("portnum",*(file->firstStatement()) ));
       func->addArg(*portarg);


       char *pbuf  = (char *)malloc(strlen(projectName)+5);
       strcpy(&pbuf[1], projectName);
       pbuf[0] = '\"';
       pbuf[strlen(projectName)+1] = '\\';
       pbuf[strlen(projectName)+2] = '0';
       pbuf[strlen(projectName)+3] = '\"';
       pbuf[strlen(projectName)+4] = 0;    
       SgVariableSymb *psymb = new SgVariableSymb(pbuf);
       SgValueExp *pexp = new SgValueExp(psymb->identifier());
       func->addArg(*pexp);

       
       func->addComment("include \"lib3f.h\"\n");
       func->addComment("character*16 portnum \n");
       func->addComment("call getarg(1,portnum)\n");

       fs->insertStmtBefore(*func,*(fs->controlParent()));
   }   
}

       
void readParseTree(char *projectFileName,int *nfiles)
{
 
   SgProject * project = new SgProject(projectFileName);

   create_functionTable_file(project,nfiles);

   /*write_summary("EROL", nfuncs, nfiles);  */
}



create_functionTable_file(SgProject * project, int *nfiles)
{
  int numOfFiles = project->numberOfFiles();
  int num_of_routines;
  int filelen;
  SgStatement *child;
  int rem,k;
  
  files = (FILES *)calloc(numOfFiles,sizeof(FILES));
  *nfiles = numOfFiles;

  for(int i=0;i< numOfFiles;i++){
    SgFile *f;
    SgStatement *mp;
    int nf;

    nf = 0;

    f = &(project->file(i));
    if (f -> mainProgram())
      files[i].isIncludeMain = 1;
    else
      files[i].isIncludeMain = 0;   

    num_of_routines = f->numberOfFunctions();

    // calculate filelen and structures    
    filelen = fileLength(f, num_of_routines);
    files[i].lineInfo = (SINFO *)calloc(filelen,sizeof(SINFO));
    files[i].functab  = (FTAB *)calloc(num_of_routines,sizeof(FTAB));

    strcpy(files[i].fileName,project->fileName(i));
    files[i].fileSize = filelen;

    char *sptr = strrchr(files[i].fileName,'.');
    strncpy(files[i].depFileName,files[i].fileName,
	 sptr - (files[i].fileName));
    strcat(files[i].depFileName,".dep");

    for(int j=0;j < num_of_routines;j++){
      SgStatement *sub;
      SgSymbol    *subsymb;

      sub     = f->functions(j);
      subsymb = sub->symbol();
      files[i].functab[nf].fExecStmt = find_first_exec_stmt(sub)->lineNumber();
      if (sub -> variant() == PROG_HEDR)
         files[i].mainFuncIndex = j;

      if (sub->controlParent()->variant() !=  INTERFACE_STMT){
         files[i].functab[nf].lineNum = sub->lineNumber();
         files[i].functab[nf].ParseTreeStmt = NULL;
         strcpy(files[i].functab[nf].funcName,subsymb->identifier());
         
         strcpy(files[i].functab[nf++].fileName,project->fileName(i));
        
      }

      SgStatement *ls = sub->childList1(sub->numberOfChildrenList1()-1);
      ls = ls->lexNext();

      SgStatement *ts = sub;
      for (; ts != ls; ts=ts->lexNext()){
         child = ts;
         k = child->lineNumber();
        
         files[i].lineInfo[k].lineNum = child->lineNumber();
         files[i].lineInfo[k].stmtID  = child->variant();
         files[i].lineInfo[k].parent  = child->controlParent()->lineNumber();
         files[i].lineInfo[k].isExecutable = 0; 
         files[i].lineInfo[k].isInLoop = isInLoop(child);
         if ((child->variant() == RETURN_STAT)||
	     (child->variant() == STOP_STAT)||
	     (ts->lexNext() == ls))
	   files[i].lineInfo[k].isReturnPoint = 1;
	 else
           files[i].lineInfo[k].isReturnPoint = 0;
       } 
    }
    files[i].nfuncs = nf;
    transform_registration(f->firstStatement(),i);       
  }
}



void action(SgStatement * s,int find)
{
  SgDeclarationStatement *decl;
  SgProgHedrStmt *spg;
  SgCallStmt *cs;
  SgForStmt *fs;
  int flag = 0;
  SgAssignStmt *ass ;
  SgForStmt *sfo ; 
  SgCallStmt *call;
  
  SgStatement *cp = s->controlParent();
  if ((isSgExecutableStatement(s))&&
      (s->variant() != CONTROL_END)&&
      (s->variant() != MODULE_STMT)&&
      (s->variant() != USE_STMT)&&
      (s->variant() != INCLUDE_STAT)&&
      ((!cp)||(!(cp->lineNumber()==s->lineNumber()))) )
          files[find].lineInfo[s->lineNumber()].isExecutable = 1; 
     
}

void transform_registration(SgStatement * s,int find) 
{
  int i=0;

  int org_numberOfChildren1=s->numberOfChildrenList1();
  int org_numberOfChildren2= s->numberOfChildrenList2();
  int org_numberOfChildren = org_numberOfChildren1 + org_numberOfChildren2;
  SgStatement * * children = new SgStatement * [org_numberOfChildren];
  SgStatement *st;
  SgStatement *line;
   
  for (i=0; i<org_numberOfChildren1; i++)
	children[i]=s->childList1(i);

  for (; i<org_numberOfChildren; i++)
     children[i]=s->childList2(i-org_numberOfChildren1);

   for (i=0; i<org_numberOfChildren; i++){

     st =  children[i];
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
         
     {
         action(st,find);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st,find);
	   i++;
	   continue;
     }
     else{
         action(st,find);
         transform_registration(st,find);
     }
   }
   delete [] children;
   return;
   
}







