#include <sys/types.h>
#include <rpc/types.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpc/xdr.h> 
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/wait.h>
#include "tags.h"
#include <string.h>
#include <dlfcn.h>
#include <setjmp.h>

#include <pthread.h>
  
#define DEBUG 1  
#define MAXFILES 128
#define HOST   (pghpf_myprocnum_() == 0)  
#define ME     (pghpf_myprocnum_())
#define SERVER_CONNECT_RETRY_DELAY 1
#define QLEN            5
#define MAX_NAME_LENGTH 256
#define MAX_DIMS        7
#define MAX_VARS        256
#define REGBUFSIZE      10240   
#define UNCONNECTED     0
#define CONNECTED       1
#define MAXVARLEN       256
#define BREAKON         1
#define BREAKOFF        2  
#define STAYINSERVER    1
#define EXITFROMSERVER  2
#define RUNTIME_ERROR   -1
#define COMPILE_ERROR   -2

/* Debugger commands */
#define PUTBREAKONOFF  11
#define STEPOVER       12
#define NSTEPOVER      13 
#define STEPINTO       14
#define NSTEPINTO      15
#define CONTINUE       16
#define RUN            17
#define STOP           18
#define NITERATION     19
#define GOUNTIL        20
#define DISPLAY        21
#define SETVARIABLE    22
#define SETPARVAR      23 /* update one parallel variable from java applet */
#define CONNECT        24 /* make conenction to HPF server */
#define GOTOCURSOR     25
#define INTERPRET      26 
#define ENDOFPROGRAM   27
#define DISARRAYSECTION 28

#define GETNEWFILE     -1
#define STOPPROGRAM    -2
#define ABNORMALTERM   -3
#define ARRAYSECERROR  -4
#define NUSEDDIM      -999
typedef struct SgStatement* stmtObj;   

typedef struct push_data{
  int  type;
  char name[MAXVARLEN];
  int  ndims;  /* 0 for simple variables,ndims for array vars */ 
  int  dim_info[MAX_DIMS] ;
  int  lbounds[MAX_DIMS];
  int  ubounds[MAX_DIMS];
  int  sockfd; /* socket file descriptor */
  int  port;
  int  sizeof_data; /* total size of push_data in bytes */
  long buf; /* includes actual data whatever it is */
  int  elsize; /* sizeof of each element */
  int  connected;
  int  nelems ; /* total element count */
  int  tind; /* index to type_list,returns the pointer to function */
  int  javabuflen;
  unsigned long  client_addr;
  char funcName[256];
  
  long desc;
  int  isDistributed;
  int  localIndex;/* If I am distributed,this points to my local data */
  /* this will be used to transfer from parallel dat ato local data */
  char attribute[5096];
  int save; /* It was saved before by using "SAVE" attribute of F90 */

}DATA;

typedef struct send_buf{   
  char *sbuf;
  int   len;
  XDR   xdrs;
}SBUF;

typedef struct functable{
 int lineNum;
 int fExecStmt; 
/*First executable statement which is used to update control stack info.*/
 char funcName[128];
 char fileName[128];
 stmtObj ParseTreeStmt;
}FTAB;

typedef struct sourceinfo{
 int lineNum;       /* line number */
 int stmtID;        /* statement ID */
 int parent;        /* its parent line number */
 int isExecutable;  /* Is this line executable */
 int isInLoop;       /* is this in loop */
 int onbreak;        /* has break on or off */
 int isReturnPoint;   /* this line has return stmt or end of one program unit */
 jmp_buf   jmpbuffer ; /* jump buffer to go back */
}SINFO;

typedef struct intermediateFiles{
  char   fileName[64];
  char   depFileName[64];
  int    isIncludeMain;
  int    mainFuncIndex;
  int    fileSize;
  SINFO  *lineInfo;
  FTAB   *functab;
  int    nfuncs;
}FILES ;

/* Control stack information */
struct StackInfo{
    char *funcName;
    char *fileName;
    int  lineNum;
};

typedef struct  CONTROLSTACKINFO{
  struct StackInfo sInfo[128];
  int top;
}CONTROLSTACK;

typedef struct errtypedef{
      int type;
      int status;
}ERRBUF;


typedef struct paramsbuf{
  int myline,sockfd;
  char *prevFileName,*curFileName,*funcName;
}PARAMBUF;
