#include <stdio.h>
#include "compilerApi.h"
#include <malloc.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>

#define SLOCAL       1
#define DISTRIBUTED  2
#define BEFORE       0
#define AFTER        1

// Find the enclosing function/subroutine unit

// decl for lexical analyser
#define VARID   509
#define MAXVARS 256
char *input_file;
FILE *inf_ptr;
#undef YYLMAX
#define YYLMAX 8192
extern char yytext[YYLMAX];
extern int yylineno;


SgFile *curSgFile;
extern "C" char *generateNewFunction(char *hpfStmtList,char *calledfunc,
           char *funcName,char *usedVars[],int *numOfusedVars,int ME);
extern "C" void openNewProject(char *depfileName);
extern "C" int find_var(char *origProgName,char *usedVars);
extern "C" int createInterpretFunc(int nextModule, char *moduleNames[], 
				   char *filename); 
int isDecl(SgStatement *chld);
insertList(char *list[],int *len,char *key);
void findUsedVars(char *inputFile,char *usedVars[],int *nVars);
int parse_find_var(char *origProgName,char *usedVars);

struct Internalfunctable{
  SgStatement *startAddr;
  char *fname;
}curFuncTable[256];
static int    numOfFuncs;
static int nextFIndex=0;
      
void openNewProject(char *depFileName)
{
   int j;
   char *myargs[3]={NULL,NULL,(char *)0};
   myargs[1] =  (char *)malloc(256);

   strcpy(myargs[1],depFileName);
   SgProject *project = new SgProject("",2,myargs);
   curSgFile = &(project->file(0));

   for(j=0;j<curSgFile->numberOfFunctions();j++){
     curFuncTable[j].startAddr = curSgFile->functions(j);
     SgSymbol *subsym  =  (curFuncTable[j].startAddr)->symbol();
     curFuncTable[j].fname     = subsym->identifier();
   }
   numOfFuncs = curSgFile->numberOfFunctions();

} 
  

int findex(char *name)
{
   int found = -1;  

   for(int j=0;j<numOfFuncs;j++){
     if (!strcmp(name,curFuncTable[j].fname )){
         found = j;
	 break;
     }
   }  
  
   if ((found<0)||(found==numOfFuncs)){
      printf("I could not find file : %s\n",name);
      exit(0);
   }
   return found;

}   
char *generateNewFunction(char *hpfStmtList,char *calledFunc,
                          char *funcName,char *usedVars[],int *numOfusedVars,
                          int ME) 
{
  SgFunctionSymb *fsymb;
  SgStatement    *prev,*s,*ptfunc;
  SgStatement *curStmt;
  char fname[256],tmp[256];
  int retval,cnt = 0,nVars ;
  char *tempBuf1,*tempBuf2;
  SgType *type;
  FILE *fTemp;   
 

  /* Store the user's HPF stsmts into temporary File */
  char *ts = new char[strlen(hpfStmtList)+1];
  strcpy(ts,hpfStmtList);
  sprintf(fname,"tmpFile_%d",ME);
  fTemp = fopen(fname,"w");
  fprintf(fTemp,"%s",ts);
  fclose(fTemp);
  free(ts);
  
  printf("in generateNewFunction ME=%d list=%s \n",ME,hpfStmtList);
 
  //Find the variables used in the HPF stmts user gave in intercative mode
  findUsedVars(fname ,usedVars,&nVars);
  *numOfusedVars = nVars;    

  // get the scope and name of enclosing program unit
  int index = findex(calledFunc);
  SgStatement *origProgUnit = curFuncTable[index].startAddr;
  char *origProgName = (((SgProgHedrStmt *)origProgUnit)->name()).identifier();

  // Create first new function
  char *newSubName          = (char *)malloc(512);
  sprintf(newSubName, "newfunction_%d", nextFIndex++);
  strcpy(funcName,newSubName);
  type = new SgType(T_VOID);
  fsymb   = new SgFunctionSymb(PROCEDURE_NAME,newSubName,
			       *type,*(origProgUnit->controlParent()));
  ptfunc  = new SgProcHedrStmt(*fsymb);

  // Add used variables as parameters to new function
  for(int i=0;i<nVars;i++){
      int index = parse_find_var(origProgName,usedVars[i]);
      if (index >= 0){    
          SgVariableSymb *ns = new SgVariableSymb(usedVars[i]);
          fsymb->insertParameter(cnt++,*ns);
       }
  }
  //add the user's HPF stmts as comments before "subroutine End" stmt
  SgStatement *endStmt =ptfunc->childList1(ptfunc->numberOfChildrenList1()-1);
  endStmt ->addComment(hpfStmtList); 
  tempBuf1 = ptfunc->unparseToBuf();
  printf("temp1 = \n %s\n",tempBuf1);

  ///////////////////////////////////////////////////////
  /////////////////////////////////////////////////////
  // Create second new function
  sprintf(newSubName, "newgen_%s", origProgName);
  fsymb   = new SgFunctionSymb(PROCEDURE_NAME,newSubName,
			       *type,*(origProgUnit->controlParent()));
  ptfunc  = new SgProcHedrStmt(*fsymb);
  // add all the declaration stmts of original function
  prev = ptfunc;
  for(int i=0;i<origProgUnit->numberOfChildrenList1()-1;i++){
    SgStatement *chld = origProgUnit->childList1(i);
    if (isDecl(chld))
     {
        s = &origProgUnit->childList1(i)->copy();
        prev->insertStmtAfter(*s,*ptfunc);
        prev = s;
     }
  }  
  //Unparse new created function to standard output
  tempBuf2 = ptfunc->unparseToBuf();
  printf("temp2 = \n %s\n",tempBuf2);

  int len = strlen(tempBuf1)+strlen(tempBuf2);
  char *fileOutput = (char *)malloc(len+1);
  strcpy(fileOutput,tempBuf1);
  strcat(fileOutput,tempBuf2);
  return  fileOutput;     

}

int createInterpretFunc(int nextModule, char *moduleNames[],char *filename)
{
    int pid,status,ret;
    char cmnd[256],depfile[256],*tempBuf;
    char command[5096];
 
    strcpy(depfile,filename);
    *(strchr(depfile,'.')) = 0;
    strcat(depfile,".dep");
    printf("DEPFILE : %s\n",depfile);

    /* make dep. file */
    sprintf(cmnd,"hpf2dep %s",filename);
    system(cmnd);

    strcpy(command,"instrumentInterpreter  ");
    strcat(command,depfile);
    
    for(int i=0;i<nextModule;i++){
      strcat(command," ");
      strcat(command,moduleNames[i]);
    }
    if ((ret=system(command)) < 0)
      printf("system error : %d\n",ret);

    printf("After last function..\n");
    return 0;

}


void findUsedVars(char *inputFile,char *usedVars[],int *nVars)
{

  int cnt=0,token;
  
  input_file = (char *)malloc(strlen(inputFile)+1);
  strcpy(input_file,inputFile);
 
  if (!(inf_ptr = fopen(input_file,"r"))){
    printf("I can not open input file : %s\n",input_file);
    exit(0);
  }
  
  while((token=yylex())){
    if (token == VARID)
      insertList(usedVars,&cnt,yytext);
  }

  fclose(inf_ptr);  
  *nVars = cnt;

}

insertList(char *list[],int *len,char *key)
{
  int found = 0;
  
  for(int i=0;i<(*len);i++)
    if (!strcmp(list[i],key)){
      found = 1;
      break;
    }
  if (!found){
    list[(*len)] = (char *)malloc(strlen(key)+1);
    strcpy(list[(*len)],key);
    (*len)++;
  }
}

