#include <stdio.h>
#include "compilerApi.h"
#include "deftypes.h"
#include <malloc.h>
#include <string.h>

#define TIMER_CLEAR    0
#define TIMER_START    1
#define TIMER_STOP     2
#define TIMER_ELAPSED  3
#define REGISTER       4
#define HPF_SERVER     5
#define INIT           6
#define SLOCAL       1
#define DISTRIBUTED 2

#define BEFORE  0
#define AFTER   1
#define MAX_FUNC_NUM 300
#define MAX_PAIR     300

enum FLAG {off, on};
FLAG in_main_program;
FLAG in_procedure;
int tabnumber=0;
void doFullType(SgType *t);
void doSymbAttribs(SgSymbol *s);
void doRoutineHeader(SgStatement *s);
void doSymb(SgSymbol *s);
void register_var();
void timer_start(),timer_clear(),timer_stop(),timer_elapsed();
void configure();
void insert_register_and_config_params(SgStatement *);
int give_type_symbol();
SgStatement *find_first_exec_stmt(SgStatement *);
void hpf_server();
void find_exec_next_stmt(SgStatement *,int *,int *);
SgStatement *find_outermost_scope(SgStatement *);
void create_one_func(SgStatement *);
SgVariableSymb *DeclareTemporaryLike(SgSymbol * ,int );
void insert_decls(SgStatement *moduleDecls);
void build_function(int ,SgStatement *);
int isInLoop(SgStatement *);
int isTemplate(SgSymbol *symb);

create_functionTable_file(SgProject * project, int *nfiles);
int isInLoop(SgStatement *s);
SgStatement *find_outermost_scope(SgStatement *s);
void action(SgStatement * s,int find);
void transform_registration(SgStatement * s,int find) ;
write_summary(char *outFile,int nfiles);
SgStatement *find_first_exec_stmt(SgStatement *s);
void insert_register_defined_vars(SgStatement *procHedr);
			
extern "C" void readParseTree(char *projectFileName ,int *nfiles,
			      FILES *in_files,FTAB  *in_functab);
void addcontains(char *s);
int isDecl(SgStatement *chld);

//char projectName[256];
typedef struct festmts
{
  char *fname;
  int  feline;
  SgStatement *s;
}FUNCS;

SgStatement *first;
int nextp;
FUNCS funcs[MAX_FUNC_NUM];
int nextf ;
FILES *parsefiles;
int   *lengthOfFiles;
FTAB  *parfunctab;
int fileIndex;
SgVariableSymb *hpf_server_symb,*jmp_symb;
char *parcurFileName,*parcurFuncName;

FILES *files;
FTAB  *functab;


SgStatement *whereDeclared(SgSymbol *x,SgStatement *start);


// Insert a C function call statement
void 
InsertFCallNode(SgStatement *s,int wfunction,int how)
{
  int myline = s->lineNumber();
  SgSymbol *symb;
  char *fsymb ;
  SgStatement *f;

  switch (wfunction) {
    case TIMER_CLEAR: symb = new SgVariableSymb("timer_clear");
                        break;
    case TIMER_START: symb = new SgVariableSymb("timer_start");
                        break;
    case TIMER_STOP:  symb = new SgVariableSymb("timer_stop");
                        break;
    case TIMER_ELAPSED:symb = new SgVariableSymb("timer_elapsed"); 
                         break;
    case REGISTER : symb = new SgVariableSymb("REGISTER");
      break;
    case HPF_SERVER: symb = hpf_server_symb;
      break;
    case INIT: symb = new SgVariableSymb("init_hpf_server");
      break;
    
 }

 

  // you can put break only at points before which HPF_Server 
  // call is inserted   
  parsefiles[fileIndex].lineInfo[myline].isExecutable = 1;
  parsefiles[fileIndex].lineInfo[myline].isInLoop = isInLoop(s);
  SgCallStmt *func = new SgCallStmt(*symb);

  /* add myline parameter */  
  SgValueExp *ev = new SgValueExp(myline);
  func->addArg(*ev);

  f = find_outermost_scope(s);
  SgProgHedrStmt *fph = (SgProgHedrStmt *)f;
  fsymb = (fph->name()).identifier();

  // add current function name  
  char *ts  = (char *)malloc(strlen(fsymb)+5);
  strcpy(&ts[1], fsymb );
  ts[0] = '\"';
  ts[strlen(fsymb)+1] = '\\';
  ts[strlen(fsymb)+2] = '0';
  ts[strlen(fsymb)+3] = '\"';
  ts[strlen(fsymb)+4] = 0;    

  SgVariableSymb *dummy = new SgVariableSymb(ts,*f);
  SgValueExp *varName = new SgValueExp(dummy->identifier());
  func->addArg(*varName);

  // add current file name
  ts  = (char *)malloc(strlen(parcurFileName)+5);
  strcpy(&ts[1], parcurFileName);
  ts[0] = '\"';
  ts[strlen(parcurFileName)+1] = '\\';
  ts[strlen(parcurFileName)+2] = '0';
  ts[strlen(parcurFileName)+3] = '\"';
  ts[strlen(parcurFileName)+4] = 0;    

  dummy = new SgVariableSymb(ts,*f);
  varName = new SgValueExp(dummy->identifier());
  func->addArg(*varName);

  switch (how) {
    case BEFORE:
         s->insertStmtBefore(*func,*(s->controlParent()));
         break;
    case AFTER:
         s->insertStmtAfter(*func,*(s->controlParent()));
         break;
  }
 
  /***********************************************************/
  /* Add setjump function call to store stack information */
   SgCallStmt *jmp_func = new SgCallStmt(*jmp_symb);
   SgValueExp *jmp_ev = new SgValueExp(myline);
   jmp_func->addArg(*jmp_ev);
   func->insertStmtBefore(*jmp_func,*(s->controlParent()));
  /***********************************************************/

}

void action(SgStatement * s)
{
  SgDeclarationStatement *decl;
  SgProgHedrStmt *spg;
  SgCallStmt *cs;
  SgForStmt *fs;
  int flag = 0;
  SgAssignStmt *ass ;
  SgForStmt *sfo ; 
  SgCallStmt *call;
  
  if (call = isSgCallStmt(s))
     parcurFuncName = call->name()->identifier();
      
  if (cs=isSgCallStmt(s)){
      if ((!strcmp(cs->name()->identifier(),"register_var"))||
          (!strcmp(cs->name()->identifier(),"setjmp"))||
          (!strcmp(cs->name()->identifier(),"hpf_server")))
	   
        return;
  }
  
  SgStatement *cp = s->controlParent();
  if ((isSgExecutableStatement(s))&&
      (s->variant() != CONTROL_END)&&
      (s->variant() != MODULE_STMT)&&
      (s->variant() != USE_STMT)&&
      (s->variant() != INCLUDE_STAT)&&
      ((!cp)||(!(cp->lineNumber()==s->lineNumber()))) )
          InsertFCallNode(s , HPF_SERVER,BEFORE);
 

  // put server function call before end of function unit
    
}

void transform_registration(SgStatement * s) 
{
  int i=0;

  int org_numberOfChildren1=s->numberOfChildrenList1();
   int org_numberOfChildren2= s->numberOfChildrenList2();
  int org_numberOfChildren = org_numberOfChildren1 + org_numberOfChildren2;
   SgStatement * * children = new SgStatement * [org_numberOfChildren];
   SgStatement *st;
   SgStatement *line;
   
  for (i=0; i<org_numberOfChildren1; i++)
	children[i]=s->childList1(i);

  for (; i<org_numberOfChildren; i++)
     children[i]=s->childList2(i-org_numberOfChildren1);

   for (i=0; i<org_numberOfChildren; i++){

     st =  children[i];
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
         
     {
         action(st);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st);
	   i++;
	   continue;
     }
     else{
         action(st);
         insert_register_and_config_params(st);
         transform_registration(st);
       }
     
   }
   delete [] children;
   return;
   
}


void transform_registrationA(SgStatement * s) 
{
  int i=0;

  
   SgStatement *st;
   SgStatement *line;
   

   for (line=s; line;line=line->lexNext()){
     st =  line;
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
         
     {
         action(st);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st);
	   i++;
	   continue;
     }
     else{
         action(st);
         insert_register_and_config_params(st);
       }
     
   }
   return;
}



 SgStatement *find_outermost_scope(SgStatement *s)
{
   SgProgHedrStmt *    prog,*curs;    
   SgProcHedrStmt *    subr;  
   SgProsHedrStmt *    pros;
   SgFuncHedrStmt *    func;  
   SgStatement *cp;


   cp = s;
   curs = NULL;
   do{
      if(subr = isSgProcHedrStmt (cp))
        curs = (SgProgHedrStmt *)subr;
      else if (pros = isSgProsHedrStmt(cp))
        curs = (SgProgHedrStmt *)pros;
      else if(func = isSgFuncHedrStmt(cp))
        curs = (SgProgHedrStmt *)func;
      else if(prog = isSgProgHedrStmt(cp))
        curs = prog;
      
      if (curs == NULL){
      
	cp = cp->controlParent();
      
      }
   }while(!curs);
   return (curs);
   
}

//find whether or not current stmt s is in one loop ,
// if so return line number of the header of innermost loop  
// otherwise return 0
int isInLoop(SgStatement *s)
{

   SgStatement *cp,*curs;
   SgForStmt      *fstmt;
   SgDoWhileStmt  *dwstmt;

   cp   = s;
   curs = NULL;
   do{
     if ((fstmt = isSgForStmt(cp))||(dwstmt = isSgDoWhileStmt(cp)))
       curs =  cp;
     else if(cp != NULL)
   	cp = cp->controlParent();
   }while((cp)&&(!curs));
   if (curs )
     return (curs->lineNumber());
   else 
     return 0;
         
}



void find_exec_next_stmt(SgStatement *s,int *lexnext,int *callnext)
{
  SgStatement *cp;
  SgCallStmt  *cs;
  int i;
  
  if (s->lexNext()->variant() != CONTROL_END){
    if (s->lineNumber() == s->lexNext()->lineNumber()){
       cp = s;
       while((cp->nextInChildList())&& 
          (cp->nextInChildList()->variant() == CONTROL_END))
       cp = cp->controlParent();
       if (cp->nextInChildList())
         *lexnext  = cp->nextInChildList()->lineNumber();
       else
         *lexnext  = -3;     
    }
    
    else
      *lexnext  = s->lexNext()->lineNumber();
  }
      
  else{
    cp = s;
    while((cp->nextInChildList())&& 
         (cp->nextInChildList()->variant() == CONTROL_END))
      cp = cp->controlParent();
    if (cp->nextInChildList())
      *lexnext  = cp->nextInChildList()->lineNumber();
    else
      *lexnext  = -2;
    
  }
  
  if (cs = isSgCallStmt(s)){
    for(i=0;i<nextf;i++)
      if(!strcmp(funcs[i].fname,cs->name()->identifier()))
        break;
    if (i == nextf) 
      *callnext = -1;
    else
      *callnext = funcs[i].feline;
  }
  else
    *callnext  = -2;
  
}

  
SgStatement *find_first_exec_stmt(SgStatement *s)
{
  int i;
  
   int nlchilds = s->numberOfChildrenList1();
   SgStatement *lasts = s-> childList1(nlchilds-1) ; 

   for(i=0;i<nlchilds;i++)
     if ((isSgExecutableStatement(s->childList1(i)))&&
        (s->childList1(i)->variant() != CONTROL_END)&&
        (s->childList1(i)->variant() != USE_STMT)&&
        (s->childList1(i)->variant() != INCLUDE_STAT)&&
	(s->childList1(i)->variant() != PROCESSORS_STMT)&&
        (s->childList1(i)->variant() != TEMPLATE_STMT)  )
	 return (s->childList1(i));
   
   return lasts;
   
}

int give_type_symbol(SgType *vtype)
{
  int variant_type = vtype->variant();
  int var_type;

  switch(variant_type){
           
     case T_CHAR: var_type = T_CHAR; break;
     case T_INT: var_type = T_INT; break;
     case T_FLOAT: var_type = T_FLOAT; break;
     case T_DOUBLE : var_type = T_DOUBLE; break;
     case T_BOOL: var_type = T_BOOL; break;
     case T_ARRAY: var_type = give_type_symbol(vtype->baseType());break;
     case T_KIND:
	int base_type = 
            vtype->thetype->entry.kind_decl.base_type->variant;
        int kind_val = (vtype->thetype->entry.kind_decl.kindexp)->entry.                   list_s.item->entry.literal.value->entry.ival; 
	
        switch(base_type){
	case T_INT:
	  switch(kind_val){
	  case 1:var_type = T_BYTE;break; /* Byte */
	  case 2:var_type = T_SHORT;break; /* Short */
	  case 4:var_type = T_INT;break; /* Integer */
	  case 8:var_type = T_LONG;break; /* Long */
	  }
	  	  
	  break;
	case T_FLOAT:  
	  switch(kind_val){
	  case 4:var_type = T_FLOAT;break; /* Float */
	  case 8:var_type = T_DOUBLE;break; /* Double */
	  }
	  
	  break;
	}
		
	break;
  }

  return (var_type);
 
}


int symb_ndims(SgSymbol *symb)

{
  int vtype =  symb->type()->variant();
  SgArrayType *atype;
  int ndims;
    
  switch(vtype){
  case T_ARRAY: if (atype = isSgArrayType(symb->type()))
                   ndims = atype->dimension();
  break;
  
  default:ndims = 0;break;
  }
  return ndims;
  
}

  
void insert_register_and_config_params(SgStatement *s)
{

    SgProgHedrStmt *    prog,*curs;    
    SgProcHedrStmt *    subr;  
    SgProsHedrStmt *    pros;
    SgFuncHedrStmt *    func;       
    SgSymbol *symb;
    SgVariableSymb *fsymb,*shape_symb;
    SgStatement    *    festmt;
    int var_type;
    int ndims,scope,var_id;
    char *name,indexName[256] ;
    SgValueExp *varType,*varName,*varNdims,*varShape,*varScope;
    SgCallStmt *scstmt;
    SgVarRefExp *e1;
    SgVariableSymb *ptsymb;
    SgFunctionCallExp *fce;
    int distFlag;
    SgStatement *decls;
    int nextTemp=0;
    char *ts ;
     
    curs = NULL;
    if(subr = isSgProcHedrStmt (s))
       curs = (SgProgHedrStmt *)subr;
    else if (pros = isSgProsHedrStmt(s))
      curs = (SgProgHedrStmt *)pros;
    else if(func = isSgFuncHedrStmt(s))
      curs = (SgProgHedrStmt *)func;
    else if(prog = isSgProgHedrStmt(s))
      curs = prog;
    
    if (curs == NULL) return;
    //printf("symbol:%s", (curs->name()).identifier());
    //    printf(" which is a subroutine.\n");
    
        /* find first executable stmt in current scope */
        festmt = find_first_exec_stmt(curs);
        
	
        if (festmt == NULL)
          return;
	
 
         /* add the first executable stmt of this func./subr. to funcs
           record array */
        funcs[nextf].feline = festmt->lineNumber();
	funcs[nextf].fname   = (curs->name()).identifier();
        funcs[nextf++].s = curs;
	
        SgVariableSymb *register_var_symb = new SgVariableSymb("register_var",*curs);
        hpf_server_symb = new SgVariableSymb("hpf_server",*curs);
        jmp_symb =  new SgVariableSymb("setjmp",*curs);        

        // create SgSymbol for function name     
	char *funcName = (curs->name()).identifier();
        char *funcNameSymb  = (char *)malloc(strlen(funcName)+5);
        strcpy(&funcNameSymb[1], funcName);
        funcNameSymb[0] = '\"';
        funcNameSymb[strlen(funcName)+1] = '\\';
        funcNameSymb[strlen(funcName)+2] = '0';
        funcNameSymb[strlen(funcName)+3] = '\"';
        funcNameSymb[strlen(funcName)+4] = 0;    
        SgVariableSymb *fSymb = new SgVariableSymb(funcNameSymb,*curs);
        SgValueExp *funcVarName = new SgValueExp(fSymb->identifier());

        // new addition
        // get the expression list of variables delcared
        for(decls=curs->lexNext();decls!=festmt;decls = decls->lexNext())
	{
	  
           if (!isSgVarDeclStmt(decls))
             continue;
           
	   SgVarDeclStmt *svar = (SgVarDeclStmt *)decls;
           int ns = svar->numberOfSymbols();
	   
           for(int i=0;i<ns;i++)
	   {
              if (svar->symbol(i)->variant() != VARIABLE_NAME) 
                continue;
                 
	      symb = svar->symbol(i);
              //printf("I am registring : %s variant = %d in fun=%s\n",
              //       symb->identifier(),symb->variant(),
              //       (curs->name()).identifier() );
            	          	   
              /* Dont register variables other than variable names */
              if (symb->variant() != VARIABLE_NAME)
                continue;
                          
              fsymb = register_var_symb;
	      scstmt = new SgCallStmt(*fsymb);
	      
              /* add var_type parameters to register_var subroutine */
	      var_type = give_type_symbol(symb->type());
              varType = new SgValueExp(var_type);
	      scstmt->addArg(*varType);

              /* add variable id = parameter var_id */	
              sprintf(indexName, "ind_%d", nextTemp++);
              ptsymb = new SgVariableSymb(indexName,*curs);
	      e1 = new SgVarRefExp(*ptsymb);
	      scstmt->addArg(*e1);
              // make declaration this new created temporary variable
	      // in this function/subroutine. Fotget if for now
              // ptsymb->declareTheSymbol(*curs);
	        

              /* add variable itself = var parameter */
              e1 = new SgVarRefExp(*symb);
	      scstmt->addArg(*e1);
                 
              /* add variable literal name = name parameter */
              ts  = (char *)malloc(strlen(symb->identifier())+5);
              strcpy(&ts[1], symb->identifier() );
              ts[0] = '\"';
              ts[strlen(symb->identifier())+1] = '\\';
              ts[strlen(symb->identifier())+2] = '0';
              ts[strlen(symb->identifier())+3] = '\"';
              ts[strlen(symb->identifier())+4] = 0;                   
   


              SgVariableSymb *dummy = new SgVariableSymb(ts,*curs);

	      varName = new SgValueExp(dummy->identifier());
              //SgExpression *varName = new SgExpression(STRING_VAL,ts);
	      scstmt->addArg(*varName);
	      //printf("ts = %s arg = %s\n",ts,varName->stringValue());
 
              /* add var_ndims parameter */
	      ndims = symb_ndims(symb);
              varNdims = new SgValueExp(ndims);
	      scstmt->addArg(*varNdims);
              // add current function name  
              scstmt->addArg(*funcVarName);

               // This NEW ADDED for attributes of vars
              // add null attribute now. later it can change 
              SgStatement *declStmtIn = symb->declaredInStmt();
              /* add variable literal name = name parameter */
	      int slen = strlen("");
              ts  = (char *)malloc(slen+5);
              strcpy(&ts[1], "");
              ts[0] = '\"';
              ts[slen+1] = '\\';
              ts[slen+2] = '0';
              ts[slen+3] = '\"';
              ts[slen+4] = 0;     
              SgVariableSymb *attdummy = new SgVariableSymb(ts,*curs);
              SgVarRefExp *attvarName = new SgVarRefExp(attdummy);
 	      scstmt->addArg(*attvarName); 

              /* Add distributed flag:is distributed or not */      
              if ((symb->attributes() & ALIGN_BIT)||
                    (symb->attributes() & DISTRIBUTE_BIT ))
		  distFlag = 1;
	      else
		  distFlag = 0;
              SgValueExp *dflag = new SgValueExp(distFlag);       
              scstmt->addArg(*dflag);
   
              /****************************************************/
              /* if its is array ,adds its ubounds and lbounds as argument */
              if ((symb->attributes()& DIMENSION_BIT)||
		  strstr(symb->identifier(),"LOCAL_")){
                 /*prepare LBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the LBOUND of var */
                 shape_symb = new SgVariableSymb("lbound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgValueExp(symb->identifier());
                 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);


                 /*prepare UBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the UBOUND of var */
                 shape_symb = new SgVariableSymb("ubound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgValueExp(symb->identifier());
                 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);
	      } 
              /****************************************************/

              /* finally add register call stmt */
              festmt->insertStmtBefore(*scstmt,*(festmt->controlParent()));

              //scstmt->unparsestdout();

	   }
	}
       	
        // put server function call before end of function unit
        SgStatement *lastStmt = curs->lastExecutable()->lexNext();
        if (curs->lastExecutable()->variant() != RETURN_STAT)
	  InsertFCallNode(lastStmt , HPF_SERVER,BEFORE);
	free(ts);
}

insertAllocateStmt(SgSymbol *symb,SgStatement *curs)
{
    SgCallStmt *callStmt;
    SgVariableSymb *parSymb;//,*callSymb;
    SgFunctionSymb *callSymb;
    SgStatement *sline;
    char varName[1024],tmpName[1024],*p;

    sline = find_first_exec_stmt(curs);
    //callSymb = new SgVariableSymb("ALLOCATE", *(sline->controlParent()));
    callSymb = new SgFunctionSymb(FUNCTION_NAME,"ALLOCATE",
		     *(new SgType(T_VOID)),*(sline->controlParent()));
    callStmt = new SgCallStmt(*callSymb);

    p = symb->identifier();
    strcpy(tmpName,p+strlen("LOCAL_"));
    parSymb = new SgVariableSymb(tmpName);

    SgArrayRefExp *arefexp = new  SgArrayRefExp(*symb);
    for(int i=1;i<= symb_ndims(symb);i++){
      SgVariableSymb *shape_symb = new SgVariableSymb("size");
      SgFunctionCallExp *fce   = new SgFunctionCallExp(*shape_symb );
      SgVarRefExp   *sv = new SgVarRefExp(parSymb);
      fce -> addArg(*sv);
      SgValueExp *sve = new SgValueExp(i);
      fce -> addArg(*sve); 
      arefexp->addSubscript(*fce);
    }
    callStmt->addArg(*arefexp);
    p = callStmt->unparseToBuf();
    p = strstr(p,"call ")+strlen("call ");
    sprintf(tmpName,"      %s",p);
    sline->addComment(tmpName);  

    //Insert ALLOCATE function call just before registration of this local
    //sline->insertStmtBefore(*callStmt,*(sline->controlParent()));
    //sline->insertStmtBefore(*ass,*(sline->controlParent()));
}



/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
/* Here we are rewriting to insert the registration of only the variables*/
/* we are interedted inone variable I know this is extra but later I will*/
/* reorganise  my coding to make it more structured                      */
/*************************************************************************/
void insert_register_defined_vars(SgStatement *procHedr)
				 
{
    
    SgProgHedrStmt *    prog,*curs;    
    SgProcHedrStmt *    subr;  
    SgProsHedrStmt *    pros;
    SgFuncHedrStmt *    func;       
    SgSymbol *symb;
    SgVariableSymb *fsymb,*shape_symb;
    SgStatement    *    festmt;
    int var_type;
    int ndims,scope,var_id;
    char *name,indexName[256] ;
    SgValueExp *varType,*varName,*varNdims,*varScope;
    SgCallStmt *scstmt;
    SgVarRefExp *e1,*varShape; 
    SgVariableSymb *ptsymb;
    SgFunctionCallExp *fce;
    int distFlag;
    SgStatement *decls;
    int nextTemp=0,found,k;
    char *ts ,useStmt[512];
    char hpfDynamic[1024];
    int dynCnt=0;
    SgVariableSymb *register_var_symb;

    curs = (SgProgHedrStmt *)(procHedr);

    /* find first executable stmt in current scope */
    festmt = find_first_exec_stmt(curs);
    //if (festmt == NULL)
    //   return;

    register_var_symb=new SgVariableSymb("register_var",*curs);

    //printf("symbol:%s", (curs->name()).identifier());
    //printf(" which is a subroutine.\n");
  
    //create SgSymbol for function name     
    char *funcName = (curs->name()).identifier();
    char *funcNameSymb  = (char *)malloc(strlen(funcName)+5);
    strcpy(&funcNameSymb[1], funcName);
    funcNameSymb[0] = '\"';
    funcNameSymb[strlen(funcName)+1] = '\\';
    funcNameSymb[strlen(funcName)+2] = '0';
    funcNameSymb[strlen(funcName)+3] = '\"';
    funcNameSymb[strlen(funcName)+4] = 0;    
    SgVariableSymb *fSymb = new SgVariableSymb(funcNameSymb,*curs);
    SgVarRefExp *funcVarName = new SgVarRefExp(fSymb);

    for(decls=curs->lexNext();decls!=festmt;decls = decls->lexNext())
    {
       if (!isSgVarDeclStmt(decls))
          continue;
           
       SgVarDeclStmt *svar = (SgVarDeclStmt *)decls;
       int ns = svar->numberOfSymbols();
	   
       for(int i=0;i<ns;i++)
       {
              symb = svar->symbol(i);
              /* Dont register variables other than variable names */
              if (symb->variant() != VARIABLE_NAME)
                continue;
            
              fsymb = register_var_symb;
	      scstmt = new SgCallStmt(*fsymb);

              /* add var_type parameters to register_var subroutine */
	      var_type = give_type_symbol(symb->type());
              varType = new SgValueExp(var_type);
	      scstmt->addArg(*varType);

              /* add variable id = parameter var_id */	
              sprintf(indexName, "ind_%d", nextTemp++);
              ptsymb = new SgVariableSymb(indexName,*curs);
	      e1 = new SgVarRefExp(*ptsymb);
	      scstmt->addArg(*e1);
           	        
  
              /* add variable itself = var parameter */
              e1 = new SgVarRefExp(*symb);
	      scstmt->addArg(*e1);
                 
              /* add variable literal name = name parameter */
              ts  = (char *)malloc(strlen(symb->identifier())+5);
              strcpy(&ts[1], symb->identifier() );
              ts[0] = '\"';
              ts[strlen(symb->identifier())+1] = '\\';
              ts[strlen(symb->identifier())+2] = '0';
              ts[strlen(symb->identifier())+3] = '\"';
              ts[strlen(symb->identifier())+4] = 0;                   
              SgVariableSymb *dummy = new SgVariableSymb(ts,*curs);
              SgVarRefExp *tvarName = new SgVarRefExp(dummy);
 	      scstmt->addArg(*tvarName);
	 
              /* add var_ndims parameter */
	      ndims = symb_ndims(symb);
              varNdims = new SgValueExp(ndims);
	      scstmt->addArg(*varNdims);
              // add current function name  
              scstmt->addArg(*funcVarName);

             
              /* Add distributed flag:is distributed or not */      
              if ((symb->attributes() & ALIGN_BIT)||
		  (symb->attributes() & DISTRIBUTE_BIT )){
		  distFlag = 1;
                  sprintf(hpfDynamic,
               "      !hpf$ DYNAMIC :: %s\n",symb->identifier());
                  curs->lexNext()->addComment(hpfDynamic);
	      }
	      else
		  distFlag = 0;
              SgValueExp *dflag = new SgValueExp(distFlag);       
              scstmt->addArg(*dflag);
   
              /****************************************************/
              /* if its is array ,adds its ubounds and lbounds as argument */
              if ((symb->attributes()& DIMENSION_BIT)||
		  strstr(symb->identifier(),"LOCAL_")){
                 /*prepare LBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the LBOUND of var */
                 shape_symb = new SgVariableSymb("lbound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgVarRefExp(symb);
                 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);


                 /*prepare UBOUND(A) call function for dim_info parameter*/
                 /* Here we are trying to calculate the UBOUND of var */
                 shape_symb = new SgVariableSymb("ubound",*curs);
	         fce   = new SgFunctionCallExp(*shape_symb );
	         varShape = new SgVarRefExp(symb);
		 fce -> addArg(*varShape);
                 /* Add dim_info parameter  */
                 scstmt->addArg(*fce);

              
	      }
              if (strstr(symb->identifier(),"LOCAL_"))
                 insertAllocateStmt(symb,curs);
              /****************************************************/
              /* finally add register call stmt */
              festmt->insertStmtBefore(*scstmt,*(festmt->controlParent()));
       }
    }
    free(ts);

}

SgStatement *whereDeclared(SgSymbol *x,SgStatement *start)
{ 
   SgStatement  *old_decl;
   SgExpression *exp;
   SgExprListExp *elist;

   for(old_decl = start;old_decl ;old_decl=old_decl->lexNext()){
       
       if (old_decl -> variant() == SAVE_DECL)
          continue; 
       
       // get the expression list of variables delcared
       exp = old_decl->expr(0);
       // look for the reference to our symbol x.
       while(elist = isSgExprListExp(exp)){
       	 if((elist->value())->IsSymbolInExpression(*x) )
               break;       
         exp = elist->next();
       }    
       if ( exp ) 
          break;
   }
   return old_decl;
}

SgVariableSymb *DeclareTemporaryLike(SgSymbol * x,int sd)
{
       SgStatement  *old_decl_stmt;
       SgVariableSymb *tmp;
       SgVarDeclStmt *decl_stmt;
       SgExpression *exp;
       char *name;
       SgExprListExp *elist;
       SgType *old_type;
       SgVarDeclStmt *decl;
       SgProgHedrStmt *ispg;
              
       // now find the declaration statement for x and get
       // the correct initializing expression.
       
       old_decl_stmt = x ->declaredInStmt();
       // get the expression list of variables declared
       exp = old_decl_stmt->expr(0);
       // look for the reference to our symbol x.
       while(elist = isSgExprListExp(exp)){
       	 if((elist->value())->IsSymbolInExpression(*x) )
               break;       
            exp = elist->next();
            }    
       if(exp == NULL) return NULL;

       // now create a new expression list and 
       // and a new declaration statement
       elist = new SgExprListExp(exp->copy());
       if (decl = isSgVarDeclStmt(old_decl_stmt))
         old_type = decl->type();
       decl_stmt = new SgVarDeclStmt(*elist,*old_type);

       // create a new variable of the same class
       // generate the name as "_Ti" where i is counter.
       name = (char *)malloc(256);
       if (sd == SLOCAL)     
	 sprintf(name, "LOCAL_%s", x->identifier());
       else
	 sprintf(name, "DISTR_%s", x->identifier());
       tmp = new SgVariableSymb(name, (x->type())->copy(), *(x->scope()));
    
       //Test if new declaration can be added
       SgArrayRefExp *arefexp = new  SgArrayRefExp(*(&tmp->copy())); 
       //prepare size(A) call fucntion for dim_info parameter
       //Here we are trying to calculate the shape of var 
       for(int i=1;i<=symb_ndims(x);i++){
	    //SgVariableSymb *shape_symb = new SgVariableSymb("size");
	    //SgFunctionCallExp *fce   = new SgFunctionCallExp(*shape_symb );
	    //SgValueExp *sve = new SgValueExp(x->identifier());
	    //SgValueExp *sve;
	    //SgVarRefExp   *sv = new SgVarRefExp(x);
            //printf("LOCAL : %s\n",x->identifier()); 
	    //fce -> addArg(*sve);
	    //fce -> addArg(*sv);
	    //sve = new SgValueExp(i);
	    //fce -> addArg(*sve); 
            arefexp->addSubscript(*(new SgExpression(COLON_OP))); //*fce);
	}
        
	SgVarDeclStmt *ndecl = new SgVarDeclStmt(*arefexp,
               *(x->type()->baseType()));
	old_decl_stmt->insertStmtAfter(*ndecl, 
				       *(old_decl_stmt->controlParent()));

        char *b = old_decl_stmt->unparseToBuf();
        char *bp = old_decl_stmt->controlParent()->unparseToBuf();
        //printf("defined at-%s- and parent=%s- \n",b,bp);

        return tmp;

        SgAttributeExp(ALLOCATABLE_OP);
    }

void insert_decls(SgStatement *moduleDecls)
{
   SgStatement *prev,*s1;
   SgSymbol *symb;
   SgVarDeclStmt *decl;
   SgProgHedrStmt *p;
   int found ,k;

   int nchilds = moduleDecls->numberOfChildrenList1();
   SgStatement *last = moduleDecls->childList1(nchilds-1);
   //printf("nchilds = %d\n",nchilds);
   for(prev = moduleDecls;(prev != last);prev = prev->lexNext()){  
     if  ((prev->variant() == ALIGN_DECL)||   
	  (prev->variant() == DISTRIBUTE_DECL)||
          (prev->variant() == COMBINED_DIRECTIVE_DECL)||
	  (prev->variant() == ALIGN_DECL)||
          (prev->variant() == TEMPLATE_STMT) )
       {
         decl = (SgVarDeclStmt *)prev;
       	 for(int i=0;i<decl->numberOfSymbols();i++){
            if (isTemplate(decl->symbol(i)))
              continue;
            DeclareTemporaryLike(decl->symbol(i),SLOCAL);
	 }
       }
   }

  
}

int isTemplate(SgSymbol *symb)
{
   SgStatement *prev = symb->declaredInStmt();

   if  ((prev->variant() == ALIGN_DECL)||   
	(prev->variant() == DISTRIBUTE_DECL)||
        (prev->variant() == COMBINED_DIRECTIVE_DECL)||
	(prev->variant() == ALIGN_DECL)||
        (prev->variant() == TEMPLATE_STMT) )
     return 1;
   else
     return 0;
}
   


int fileLength(SgFile *f,int num_of_routines)
{
  SgStatement *s = f->functions(num_of_routines-1);
  s = s->childList1(s->numberOfChildrenList1()-1);
  return (s->lineNumber()+1);

}

       
void readParseTree(char *projectFileName,int *nfiles,
		   FILES *in_files,FTAB  *in_functab)
{
   SgProject * project = new SgProject(projectFileName);
   
   files   = in_files;
   functab = in_functab;
   create_functionTable_file(project,nfiles);
}



create_functionTable_file(SgProject * project, int *nfiles)
{
  int numOfFiles = project->numberOfFiles();
  int num_of_routines;
  int filelen;
  SgStatement *child;
  int rem,k;
  
  files = (FILES *)calloc(numOfFiles,sizeof(FILES));
  *nfiles = numOfFiles;

  for(int i=0;i< numOfFiles;i++){
    SgFile *f;
    SgStatement *mp;
    int nf;

    nf = 0;

    f = &(project->file(i));
    if (f -> mainProgram())
      files[i].isIncludeMain = 1;
    else
      files[i].isIncludeMain = 0;   

    num_of_routines = f->numberOfFunctions();

    // calculate filelen and structures    
    filelen = fileLength(f, num_of_routines);
    files[i].lineInfo = (SINFO *)calloc(filelen,sizeof(SINFO));
    files[i].functab  = (FTAB *)calloc(num_of_routines,sizeof(FTAB));

    strcpy(files[i].fileName,project->fileName(i));
    files[i].fileSize = filelen;

    char *sptr = strrchr(files[i].fileName,'.');
    strncpy(files[i].depFileName,files[i].fileName,
	 sptr - (files[i].fileName));
    strcat(files[i].depFileName,".dep");

    for(int j=0;j < num_of_routines;j++){
      SgStatement *sub;
      SgSymbol    *subsymb;

      sub     = f->functions(j);
      subsymb = sub->symbol();
      files[i].functab[nf].fExecStmt = find_first_exec_stmt(sub)->lineNumber();
      if (sub -> variant() == PROG_HEDR)
         files[i].mainFuncIndex = j;

      if (sub->controlParent()->variant() !=  INTERFACE_STMT){
         files[i].functab[nf].lineNum = sub->lineNumber();
         files[i].functab[nf].ParseTreeStmt = NULL;
         strcpy(files[i].functab[nf].funcName,subsymb->identifier());
         
         strcpy(files[i].functab[nf++].fileName,project->fileName(i));
        
      }

      SgStatement *ls = sub->childList1(sub->numberOfChildrenList1()-1);
      ls = ls->lexNext();

      SgStatement *ts = sub;
      for (; ts != ls; ts=ts->lexNext()){
         child = ts;
         k = child->lineNumber();
        
         files[i].lineInfo[k].lineNum = child->lineNumber();
         files[i].lineInfo[k].stmtID  = child->variant();
         files[i].lineInfo[k].parent  = child->controlParent()->lineNumber();
         files[i].lineInfo[k].isExecutable = 0; 
         files[i].lineInfo[k].isInLoop = isInLoop(child);
         if ((child->variant() == RETURN_STAT)||
	     (child->variant() == STOP_STAT)||
	     (ts->lexNext() == ls))
	   files[i].lineInfo[k].isReturnPoint = 1;
	 else
           files[i].lineInfo[k].isReturnPoint = 0;
       } 
    }
    files[i].nfuncs = nf;
    transform_registration(f->firstStatement(),i);       
  }
}



void action(SgStatement * s,int find)
{
  SgDeclarationStatement *decl;
  SgProgHedrStmt *spg;
  SgCallStmt *cs;
  SgForStmt *fs;
  int flag = 0;
  SgAssignStmt *ass ;
  SgForStmt *sfo ; 
  SgCallStmt *call;
  
  SgStatement *cp = s->controlParent();
  if ((isSgExecutableStatement(s))&&
      (s->variant() != CONTROL_END)&&
      (s->variant() != MODULE_STMT)&&
      (s->variant() != USE_STMT)&&
      (s->variant() != INCLUDE_STAT)&&
      ((!cp)||(!(cp->lineNumber()==s->lineNumber()))) )
          files[find].lineInfo[s->lineNumber()].isExecutable = 1; 
     
}

void transform_registration(SgStatement * s,int find) 
{
  int i=0;

  int org_numberOfChildren1=s->numberOfChildrenList1();
  int org_numberOfChildren2= s->numberOfChildrenList2();
  int org_numberOfChildren = org_numberOfChildren1 + org_numberOfChildren2;
  SgStatement * * children = new SgStatement * [org_numberOfChildren];
  SgStatement *st;
  SgStatement *line;
   
  for (i=0; i<org_numberOfChildren1; i++)
	children[i]=s->childList1(i);

  for (; i<org_numberOfChildren; i++)
     children[i]=s->childList2(i-org_numberOfChildren1);

   for (i=0; i<org_numberOfChildren; i++){

     st =  children[i];
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
         
     {
         action(st,find);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st,find);
	   i++;
	   continue;
     }
     else{
         action(st,find);
         transform_registration(st,find);
     }
   }
   delete [] children;
   return;
   
}

int isDecl(SgStatement *chld)
{       if ((chld->variant() == VAR_DECL)||
            (chld->variant() == COMBINED_DIRECTIVE_DECL)||
	    (chld->variant() == DISTRIBUTE_DECL)||
            (chld->variant() == ALIGN_DECL)||
            (chld->variant() == PARAM_DECL)||
            (chld->variant() == PROCESSORS_STMT)||
            (chld->variant() == TEMPLATE_STMT)  )
         return 1;
     else
        return 0;

}

char *InstrumentModule(SgFile *file,int nextModule,
		       char *moduleNames[])
{
    int nextIndex;
    SgStatement *firstExecStmt;
    SgFunctionSymb *functionSymb,*moduleSymb ;
    SgStatement *fe,*prev,*s,*stemp;
    char funcName[1024], moduleName[1024],*tempbuf,line[1024];
    char *p,useStmt[1024];
    SgType *type = new SgType(T_VOID);

    nextIndex = nextModule-1;
    SgStatement *firstFunction  = file->functions(0);
    SgStatement *secondFunction = file->functions(1);

    //start instrumentation
    //First add local declarations for distributed variables
    //and then register the variables user defined in interpreter mode
    ///////////////////////////////////////////////////////////////////
    insert_decls(firstFunction);
    insert_register_defined_vars(firstFunction);
    firstExecStmt = find_first_exec_stmt(firstFunction);
   
    sprintf(funcName,"newfunction_%d",nextIndex);
    sprintf(moduleName,"newmodule_%d",nextIndex);   
    functionSymb = new SgFunctionSymb(PROCEDURE_NAME,funcName,
			       *type,*(firstFunction->controlParent()));
    SgProcHedrStmt *procHedr = new SgProcHedrStmt(*functionSymb);

      
    // Add parameters for new created function
    SgProgHedrStmt *pars = (SgProgHedrStmt *)firstFunction;
    SgFunctionSymb *parSymb = (SgFunctionSymb *)(&pars->name());
    for(int i=0;i<parSymb->numberOfParameters();i++){
      SgSymbol *symb;
      symb = &parSymb->parameter(i)->copy();
      functionSymb->insertParameter(i,*symb);
     }

    // Add the declarations coming from main function unit in which
    // user submit interpreter request to newly created function
    prev = procHedr;
    for(int i=0;i<secondFunction->numberOfChildrenList1()-1;i++){
      SgStatement *chld = secondFunction->childList1(i);
      if (isDecl(chld))
      {
	   s = &secondFunction->childList1(i)->copy();
           prev->insertStmtAfter(*s,*procHedr);
           prev = s;
      }
    }  

    // finally add user executable statements;
    for(fe = firstExecStmt;fe->nextInChildList() != NULL;
	fe=fe->nextInChildList())
    {
	 s = &fe->copy();
         prev->insertStmtAfter(*s,*procHedr);
         prev = s;
         s->addComment(fe->comments());
    }
    
    //And Eventually ,Add USE stmts for generated modules upto now
    for(int i=0;i<nextModule-1;i++){
      sprintf(useStmt,"      USE %s\n",moduleNames[i]);
      procHedr->lexNext()->addComment(useStmt);
    }


    char *output = (char *)malloc(10240); // say maximum program is 10K
    sprintf(line,"      MODULE  %s\n",moduleName);
    strcpy(output,line);

    // Add local definitions for new module
    for(s=firstFunction->lexNext(); (s!=firstExecStmt);s=s->lexNext()){
        tempbuf = s->unparseToBuf();
        strcat(output,tempbuf);
      }

    //Add CONTAINS stmt
    strcat(output,"      CONTAINS\n");

    // Add module method
    tempbuf = procHedr -> unparseToBuf();

  
   strcat(output,tempbuf); 
   //Add module END stmt
   strcat(output,"      END\n");
     
   //modify to free format
   char *loutput = (char *)malloc(10240);
   strcpy(loutput,"");
   for(p=output;p;){
     char *pe;
     pe = strstr(p,"\n     +");
     if (pe == NULL)
       break;
     strncat(loutput,p,pe-p);
     strncat(loutput,"&\n     &",8);
     p = pe+7;
   }
   strcat(loutput,p);
   free(output);
   printf("result : \n%s\n",loutput);

   return loutput;
}

