!=========================================================================
!this subroutine performs bit-reversal of distributed complex data,
!that is, it calculates  
!   g(j,k) = f(jj,kk), where
!   kk = (mixup(j,k)-1)/kxp+1 and jj = mixup(j,k)-kxp*(kk-1), and
!   mixup(j,k) = 1 + reversed bits of (j + kxp*(k + kstrt - 2) - 1)
!   f = complex input array
!   g = complex output array
!   mixup = array of bit reversed addresses
!   if isign = 0, bit-reverse array is prepared
!   if isign = 1, bit-reversal is performed
!written by viktor k. decyk, ucla
!=========================================================================
      subroutine bitrv1(f,g,mixup,isign,indx,kstrt,kxp,kblok,npp)
     
      complex, dimension(kxp,kblok) ::  f, g
      dimension  mixup(kxp,kblok),npp(kblok)
     
      !hpf$ distribute *(block)        :: npp 
      !hpf$ align (*,i) with *npp(i)   :: f,g,mixup

      nx = 2**indx
      if (kstrt.gt.nx) return
      ks = kstrt - 2
      if (isign.ne.0) go to 40
!---------------------------------
!prepare bit-reverse index table
!---------------------------------

         !HPF$ INDEPENDENT,NEW(k,j,koff,lb,ll,jb,it) 
      do 30 k = 1, kblok
         koff = kxp*(k + ks) - 1
         do 20 j = 1, kxp
            lb = j + koff
            ll = 0
            do 10 l = 1, indx
               jb = lb/2
               it = lb - 2*jb
               lb = jb
               ll = 2*ll + it
   10       continue
            mixup(j,k) = ll + 1
   20    continue
   30 continue
      return
!---------------------------------
!bit-reverse array elements
!---------------------------------

      !HPF$ INDEPENDENT,NEW(k,j,ll,kk,jj)
  40  do 60 k = 1, kblok
      do 50 j = 1, kxp
        ll = mixup(j,k)
        kk = (ll - 1)/kxp + 1
        jj = ll - kxp*(kk - 1)
        g(j,k) = f(jj,kk)
  50  continue
  60  continue

      return
      end
!=========================================================================
!this program calculates a random number y from a gaussian distribution
!with zero mean and unit variance, according to the method of
!mueller and box:
!   y(k) = (-2*ln(x(k)))**1/2*sin(2*pi*x(k+1))
!   y(k+1) = (-2*ln(x(k)))**1/2*cos(2*pi*x(k+1)),
!where x is a random number uniformly distributed on (0,1).
!written for the ibm by viktor k. decyk, ucla
!=========================================================================
      function ranorm(d)
      integer r1,r2,r4,r5
      double precision ranorm,h1l,h1u,h2l,r0,r3,asc,bsc,temp
      save iflg,r1,r2,r4,r5,h1l,h1u,h2l,r0
      data r1,r2,r4,r5 /885098780,1824280461,1396483093,55318673/
      data h1l,h1u,h2l /65531.0d0,32767.0d0,65525.0d0/
      data iflg,r0 /0,0.0d0/
      if (iflg.eq.0) go to 10
      ranorm = r0
      r0 = 0.0d0
      iflg = 0
      return
   10 isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r1 - (r1/isc)*isc
      r3 = h1l*dfloat(r1) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r2/isc
      isc = r2 - i1*isc
      r0 = h1l*dfloat(r2) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r2 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r1 = r3 - dfloat(isc)*bsc
      temp = dsqrt(-2.0d0*dlog((dfloat(r1) + dfloat(r2)*asc)*asc))
      isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r4 - (r4/isc)*isc
      r3 = h2l*dfloat(r4) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r5/isc
      isc = r5 - i1*isc
      r0 = h2l*dfloat(r5) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r5 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r4 = r3 - dfloat(isc)*bsc
      r0 = 6.28318530717959d0*((dfloat(r4) + dfloat(r5)*asc)*asc)
      ranorm = temp*dsin(r0)
      r0 = temp*dcos(r0)
      iflg = 1
      return
      end
    
