!======================================================================
! NPAC $Id: bsort.f90,v 1.6 1996/08/26 03:41:14 hwyau Exp $
!======================================================================
!
      program sort
! _____________________________________________________________________
!                                                                     |
! Bubble sort code.                                                   |
!                                                                     |
! H W Yau.  20th of August, 1996.                                     |
! Northeast Parallel Architectures Center.                            |
! Syracuse University.                                                |
!_____________________________________________________________________|
!                                                                     |
! Edit record:                                                        |
! 20/Aug/1996: First edit.                                            |
!              Rewrote initialisation to allow size to be increased.  |
!_____________________________________________________________________|

      integer, parameter:: NELEMENTS=1000
      INTEGER, PARAMETER :: DP = KIND(1.0D0)

      real, dimension(NELEMENTS):: sort_array
      real :: ti,tf,tcalc, tcom, a,b,start
      integer, dimension(NELEMENTS)::segment_array
!hpf$ distribute (block)  :: sort_array, segment_array

      interface
	subroutine bsort(Fdist,Flabel,tcalc,tcom)
	implicit none
	real,dimension(:), intent(inout) :: Fdist
        real tcalc,tcom
	integer,dimension(:), intent(in)::  Flabel
!hpf$   distribute *(block) :: Fdist,Flabel
        end subroutine bsort

      end interface

      interface
        subroutine timer(return_time,initial_time)
        real,intent(in) :: initial_time
        real,intent(out) :: return_time
        end subroutine
      end interface


!      interface
!      EXTRINSIC (F77_LOCAL) subroutine sciviz(name,time,coords,fx,np)
!         character (LEN=20) :: name  
!         real(DP) :: time      
!         real(DP), dimension(:) :: coords,fx 
!         integer np
!         end subroutine
!      end interface 

      integer i,li,ui

      nseg=10
      tcom=0.0
      tcalc=0.0
      call timer(start,0.0)
      call RANDOM_SEED()
      call RANDOM_NUMBER(sort_array)

      DO i=1,nseg
         li = (NELEMENTS/nseg)*(i-1) + 1
         ui = (NELEMENTS/nseg)*i
         segment_array(li:ui) = i
      END DO
      call timer(tcalc,start)

      call bsort(sort_array, segment_array,a,b)
      tcalc=tcalc+a
      tcom=tcom+b

      call timer(tf,start)

      WRITE(6,6001) NUMBER_OF_PROCESSORS(), NELEMENTS,tcom,tcalc,(tf-tcom-tcalc),tf

      STOP

 6001 FORMAT('Number of Processors = ' , I4/ 'Problem size =',I12/ &
      'Communications = ' , F9.3/ &
      'Compute = ' , F9.3/        &
      'Others = ', F9.3/          &
      'Total time =',F9.3)

      END PROGRAM sort

!======================================================================

      SUBROUTINE bsort(Fdist,Flabel,tcalc,tcom)

      INTEGER, PARAMETER :: DP = KIND(1.0D0)
       real(KIND=8) bb(2)


      real, dimension(:), intent(inout) :: Fdist
      real  tcalc,tcom
      integer, dimension(:), intent(in) :: Flabel
!hpf$ distribute *(block) :: Fdist,Flabel

      real, dimension(size(Fdist)) :: Fsave,Frdist,Fldist
      real(KIND=8) coords(1000),fx(1000)
      real x,ti
      integer, dimension(size(Fdist)) :: even,odd,Fsegment,mask
!hpf$ align with Fdist :: Fsave,Frdist,Fldist
!hpf$ align with Fdist :: even,odd,Fsegment,mask
! HPF$ ALIGN with Fdist :: coords,fx
!HPF$ DISTRIBUTE (BLOCK) :: coords,fx
      integer n,np,n1,t1,t8
      real(KIND=8) timei
      
!______________________________________________________________________
!
! Executable Code.
!______________________________________________________________________      
!
      bb(1)=1.0d0
      bb(2)=size(Fdist)*1.0d0
      forall(i=1:size(Fdist)) coords(i)=i*1.0d0

      tcom=0.0
      tcalc=0.0
      call timer(ti,0.0)
      n=size(Fdist)
      even=0
      odd=0
      even(2:n:2)=1
      odd(1:n:2) =1
      Fsave=0.

      call timer(tcalc,ti)

      !call timer(ti,0.0)
      where(Flabel.eq.cshift(Flabel,dim=1,shift=+1))
           Fsegment=1
      else where
           Fsegment=0
      end where
      call timer(tcom,ti)

      iter=0
      timei=iter

      do while(any(Fdist.ne.Fsave))

      iter=iter+1
      timei=iter*1.0d0
      fx=Fdist*1.0d0
      n1=size(fx)
      t1 = 1
      t8 = 8

      np=NUMBER_OF_PROCESSORS()
!      call sciviz("Sorted Array P",timei,coords,fx,np)

!      vsrc = java_bbser("Sorted Array BB",timei,bb,size(fx),fx)       
!      vsrc = java_ser("Sorted Array S",timei,coords,fx,8)       
!      vsrc = java_pser("Sorted Array P",timei,coords,fx,8,1)       

         call timer(ti,0.0)
         Fsave=Fdist
         call timer(x,ti)
         tcalc=tcalc+x

         call timer(ti,0.0)
         Frdist=eoshift(Fdist,dim=1,shift=1)
         Fldist=eoshift(Fdist,dim=1,shift=-1)
         call timer(x,ti)
         tcom=tcom+x

         call timer(ti,0.0)
         where((even.eq.1).and.(Fsegment.eq.1).and.(Fdist.gt.Frdist))
            mask=1
         else where
            mask=0
         end where
         call timer(x,ti)
         tcalc=tcalc+x

         call timer(ti,0.0)
         mask=mask+eoshift(mask,dim=1,shift=-1)
         call timer(x,ti)
         tcom=tcom+x

         call timer(ti,0.0)
         where(mask.eq.1)
                Fdist=Frdist*even + Fldist*odd
         end where
         call timer(x,ti)
         tcalc=tcalc+x

         call timer(ti,0.0)
         Frdist=eoshift(Fdist,dim=1,shift=1)
         Fldist=eoshift(Fdist,dim=1,shift=-1)
         call timer(x,ti)
         tcom=tcom+x


         call timer(ti,0.0)
         where((odd.eq.1).and.(Fsegment.eq.1) .and. (Fdist.gt.Frdist))
            mask=1
         else where
            mask=0
         end where
         call timer(x,ti)
         tcalc=tcalc+x

         call timer(ti,0.0)
         mask=mask+eoshift(mask,dim=1,shift=-1)
         call timer(x,ti)
         tcom=tcom+x

         call timer(ti,0.0)
         where(mask.eq.1)
                Fdist=Frdist*odd + Fldist*even
         end where
         call timer(x,ti)
         tcalc=tcalc+x

      end do

      iter=iter+1
      timei=iter
      fx=Fdist

      RETURN
      END SUBROUTINE bsort

!======================================================================
      
      subroutine timer(return_time, initial_time)
      real, intent(IN)  :: initial_time
      real, intent(OUT)  :: return_time
      integer finish,rate

      CALL system_clock(COUNT=finish,COUNT_RATE=rate)
      return_time = FLOAT(finish) / FLOAT (rate) -initial_time
      RETURN
      END SUBROUTINE timer
!======================================================================

!      subroutine sciviz(name,time,coords,fx,np)
!      INTEGER, PARAMETER :: DP = KIND(1.0D0)
!      character (LEN=20) :: name  
!      real(DP) :: time      
!      real(DP), dimension(:) :: coords,fx 
!      integer :: np

!      integer n,vsrc
!      n=size(fx)

!!      vsrc = java_pser(name,time,coords,fx,n,np)       

!      return
!      end





