#include "compilerApi.h"
#include "debugger.h"
int fileLength(SgFile *f,int num_of_routines);
create_functionTable_file(SgProject * project,int *nfuncs, int *nfiles);
int isInLoop(SgStatement *s);
SgStatement *find_outermost_scope(SgStatement *s);
void action(SgStatement * s,int find);
void transform_registration(SgStatement * s,int find) ;
write_summary(char *outFile,int nfuncs,int nfiles);
SgStatement *find_first_exec_stmt(SgStatement *s);

extern "C" void readParseTree(char *projectFileName ,int *nfuncs, int *nfiles);
typedef class SgStatement* obj;


/*main(int argc,char *argv[]){
  int n,m;
  readParseTree("test.prj",&n,&m);
  write_summary("EROL",n,m);
}*/


void readParseTree(char *projectFileName,int *nfuncs,int *nfiles)
{

   SgProject * project = new SgProject(projectFileName);

   create_functionTable_file(project,nfuncs,nfiles);
   /*write_summary("EROL", nfuncs, nfiles); */
}


int fileLength(SgFile *f,int num_of_routines)
{
  SgStatement *s = f->functions(num_of_routines-1);
  s = s->childList1(s->numberOfChildrenList1()-1);
  return (s->lineNumber()+1);

}


create_functionTable_file(SgProject * project,int *nfuncs, int *nfiles)
{
  int numOfFiles = project->numberOfFiles();
  int num_of_routines;
  int filelen,nroutines=0,nf=0;
  SgStatement *child;
  int rem,k,ninterfaces=0;

  
  for(int i=0 ; i< numOfFiles;i++){
    SgFile *f = &(project->file(i));
    nroutines += (f->numberOfFunctions());
  }
  
  functab = (FTAB *)calloc(nroutines,sizeof(FTAB));
  files = (FILES *)calloc(numOfFiles,sizeof(FILES));
  *nfiles = numOfFiles;

  for(int i=0;i< numOfFiles;i++){
    SgFile *f;
    SgStatement *mp;

    f = &(project->file(i));
    if (f -> mainProgram())
      files[i].isIncludeMain = 1;
    else
      files[i].isIncludeMain = 0;   

    num_of_routines = f->numberOfFunctions();

    // calculate filelen and structures    
    filelen = fileLength(f, num_of_routines);
    files[i].lineInfo = (SINFO *)calloc(filelen,sizeof(SINFO));
    strcpy(files[i].fileName,project->fileName(i));
    files[i].fileSize = filelen;

    for(int j=0;j < num_of_routines;j++){
      SgStatement *sub;
      SgSymbol    *subsymb;

      sub     = f->functions(j);
      subsymb = sub->symbol();
      functab[nf].fExecStmt = find_first_exec_stmt(sub)->lineNumber();
      if (sub -> variant() == PROG_HEDR)
         files[i].mainFuncIndex = j;

      if (sub->controlParent()->variant() !=  INTERFACE_STMT){
         functab[nf].lineNum = sub->lineNumber();
         functab[nf].ParseTreeStmt = sub;
         strcpy(functab[nf].funcName,subsymb->identifier());
         strcpy(functab[nf++].fileName,project->fileName(i));
         
        
      }
      else 
        ninterfaces++;

      SgStatement *ls = sub->childList1(sub->numberOfChildrenList1()-1);
      ls = ls->lexNext();

      SgStatement *ts = sub;
      for (; ts != ls; ts=ts->lexNext()){
         child = ts;
         k = child->lineNumber();
        
         files[i].lineInfo[k].lineNum = child->lineNumber();
         files[i].lineInfo[k].stmtID  = child->variant();
         files[i].lineInfo[k].parent  = child->controlParent()->lineNumber();
         files[i].lineInfo[k].isExecutable = 0; 
         files[i].lineInfo[k].isInLoop = isInLoop(child);
       }
    }
    transform_registration(f->firstStatement(),i);       
  }
  *nfuncs = nroutines-ninterfaces;
}

int isInLoop(SgStatement *s)
{

   SgStatement *cp,*curs;
   SgForStmt      *fstmt;
   SgDoWhileStmt  *dwstmt;

   cp   = s;
   curs = NULL;
   do{
     if ((fstmt = isSgForStmt(cp))||(dwstmt = isSgDoWhileStmt(cp)))
       curs =  cp;
     else if(cp != NULL)
   	cp = cp->controlParent();
   }while((cp)&&(!curs));
   if (curs )
     return (curs->lineNumber());
   else 
     return 0;
         
}


void action(SgStatement * s,int find)
{
  SgDeclarationStatement *decl;
  SgProgHedrStmt *spg;
  SgCallStmt *cs;
  SgForStmt *fs;
  int flag = 0;
  SgAssignStmt *ass ;
  SgForStmt *sfo ; 
  SgCallStmt *call;
  
  SgStatement *cp = s->controlParent();
  if ((isSgExecutableStatement(s))&&
      (s->variant() != CONTROL_END)&&
      (s->variant() != MODULE_STMT)&&
      (s->variant() != USE_STMT)&&
      (s->variant() != INCLUDE_STAT)&&
      ((!cp)||(!(cp->lineNumber()==s->lineNumber()))) )
          files[find].lineInfo[s->lineNumber()].isExecutable = 1; 
     
}

void transform_registration(SgStatement * s,int find) 
{
  int i=0;

  int org_numberOfChildren1=s->numberOfChildrenList1();
  int org_numberOfChildren2= s->numberOfChildrenList2();
  int org_numberOfChildren = org_numberOfChildren1 + org_numberOfChildren2;
  SgStatement * * children = new SgStatement * [org_numberOfChildren];
  SgStatement *st;
  SgStatement *line;
   
  for (i=0; i<org_numberOfChildren1; i++)
	children[i]=s->childList1(i);

  for (; i<org_numberOfChildren; i++)
     children[i]=s->childList2(i-org_numberOfChildren1);

   for (i=0; i<org_numberOfChildren; i++){

     st =  children[i];
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
         
     {
         action(st,find);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st,find);
	   i++;
	   continue;
     }
     else{
         action(st,find);
         transform_registration(st,find);
     }
   }
   delete [] children;
   return;
   
}

write_summary(char *outFile,int nfuncs,int nfiles)
{


    FILE *fp = fopen(outFile,"w");
 
    fprintf(fp,"// number of functions \n");
    fprintf(fp,"%d\n",nfuncs);
    fprintf(fp,"//LineNumber  FunctionName FileName \n");
    for(int i=0;i<nfuncs;i++)
      fprintf(fp,"%d   %s  %s\n", functab[i].lineNum,functab[i].funcName,
         functab[i].fileName);
 
    fprintf(fp,"//Number of Files \n");
    fprintf(fp,"%d\n",nfiles);
    fprintf(fp,"// LineNumber  stmtID parent isExecutable isInLoop \n");
    for(int i=0;i<nfiles;i++){
      fprintf(fp,"//File name \n");
      fprintf(fp,"%s\n",files[i].fileName);
      fprintf(fp,"//Length of File \n");
      fprintf(fp,"%d\n",files[i].fileSize);
      for(int k=0;k<files[i].fileSize;k++)
        fprintf(fp,"%5d    %5d   %5d   %5d   %5d \n",
	  files[i].lineInfo[k].lineNum,
          files[i].lineInfo[k].stmtID  ,
          files[i].lineInfo[k].parent  ,
          files[i].lineInfo[k].isExecutable,
          files[i].lineInfo[k].isInLoop );
    }
 
    fclose(fp);

}

SgStatement *find_first_exec_stmt(SgStatement *s)
{
  int i;
  
   int nlchilds = s->numberOfChildrenList1();
   
   for(i=0;i<nlchilds;i++)
     if ((isSgExecutableStatement(s->childList1(i)))&&
        (s->childList1(i)->variant() != CONTROL_END)&&
        (s->childList1(i)->variant() != USE_STMT)&&
        (s->childList1(i)->variant() != INCLUDE_STAT))
	 return (s->childList1(i));
   
   return NULL;
   
}
