      Program Laplace
!************************************************************
!*
!* This HPF program implements a finite difference method for
!* solving the Laplace Heat Equation via iteration.  Iteration
!* is continued until a steady state is reached to within some
!* epsilon.
!*
!************************************************************

      INTEGER maxX$,maxY$
      INTEGER count
      PARAMETER (maxX$=6)
      PARAMETER (maxY$=6)

      INTEGER S_ID,s_id1

      REAL surface(maxX$,maxY$) ,temp(maxX$,maxY$)

!HPF$ DISTRIBUTE surface(BLOCK,BLOCK)
      DOUBLE PRECISION epsilon,maxErr

      integer ndims,useless,uint
      integer dims(2),xcoor(5)

      ndims = 2
      port = 3336
      dims(1) = maxX$ 
      dims(2) = maxY$
      print *, maxX$, "**", maxY$ 
      temp = 101

      epsilon = 1.0E-6
      maxErr = 1.0E10
            
      CALL Init(surface,maxX$,2,maxX$-1,maxY$,2,maxY$-1,0.0)

      
      count = 0
      temp = surface
      do while (maxErr > epsilon)
         print *,"ITERATION: ",count
         
         surface(2:maxX$-1,2:maxY$-1) = (temp(1:maxX$-2,2:maxY$-1) + & 
                                        temp(3:maxX$,2:maxY$-1) + &
                                        temp(2:maxX$-1,1:maxY$-2) + &
                                        temp(2:maxX$-1,3:maxY$))/4.0 
       	 maxErr = MAXVAL(ABS(surface-temp))
         if (MOD(count,100) .EQ. 0) then
          print *,count,maxErr
         end if
         count = count + 1

      end do
      print *,'end of loop'

      End program Laplace

!************************************************************
!*
!* Subroutine Init initializes a 2D surface with a random 
!* distribution of temperatures in the interior and with 
!* temperature, temp, at the edges.
!*
!************************************************************
      Subroutine Init(surface,sizeX,x0,xn,sizeY,y0,yn,temp)

      INTEGER sizeX,sizeY
      REAL surface(sizeX,sizeY)
      INTEGER x0,xn,y0,yn
      !DOUBLE PRECISION temp
      !DOUBLE PRECISION rand
      real temp
      real rnum

      INTEGER i,j,seed

      seed = time()
      Call srand(seed)
	
      do i = x0,xn
	do j = y0,yn
          call random_number(rnum)
      	  surface(i,j) = 10000*rnum 
	end do
      end do

      print *, "Init - rand OK"
      surface(1,1:sizeY) = temp
      surface(sizeX,1:sizeY) = temp
      surface(1:sizeX,1) = temp
      surface(1:sizeX,sizeY) = temp

      print *, "init done"
      return
      end



