c
c ----------------------------------------------------------------------
c                                                                       
      subroutine dissp
c                                                                       
c compute added dissipitave model of Jameson
c Ref:  Jameson, Schmidt, Turkel
c       Numerical solution of the Euler equations by FV method ...
c       AIAA-81-1259
c dwx and dwy stored at boundaries of cells
c                                                                       
      include 'dimpar.inc'
      include 'common.inc'
	
c
    
      dimension dp1d(icell),ep1d(icell),d1q1d(4,icell)
     $  ,d3q1d(4,icell)
      dimension xlamdax(icell,jcell),xlamday(icell,jcell)
	real start,end
	real phi1,phi2,sigma
!HPF$ ALIGN (:,*) with BOARD (*,:,*) :: xlamdax,xlamday
!HPF$ ALIGN (*,:) with BOARD (*,:,*) :: d1q1d,d3q1d
!HPF$ ALIGN (:) with BOARD (*,:,*) :: dp1d,ep1d
c
      if (visc .eq. 0.) return
c
      
      alpha2       = visc*(1./2.)
      alpha4       = visc*(1./64.) 
c
      sigma        = 2./3.
c     sigma        = 0.4
c

	


c compute cell-centered static pressure
c
!HPF$ INDEPENDENT,NEW(j,i)
      do 10 j=1,my
!HPF$ INDEPENDENT,NEW(rav,rho,u,v)
      do 10 i=2,mx
      rav          = .5*( r(i) +r(i-1) )
      rho          = qnew(1,i,j)
      u            = qnew(2,i,j)/qnew(1,i,j)
      v            = qnew(3,i,j)/qnew(1,i,j)
   10 pres(i,j)    = (gam  -1.)*(qnew(4,i,j)  -.5*rho*(u*u  +v*v 
     &               -(rav*omega)**2))
c
          k   = 0
!HPF$ INDEPENDENT,NEW(rav,k)
      do 12 i=1,mx+1,mx
         rav       = r(i-k)
         k         = k+1
!HPF$ INDEPENDENT,NEW(rav,rho,u,v)
      do 12 j=1,my
      rho          = qnew(1,i,j)
      u            = qnew(2,i,j)/qnew(1,i,j)
      v            = qnew(3,i,j)/qnew(1,i,j)
 12   pres(i,j)    = (gam  -1.)*(qnew(4,i,j)  -.5*rho*(u*u  +v*v 
     &               -(rav*omega)**2))
c
c **********************
c CALCULATING XLAMDA AT CELL CENTER LOCATION 
c **********************
c
!HPF$ INDEPENDENT,NEW(i)
      do j = 2,my
!HPF$ INDEPENDENT,NEW(dx,dy,dl,rav,hav,rho,u,v,et,p,c,qx,qy)
      do i = 2,mx
c
c XLAMDAX
c
      dx           = .5*(x(i,j) -x(i,j-1) +x(i-1,j) -x(i-1,j-1))
      dy           = .5*(r(i)*(y(i,j) -y(i,j-1)) 
     &                  +r(i-1)*(y(i-1,j) -y(i-1,j-1))) 
      dl           = sqrt(dx*dx  +dy*dy)
      rav          = .5*( r(i) + r(i-1) )
      hav          = .5*( h(i) + h(i-1) )
      rho          = qnew(1,i,j)
      u            = qnew(2,i,j)/qnew(1,i,j)
      v            = qnew(3,i,j)/qnew(1,i,j)
      et           = qnew(4,i,j)
      p            = (gam  -1.)*(et  -.5*rho*(u*u  +v*v 
     &               -(rav*omega)**2))
      c            = sqrt(gam*p/rho)
      qx           = (u*dy  -v*dx)*hav
      xlamdax(i,j) = abs(qx)  +c*dl*hav
c
c XLAMDAY
c
      dx           = .5*(x(i,j) +x(i,j-1) -x(i-1,j) -x(i-1,j-1))
      dy           = .5*(r(i)*(y(i,j) +y(i,j-1)) 
     &                  +r(i-1)*(-y(i-1,j) -y(i-1,j-1))) 
      dl           = sqrt(dx*dx  +dy*dy)
      rav          = .5*( r(i) + r(i-1) )
      hav          = .5*( h(i) + h(i-1) )
      rho          = qnew(1,i,j)
      u            = qnew(2,i,j)/qnew(1,i,j)
      v            = qnew(3,i,j)/qnew(1,i,j)
      et           = qnew(4,i,j)
      p            = (gam  -1.)*(et  -.5*rho*(u*u  +v*v 
     &               -(rav*omega)**2))
      c            = sqrt(gam*p/rho)
      qy           = (u*dy  -v*dx)*hav
      xlamday(i,j) = abs(qy)  +c*dl*hav
      
      enddo
      enddo
c
!HPF$ INDEPENDENT

      do i = 2,ile
      xlamdax(i,1)    = xlamdax(i,my)
      xlamdax(i,my+1) = xlamdax(i,2)
      xlamday(i,1)    = xlamday(i,my)
      xlamday(i,my+1) = xlamday(i,2)
      enddo
c
!HPF$ INDEPENDENT
      do i = ite+1,mx
      xlamdax(i,1)    = xlamdax(i,my)
      xlamdax(i,my+1) = xlamdax(i,2)
      xlamday(i,1)    = xlamday(i,my)
      xlamday(i,my+1) = xlamday(i,2)
      enddo
      
        

c
c ******************
c calculation of dwx
c ******************
c
c calculate cell-centered pressure gradient in streamwise direction
c

	
!!HPF$ INDEPENDENT
      do 100 j=2,my

c
c calculate cell-centered pressure gradient in streamwise direction
c

!!HPF$ INDEPENDENT,NEW(xnum,xden)
!      do 110 i=2,mx
!      xnum         = pres(i+1,j)  -2.*pres(i,j)  +pres(i-1,j)
!      xden         = pres(i+1,j)  +2.*pres(i,j)  +pres(i-1,j)
!  110 dp1d(i)        = abs(xnum)/xden

	dp1d(2:mx)=pres(3:mx+1,j)-2.*pres(2:mx,j)+pres(1:mx-1,j)
	dp1d(2:mx)=abs(dp1d(2:mx))/(pres(3:mx+1,j)+2.*pres(2:mx,j)
     $ +pres(1:mx-1,j))
      i            = 1
      dp1d(i)        = 2.*dp1d(i+1)  -dp1d(i+2)
      i            = mx  +1
      dp1d(i)        = 2.*dp1d(i-1)  -dp1d(i-2)

	
c
c calculate cell-boundary pressure-gradient switching function
c

!!HPF$ INDEPENDENT
!      do 120 i=1,mx
!  120 ep1d(i)        = amax1(dp1d(i),dp1d(i+1))

	ep1d(1:mx)=amax1(dp1d(1:mx),dp1d(2:mx+1))
c
c calculate cell-boundary dQ/dx 
c
	
!!HPF$ INDEPENDENT
!      do 130 i=2,mx-1
!!HPF$ INDEPENDENT
!      do 130 n=1,4
!  130 d1q1d(n,i)     = qnew(n,i+1,j)  -qnew(n,i,j)

	d1q1d(1:4,2:mx-1)=qnew(1:4,3:mx,j)-qnew(1:4,2:mx-1,j)
	
!!HPF$ INDEPENDENT
!      do 131 n=1,4
!      d1q1d(n,1)     = 0.
!  131 d1q1d(n,mx)    = 0.


	d1q1d(1:4,1)=0.
	d1q1d(1:4,mx)=0.
c
c calculate cell-boundary d3Q/dx3
c
	
!HPF$ INDEPENDENT
      do 140 i=2,mx-1
!HPF$ INDEPENDENT
      do 140 n=1,4
  140 d3q1d(n,i)     = d1q1d(n,i+1) -2.*d1q1d(n,i) +d1q1d(n,i-1)
!	d3q1d(1:4,2:mx-1)=d1q1d(1:4,3:mx)-2.*d1q1d(1:4,2:mx-1)
!     $ + d1q1d(1:4,1:mx-2)
	
!!HPF$ INDEPENDENT
!      do 141 n=1,4
!      d3q1d(n,1)     = 0.
!  141 d3q1d(n,mx)    = 0.

	d3q1d(1:4,1)=0.
	d3q1d(1:4,mx)=0.
c
c calculate Dx{Q} term
c
	call timer(start,0.)
!HPF$ INDEPENDENT

      do 150 i=2,mx-1
      eps2x        = alpha2*ep1d(i)
      eps4x        = amax1(0.,alpha4  -eps2x)
      phi1         = 1. + (xlamday(i,j)/xlamdax(i,j))**sigma
      xlamda1      = xlamdax(i,j)*phi1
      phi2         = 1. + (xlamday(i+1,j)/xlamdax(i+1,j))**sigma
      xlamda2      = xlamdax(i+1,j)*phi2
      xlamda       =.5*( xlamda1 +xlamda2 )
      do 150 n=1,4
  150 dwx(n,i,j)   = xlamda*(eps2x*d1q1d(n,i)-eps4x*d3q1d(n,i))
	
	call timer(end,start)
      
	
!!HPF$ INDEPENDENT
! !     do 151 n=1,4
!      dwx(n,1,j)   = 0.
!  151 dwx(n,mx,j)  = 0.

	dwx(1:4,1,j)=0.
	dwx(1:4,mx,j)=0.
	
	
c
  100 continue


c
c ***********************************
c calculation of dwy in bladed region
c ***********************************
c

!HPF$ INDEPENDENT
      do 300 i=ile+1,ite
c
c calculate cell-centered pressure gradient in cross-stream direction
c
!HPF$ INDEPENDENT,NEW(xnunm,xden)
      do 310 j=3,my-1
      xnum         = pres(i,j+1)  -2.*pres(i,j)  +pres(i,j-1)
      xden         = pres(i,j+1)  +2.*pres(i,j)  +pres(i,j-1)
  310 dp(i,j)        = abs(xnum)/xden

       


      j            = 2
      dp(i,j)        = 2.*dp(i,j+1)  -dp(i,j+2)
      j            = my
      dp(i,j)        = 2.*dp(i,j-1)  -dp(i,j-2)
c
c calculate cell-boundary pressure-gradient switching function
c
!HPF$ INDEPENDENT
      do 320 j=2,my-1
  320 ep(i,j)        = amax1(dp(i,j),dp(i,j+1))
      j            = 1
      ep(i,j)        = dp(i,j+1)
      j            = my
      ep(i,j)        = dp(i,j)
c
c calculate cell-boundary dQ/dy 
c
!HPF$ INDEPENDENT
      do 330 j=2,my-1
!HPF$ INDEPENDENT
      do 330 n=1,4
  330 d1q(n,i,j)     = qnew(n,i,j+1)  -qnew(n,i,j)
!HPF$ INDEPENDENT
      do 332 n=1,4
      d1q(n,i,1)     = 0.
  332 d1q(n,i,my)    = 0.
c
c calculate cell-boundary d3Q/dy3
c
!HPF$ INDEPENDENT
      do 340 j=2,my-1
!HPF$ INDEPENDENT
      do 340 n=1,4
  340 d3q(n,i,j)     = d1q(n,i,j+1)-2.*d1q(n,i,j)+d1q(n,i,j-1)
!HPF$ INDEPENDENT
      do 342 n=1,4
      d3q(n,i,1)     = 0.
  342 d3q(n,i,my)    = 0.
c
c calculate Dy{Q} term
c
!HPF$ INDEPENDENT
      do 350 j=2,my-1
      eps2y        = alpha2*ep(i,j)
      eps4y        = amax1(0.,alpha4  -eps2y)
      phi1         = 1. + (xlamdax(i,j)/xlamday(i,j))**sigma
      xlamda1      = xlamday(i,j)*phi1
      phi2         = 1. + (xlamdax(i,j+1)/xlamday(i,j+1))**sigma
      xlamda2      = xlamday(i,j+1)*phi2
      xlamda       =.5*( xlamda1 +xlamda2 )
!HPF$ INDEPENDENT
      do 350 n=1,4
  350 dwy(n,i,j)   = xlamda*(eps2y*d1q(n,i,j)-eps4y*d3q(n,i,j))
!HPF$ INDEPENDENT
      do 352 n=1,4
      dwy(n,i,1)   = 0.
  352 dwy(n,i,my)  = 0.
c
  300 continue

      
	





c ****************************************
c calculation of dwy outside bladed region
c ****************************************
c

      do 500 nreg=1,2
c
      if (nreg .eq. 1) then
      i1           = 2
      i2           = ile
      else
      i1           = ite  +1
      i2           = mx
      endif
c
!HPF$ INDEPENDENT
      do 500 i=i1,i2
c
c calculate cell-centered pressure gradient in cross-stream direction
c
!HPF$ INDEPENDENT,NEW(xnum,xden)
      do 510 j=2,my
      xnum         = pres(i,j+1)  -2.*pres(i,j)  +pres(i,j-1)
      xden         = pres(i,j+1)  +2.*pres(i,j)  +pres(i,j-1)
  510 dp(i,j)        = abs(xnum)/xden
      dp(i,1)        = dp(i,my)
      dp(i,my+1)     = dp(i,2)
c
c calculate cell-boundary pressure-gradient switching function
c
!HPF$ INDEPENDENT
      do 520 j=1,my
  520 ep(i,j)        = amax1(dp(i,j),dp(i,j+1))
c
c calculate cell-boundary dQ/dy 
c
!HPF$ INDEPENDENT
      do 530 j=1,my
!HPF$ INDEPENDENT
      do 530 n=1,4
  530 d1q(n,i,j)     = qnew(n,i,j+1)  -qnew(n,i,j)
c
c calculate cell-boundary d3Q/dy3
c
!HPF$ INDEPENDENT
      do 540 j=2,my-1
!HPF$ INDEPENDENT
      do 540 n=1,4
  540 d3q(n,i,j)     = d1q(n,i,j+1)-2.*d1q(n,i,j)+d1q(n,i,j-1)
!HPF$ INDEPENDENT
      do 542 n=1,4
      d3q(n,i,1)     = d1q(n,i,2)  -2.*d1q(n,i,1)  +d1q(n,i,my-1)
  542 d3q(n,i,my)    = d3q(n,i,1)
c
c calculate Dy{Q} term
c
!HPF$ INDEPENDENT
      do 550 j=1,my
      eps2y        = alpha2*ep(i,j)
      eps4y        = amax1(0.,alpha4  -eps2y)
      phi1         = 1. + (xlamdax(i,j)/xlamday(i,j))**sigma
      xlamda1      = xlamday(i,j)*phi1
      phi2         = 1. + (xlamdax(i,j+1)/xlamday(i,j+1))**sigma
      xlamda2      = xlamday(i,j+1)*phi2
      xlamda       =.5*( xlamda1 +xlamda2 )
!HPF$ INDEPENDENT
      do 550 n=1,4
  550 dwy(n,i,j)   = xlamda*(eps2y*d1q(n,i,j)  -eps4y*d3q(n,i,j))
c
  500 continue

	
      
		
     
      return                                                            
      end                                                               












