c        
c ----------------------------------------------------------------------
c        
      subroutine diff (n,iend)
c        
c compute derivative of the input function at the input points
c   n = number of input points
c   x = input location
c   f = input function
c   fp = computed derivative of f at all input x locations
c   iend = 0 --- extrapolate for end values
c        = 1 --- derivatives are continuous across ends
c        
      include 'dimpar.inc'
      include 'common.inc'
c        
      ip = 2
      n1 = n - 1
      dx = x1(ip) - x1(1)
      df = f1(ip) - f1(1)
      ratio = df/dx
      if (abs(ratio) .le. 0.000001) then
        ang2 = 0.
      else
        ang2 = atan(ratio)
      endif
      dli = dx
      dfi = df
      ang1i = ang2
      dl2 = dx
      do 10 i=2,n1
      ang1 = ang2
      dl1 = dl2
      i1 = i + 1
      dx = x1(i1) - x1(i)
      df = f1(i1) - f1(i)
      ratio = df/dx
      if (abs(ratio) .le. 0.000001) then
        ang2 = 0.
      else
        ang2 = atan2(df,dx)
      end if
      dl2 = dx
      ang = (dl2*ang1 + dl1*ang2)/(dl1 + dl2)
      fp1(i) = tan(ang)
   10 continue
c
      ang2f = ang2
      dff = df
      dlf = dl2
c
      if (iend.eq.1) go to 101
      fp1(1) = 2.*dfi/dli - fp1(ip)
      fp1(n) = 2.*dff/dlf - fp1(n1)
      go to 100
c
  101 ang = (dli*ang2f + dlf*ang1i)/(dli + dlf)
      fp1(1) = tan(ang)
      fp1(n) = fp1(1)
c
  100 return
      end
c
c ----------------------------------------------------------------------
c
      subroutine interp(n1,n2)
c
c cubic interpolation
c
c given the values of a function (f1) and its first derivatives (fp1)
c at n1 values of the independent variable (x1)
c
c find the values of the function (f2) at n2 values of the independent
c variable (x2)
c
c x1 and x2 must be in ascending order
c        
      include 'dimpar.inc'
      include 'common.inc'
c
        jt = 2
      do 10 i=1,n2
      do 20 j=jt,n1
        if (x1(j).ge.x2(i)) go to 30
   20 continue
        j = n1
   30   jt = j
        j1 = j - 1
        dx = x1(j) - x1(j1)
        df = f1(j) - f1(j1)
        dx2 = dx*dx
        dx3 = dx2*dx
        df1 = fp1(j) - fp1(j1)
        fpp = fp1(j) + fp1(j1)
        fppp1 = (2.*df - dx*fpp)/(-dx3/6.)
        fpp1 = df1/dx - .5*dx*fppp1
        dxx = x2(i) - x1(j1)
        dxx2 = .5*dxx*dxx
        dxx3 = dxx2*dxx/3.
   10   f2(i) = f1(j1) + dxx*fp1(j1) + dxx2*fpp1 + dxx3*fppp1
c
      return
      end
