c                                                                       
c ----------------------------------------------------------------------
c                                                                       
      subroutine rk4 (ires,jres,resmax,resavg,
     .                iq1,jq1,q1max,q1avg,nsup)
c                                                                       
c Four-stage Runge Kutta time-stepping scheme
c with added 2nd and 4th order dissipation model of Jameson
c                                                                       
      include 'dimpar.inc'
      include 'common.inc'
	integer temp(2)
	real start,end
	
c

	
      dimension rk(4)
c
	res=0.
      rk(1)        = 0.25
      rk(2)        = 1./3.
      rk(3)        = 0.5
      rk(4)        = 1.0
c
      ires         = -1
      jres         = -1
      resmax       = 0.  
      resavg       = 0. 
      sumres       = 0.
      iq1          = -1
      jq1          = -1
      q1max        = 0.  
      q1avg        = 0. 
      sumq1        = 0.
      nsup         = 0
c
c initialize solution
c

	

!HPF$ INDEPENDENT,NEW(i)
      do 1 j=1,my+1
!HPF$ INDEPENDENT,NEW(n)
      do 1 i=1,mx+1
!HPF$ INDEPENDENT
      do 1 n=1,4
 1       qnew(n,i,j)  = q(n,i,j)
	
c
c Four-stage Runge-Kutta time-stepping scheme
c

      do 100 nstg=1,4
c
c compute artificial viscosity vectors
c
c      if (idissp.eq.1 .and. nstg.le.2) call dissp


	call timer(start,0.)
      if (idissp.eq.1 .and. (nstg.eq.1 .or. nstg.eq.3) ) call dissp
	call timer(end,start)
	
	
c

c compute flux vectors
c
	call timer(start,0.0)
      call flux
	call timer(end,start)
	
	
	
c
c form residual
c
	call timer(start,0.0)
      call residual
	call timer(end,start)
	

	

c residual smoothing for stages 2 and 4
c
	call timer(start,0.0)
!      if ( (nstg .eq. 2) .or. (nstg .eq. 4) ) call smoothing
	call timer(end,start)

c
c time-stepping step
c
	
	
!HPF$ INDEPENDENT,NEW(i)
      do 200 j=2,my
!HPF$ INDEPENDENT,NEW(n)
      do 200 i=2,mx
!HPF$ INDEPENDENT
      do 200 n=1,4
  200 qnew(n,i,j)  = q(n,i,j)  +rk(nstg)*dt(i,j)*res(n,i,j)
	
c
c boundary conditions
c


	call timer(start,0.)
      call bcin
	      call bcout
      call bcper
	call timer(end,start)

		
c
  100 continue
	
c 
c locate max. residual in the continuity equation                       
c                       
	
!      do 900 j=2,my
!      do 900 i=2,mx
!      if (abs(res(1,i,j)) .lt. abs(resmax)) go to 900
!      ires         = i
!      jres         = j
!      resmax       = res(1,i,j)
!  900 sumres       = sumres  +abs(res(1,i,j))
!      resavg       = sumres/float(nx*ny)

	resavg=sum(abs(res(1,:,:)))/float(nx*ny)
	temp=maxloc(abs(res(1,:,:)))
	resmax=res(1,temp(1),temp(2))
	ires=temp(1)
	jres=temp(2)
	
	
	
c 
c locate max. error in density
c                                                                       
!      do 910 j=2,my
!      do 910 i=2,mx
!      error        = qnew(1,i,j)  -q(1,i,j)
!      if (abs(error) .lt. abs(q1max)) go to 910
!      iq1          = i
!      jq1          = j
!      q1max        = error   
!  910 sumq1        = sumq1  +abs(error)
!      q1avg        = sumq1/float(nx*ny)
	temp=maxloc(abs(qnew(1,:,:)-
     $  q(1,:,:)))
	q1max=qnew(1,temp(1),temp(2))-q(1,temp(1),temp(2))
	q1avg=sum(abs(qnew(1,:,:)-q(1,:,:)))/float(nx*ny)
	iq1=temp(1)
	jq1=temp(2)
c
c compute number of supersonic points
c ******** NOTE HERE THE BLADE RELATIVE MACH NO. IS USED ********
c
!      do 920 j=2,my
!      do 920 i=2,mx
!      rho          = qnew(1,i,j)
!      u            = qnew(2,i,j)/qnew(1,i,j)
!      v            = qnew(3,i,j)/qnew(1,i,j)
!      et           = qnew(4,i,j)
!      p            = (gam  -1.)*(et  -.5*rho*(u*u  +v*v 
!     &               -(r(i)*omega)**2))
!      xm2          = (u*u  +v*v)/(gam*p/rho)
!     if (xm2 .ge. 1.) nsup = nsup  +1
!  920 continue
	
	
c
c update q
c
!HPF$ INDEPENDENT
      do 930 j=1,my+1
!HPF$ INDEPENDENT
      do 930 i=1,mx+1
!HPF$ INDEPENDENT
      do 930 n=1,4
  930 q(n,i,j)     = qnew(n,i,j)



c
c update mass flow rate
c
      i            = 2
      sum1         = 0.

!      do 940 j=2,my
!      rdy          = .5*( y(i,j)*r(i)  -y(i,j-1)*r(i) 
!     &                 + y(i-1,j)*r(i-1)  -y(i-1,j-1)*r(i-1))
!      hav          = .5* ( h(i) + h(i-1) )
!      rho          = q(1,i,j)
!      u            = q(2,i,j)/q(1,i,j)
!  940 sum1         = sum1  +(rho*u*rdy*hav)
!      xmdot        = sum1

	
	
c        

       return                                                            
      end                                                               

c
c --------------------------------------------------------
c
      subroutine smoothing
c
      include 'dimpar.inc'
      include 'common.inc'
	real start,end
c
      parameter(l=4,imx=270,imy=70)
      dimension a(icell),b(icell),c(icell)
      dimension res_new(4,icell,jcell)
!!HPF$ ALIGN (:) with BOARD (*,:,*) :: a,b,c
!!HPF$ ALIGN (*,:,*) with BOARD (*,:,*) :: res_new
c
c      rfactor = 0.0

c      cflnorm  = 1.5
c      factor   = 1.0
	res_new=0.
c
      if (cfl .le. cflnorm) return 
c
      eps    = factor*.25*( (cfl/cflnorm)**2 -1.  )

	
	
c
!!HPF$ INDEPENDENT,NEW(i)
!      do j = 2,my
!!HPF$ INDEPENDENT,NEW(n)                                                
!         do i = 2,mx
!            do n = 1,4
!               res_new(n,i,j) = res(n,i,j)
!            enddo
!         enddo
!      enddo
	res_new(1:4,2:mx,2:my)=res(1:4,2:mx,2:my)

	

c
!!HPF$ INDEPENDENT,NEW(i,j)
!      do n = 1,4
!!HPF$ INDEPENDENT                                                
!         do j = 2,my
!1            res_new(n,1,j)      = 0.
!!            res_new(n,mx+1,j)   = 0.
!         enddo
!!HPF$ INDEPENDENT                                                
!         do i = 2,mx
!            res_new(n,i,1)      = 0.
!            res_new(n,i,my+1)   = 0.
!         enddo
!      enddo



	res_new(1:4,1,2:my)=0.
	res_new(1:4,mx+1,2:my)=0.
	res_new(1:4,2:mx,1)=0.
	res_new(1:4,2:mx,my+1)=0.

	

c**********
c doing the I-sweep : from i = 2,mx
c**********
!!HPF$ INDEPENDENT                                                
!      do i = 2,mx+1
!         a(i) = -eps
!         b(i) = (1. +2.*eps)
!         c(i) = -eps
!      enddo

	a(2:mx+1)=-eps
	b(2:mx+1)=(1.+2.*eps)
	c(2:mx+1)=-eps
c
c FORWARD SUBSTITUTION
c
!!HPF$ INDEPENDENT,NEW(j)
      do i = 2,mx
         c(i) = c(i) / b(i)
!!HPF$ INDEPENDENT,NEW(n)
         do j = 2,my
            do n = 1,4
               res_new(n,i,j) = res_new(n,i,j)/b(i)
               res_new(n,i+1,j) =  res_new(n,i+1,j) 
     &                            -res_new(n,i,j)*a(i+1)
            enddo
         enddo
         b(i+1) = b(i+1) -c(i)*a(i+1)
      enddo

	

c
c BACKWARD SUBSTITUTION
c
!!HPF$ INDEPENDENT,NEW(i)                                                
      do j = 2,my
!!HPF$ INDEPENDENT,NEW(n)
         do i = mx-1,2,-1
            do n = 1,4
               res_new(n,i,j) = res_new(n,i,j) 
     $ -c(i)*res_new(n,i+1,j)
            enddo
         enddo
      enddo



c**********
c doing the J-sweep : from j = 2,my
c**********
!!HPF$ INDEPENDENT                                                
      do j = 2,my+1
         a(j) = -eps
         b(j) = (1. +2.*eps)
         c(j) = -eps
      enddo
c
!!HPF$ INDEPENDENT,NEW(i)                                                
      do j = 2,my
!!HPF$ INDEPENDENT,NEW(n)                                                
         do i = 2,mx
            do n = 1,4
               res(n,i,j) = res_new(n,i,j)
            enddo
         enddo
      enddo

c
!!HPF$ INDEPENDENT,NEW(j,i)                                                
      do n = 1,4
!!HPF$ INDEPENDENT
         do j = 2,my
c            res(n,1,j)      = 0.
            res(n,mx+1,j)   = 0.
         enddo
c         do i = 2,mx
c            res(n,i,1)      = 0.
c            res(n,i,my+1)   = 0.
c         enddo
      enddo

c
c FORWARD SUBSTITUTION
c
!!HPF$ INDEPENDENT,NEW(i)                                                
      do j = 2,my
         c(j) = c(j) / b(j)
!!HPF$ INDEPENDENT
         do i = 2,mx
            do n = 1,4
               res(n,i,j) = res(n,i,j)/b(j)
               res(n,i,j+1) = res(n,i,j+1) -res(n,i,j)*a(j+1)
            enddo
         enddo
         b(j+1) = b(j+1) -c(j)*a(j+1)
      enddo
c
c BACKWARD SUBSTITUTION
c
!!HPF$ INDEPENDENT,NEW(i)                                                
      do j = my-1,2,-1
!!HPF$ INDEPENDENT,NEW(n)                                                
         do i = 2,mx
            do n = 1,4
               res(n,i,j) = res(n,i,j) -c(j)*res(n,i,j+1)
            enddo
         enddo
      enddo
c
		
	
      return
      end


