!* -------------------------------------------------------------------
!
!* The following !ode is Dr. Decyk's 1-D PIC code for single-processor
!* computers. [KD]
!*
!* -------------------------------------------------------------------

! * * * periodic 1d electrostatic particle simulation kernel code * * *
! 1D sm version  for public  
! This is a simple 1d skeleton particle-in-cell code designed for
! exploring new computer architectures.  it contains the critical pieces
! needed for depositing charge, advancing particles, and solving the
! field.  the code moves only electrons, with periodic electrostatic
! forces obtained by solving poisson's equation with fast fourier
! transforms.  the only diagnostic is particle and field energy.
! portable gcpic kernel code, using algorithm described in:
! p. c. liewer and v. k. decyk, j. computational phys. 85, 302 (1989).
! written by viktor k. decyk, ucla

     program beps1k

!      parameter( indx =   7, npx =    512, npxb = 512)
       parameter( indx =   7, npx =    5120, npxb = 512)
!        parameter( indx =  11, npx =   409600, npxb = 40960)
!         parameter( indx =  14, npx =  4096000, npxb =   409600)
!benc3 parameter( indx =  18, npx = 40960000, npxb =  4096000)

! Time
      parameter( tend =  5.000, dt = 0.2000000e+00)
! Velocity
      parameter( vtx =   1.000,  vdx= 5.000, vtdx =   0.500)
      parameter( avdx=   5.000)
! Arch , vprocs, boundaries
      parameter( indnvp= 1, idps =    2, idimp =   2)
! Electrons   
      parameter(np=npx+npxb)
      parameter(nx=2**indx,nxh=nx/2)
! Time steps
      parameter(nloop=tend/dt+.0001)
! Vprocs, partitions, particle/partition, 
      parameter(nvp=2**indnvp,nblok=1+1*(nvp-1))
      parameter(npmax=(np/nvp)*1.01+250,nxpmx=(nx-1)/nvp+4)
! Grids/dim
      parameter(kxp=(nxh-1)/nvp+1,kblok=1+1*(nxh/kxp-1))
      parameter(nbmax=1+(2*(npx*vtx+npxb*vtdx)+1.4*npxb*avdx)*dt/nx)
      parameter(ntmax=2*nbmax)

      parameter(npd=128)
!     complex ffc, fc, qc, sct
!      common /large/ part

      dimension part(idimp,npmax,nblok)
      dimension maskp(npmax,nblok)
      dimension q(nxpmx,nblok), fx(nxpmx,nblok)
      complex ,dimension(kxp,kblok) :: qc, fc
      complex ,dimension(kxp,kblok) :: ffc, sct
      dimension mixup(kxp,kblok)
      dimension edges(idps,nblok)
      dimension nxp(nblok), noff(nblok)
      dimension npp(nblok), nps(nblok)

      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)

      dimension ihole(ntmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok)
      dimension jss(idps,nblok), scr(idps,nblok)
      dimension   SUM_0(nblok)
      integer nn
      real dx
      data qme,ax /-1.,.8666667/
     
   
!hpf$ distribute (block)         :: npp
!hpf$ align (i) with npp(i)      :: NOFF,NXP,NPS,sum_0
!hpf$ align (*,i) with npp(i)    :: EDGES,MASKP,Q,FX,IHOLE,JSL,JSR,JSS
!hpf$ align (*,i) with npp(i)    :: SCR,QC,FC,FFC,MIXUP,SCT 
!hpf$ align (*,*,i) with npp(i)  :: PART,SBUFL,SBUFR,RBUFL,RBUFR


!  991 format (5h t = ,i7)
!  992 format (19h * * * q.e.d. * * *)

   interface

      subroutine dcomp1(edges,nxp,noff,nx,kstrt,nvp,idps,nblok,npp)
      dimension edges(idps,nblok)
      dimension nxp(nblok), noff(nblok),npp(nblok)
    
      !hpf$ distribute *(block)      :: npp
      !hpf$ align (*,i) with *npp(i) :: edges
      !hpf$ align (i) with *npp(i  ) :: noff,nxp
      
      end subroutine

      subroutine pistr1 (part,edges,npp,nps,vtx,vdx,npx,nx,idimp,npmax,&
       nblok,idps)

      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok), nps(nblok)
      
      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,*,i) with *npp(i) :: part
      !hpf$ align (*,i) with  *npp(i)  :: edges
      !hpf$ align (i) with *npp(i)     :: nps

      end subroutine
 
      subroutine pdost1 (part,q,npp,noff,qm,idimp,npmax,nblok,nxpmx)

     dimension part(idimp,npmax,nblok), q(nxpmx,nblok)
     dimension npp(nblok), noff(nblok)


     !hpf$ distribute *(block)        :: npp
     !hpf$ align (*,*,i) with *npp(i) :: part
     !hpf$ align (*,i) with  *npp(i)  :: q
     !hpf$ align (i) with *npp(i)     :: noff       
 
     end subroutine
     
      subroutine cppfp1 (f,fc,isign,scr,kstrt,nvp,nxpmx,nblok,kxp,kblok,&
     idps,npp)
      complex, dimension(kxp,kblok) :: fc 
      dimension f(nxpmx,nblok)
      dimension scr(idps,nblok),npp(nblok)
      
      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,i) with  *npp(i)  :: f,fc,scr
     
      end subroutine

      subroutine ppois1 (q,fx,isign,ffc,ax,affp,wp,nx,kstrt,kxp,kblok,npp)

      complex,dimension(kxp,kblok):: q, fx, ffc
      dimension npp(kblok)

      !hpf$ distribute *(block)         :: npp
      !hpf$ align (*,i) with  *npp(i)   :: q,fx,ffc
          
      end subroutine

      subroutine ppush1(part,fx,npp,noff,qtm,dt,ek,idimp,npmax,nblok,&
     nxpmx)

      dimension part(idimp,npmax,nblok), fx(nxpmx,nblok)
      dimension npp(nblok), noff(nblok)

      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,*,i) with *npp(i) :: part
      !hpf$ align (*,i) with *npp(i)   :: fx
      !hpf$ align (i) with *npp(i)     :: noff
      
      end subroutine

          subroutine pmove1 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,&
     jsr,jsl,jss,nx,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
 
      
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)
    
      !hpf$ distribute *(block)        :: npp 
      !hpf$ align (*,*,i) with *npp(i) :: part,sbufr,sbufl,rbufr,rbufl
      !hpf$ align (*,i) with *npp(i)   :: edges,jsl,jsr,jss,ihole
          
      end subroutine

      subroutine pfft1r(f,g,isign,mixup,sct,indx,kstrt,kxp,kblok,npp)
      complex, dimension(kxp,kblok) :: f, g, sct
      dimension mixup(kxp,kblok),npp(kblok)
      
      !hpf$ distribute *(block)        :: npp 
      !hpf$ align (*,i) with *npp(i)   :: f,g,mixup,sct 
      
       end subroutine

      subroutine bitrv1(f,g,mixup,isign,indx,kstrt,kxp,kblok,npp)
      complex, dimension(kxp,kblok) :: f, g
      dimension  mixup(kxp,kblok),npp(kblok)
      
      !hpf$ distribute *(block)        :: npp 
      !hpf$ align (*,i) with *npp(i)   :: f,g,mixup
      end subroutine

      subroutine store_to_temp(position,velocity,part,npp,npmax,nblok,&
         idimp,npps,bb)
      
      real(kind=8),dimension(npmax,nblok) :: position,velocity 
      real(kind=8),intent(OUT) :: bb(2)
      real,dimension(idimp,npmax,nblok) :: part
      integer :: npp(nblok)

      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,i) with *npp(i)    :: position,velocity
      !hpf$ align (*,*,i) with *npp(i)    :: part
      integer ,intent(OUT) :: npps
      end subroutine      


      end interface 
      double precision timebegin,timeend
          
      integer k,j,itime_1,npps
      real(kind=8) :: position(npmax,nblok),velocity(npmax,nblok) 
      real(kind=8) :: bb(2)  ,vel(nblok*npmax),fitime
      !hpf$ align (*,i) with npp(i)  :: position,velocity

!------------------------------------
!initialize 
!------------------------------------
      kstrt = 1
!      open(unit=6,file='output1',form='formatted',status='unknown')
! FOr now , run on only two processor
    
      itime_1 = 0
      anx = float(nx)
      qtme = qme*dt
      affp = anx/float(np)
      zero = 0.
!------------------------------
!calculate partition variables
!------------------------------
      print *,'beg',elapsed_time()
      call dcomp1(edges,nxp,noff,nx,kstrt,nvp,idps,nblok,npp)
      print *,'after decomp',elapsed_time()
!prepare fft tables
      
      isign = 0
      call pfft1r(qc,fc,isign,mixup,sct,indx,kstrt,kxp,kblok,npp)
      print *,'after fft',elapsed_time()
!calculate form factors
      call ppois1(qc,fc,isign,ffc,ax,affp,we,nx,kstrt,kxp,kblok,npp)
      print *,'after ppois',elapsed_time()
!-------------------------------------------------------------
!initial density profile and maxwellian velocity distribution
!background electrons
!-------------------------------------------------------------
      !HPF$ INDEPENDENT,NEW(k,nxp3)
      do 120 k = 1, nblok
         nps(k) = 1
         npp(k) = 0
         nxp3 = nxp(k) + 3
         !HPF$ INDEPENDENT,NEW(j)
         do 110 j = 1, nxp3
               q(j,k) = 0.
  110    continue
  120 continue
      print *,'after loop',elapsed_time()
      if (npx.gt.0) call pistr1(part,edges,npp,nps,vtx,zero,npx,nx,&
        idimp, npmax,nblok,idps)
        print *,'after pistr1',elapsed_time()
!-----------------
!beam electrons
!-----------------

      !HPF$ INDEPENDENT,NEW(k)
      do 140 k = 1, nblok
      nps(k) = npp(k) + 1
  140 continue
      if (npxb.gt.0) call pistr1 (part,edges,npp,nps,vtdx,vdx,npxb,nx,&
       idimp, npmax,nblok,idps)
        print *,'after pistr1',elapsed_time()
!----------------------------------------
!deposit charge for initial distribution
!----------------------------------------
      call pdost1 (part,q,npp,noff,qme,idimp,npmax,nblok,nxpmx)
      print *,'after pdost',elapsed_time()
!add background ion density
      qi0 = -qme/affp
     
      !HPF$ INDEPENDENT,NEW(k)
      do 180 k = 1, nblok
      !HPF$ INDEPENDENT,NEW(j)
      do 170 j = 1, nxp(k)
         q(j+1,k) = q(j+1,k) + qi0
  170 continue
  180 continue 

!=======================================

!* * * start main iteration loop * * *

!=======================================
      call timer(timebegin)
!  500 if (nloop.le.itime_1) go to 2000
   do itime_1 = 1,nloop
      print *,'t = ',itime_1
!     write (6,991) itime_1
!----------------------------------
!transform charge to fourier space
!----------------------------------
      fitime = itime_1
      call store_to_temp(position,velocity,part,npp,npmax,nblok,&
         idimp,npps,bb)

      print *,'after indep loop ..npmax*nblok ',npmax*nblok,' np ',np  
!     java_bbser_("win",fitime,bb,npps,vel)
!     java_bbser("win",fitime,bb,npps,position)
!     java_bbser_("win",fitime,bb,npps,velocity)
      isign = -1
!-------------------------------------------------------------------
!copy data from particle to field partition, and add up guard cells
!-------------------------------------------------------------------
      call cppfp1 (q,qc,isign,scr,kstrt,nvp,nxpmx,nblok,kxp,kblok,idps,npp)
      print *,'after cppfp',elapsed_time()
      call pfft1r(qc,fc,isign,mixup,sct,indx,kstrt,kxp,kblok,npp)
      print *,'after ifft',elapsed_time()

!----------------------------------------
!calculate force/charge in fourier space
!----------------------------------------
      call ppois1 (qc,fc,isign,ffc,ax,affp,we,nx,kstrt,kxp,kblok,npp)
      print *,'after ppois',elapsed_time()
!-------------------------------------
!transform force/charge to real space
!-------------------------------------
      isign = 1
      call pfft1r(fc,qc,isign,mixup,sct,indx,kstrt,kxp,kblok,npp)
      print *,'after dfft',elapsed_time()
!--------------------------------------------------------------------
!copy data from field to particle partition, and copy to guard cells
!--------------------------------------------------------------------
      call cppfp1 (fx,fc,isign,scr,kstrt,nvp,nxpmx,nblok,kxp,kblok,idps,npp)
      print *,'after cppfp',elapsed_time() 
!----------------------------------------
!particle push and charge density update
!----------------------------------------
!    call timera(TIME_PUSH)
!-----------------------------------
!initialize charge density to zero
!-----------------------------------

      !HPF$ INDEPENDENT,NEW(k,nxp3)
      do 1170 k = 1, nblok
      nxp3 = nxp(k) + 3
      !HPF$ INDEPENDENT,NEW(j)
      do 1160 j = 1, nxp3
      q(j,k) = 0.
 1160 continue
 1170 continue
      wke = 0.
!---------------
!push particles
!---------------
       

      call ppush1(part,fx,npp,noff,qtme,dt,wke,idimp,npmax,nblok,nxpmx)
      print *,'after ppush',elapsed_time() 
!------------------------------------------------
!move particles into appropriate spatial regions
!------------------------------------------------
      call pmove1 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,jsr,jsl,&
     jss,nx,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
      print *,'after pmove',elapsed_time()
!---------------
!deposit charge
!---------------
      call pdost1 (part,q,npp,noff,qme,idimp,npmax,nblok,nxpmx)
      print *,'after pdost',elapsed_time() 
!----------------------
!add background charge
!----------------------
 
      !HPF$ INDEPENDENT,NEW(k)
      do 1190 k = 1, nblok
      !HPF$ INDEPENDENT,NEW(j)
      do 1180 j = 1, nxp(k)
      q(j+1,k) = q(j+1,k) + qi0
 1180 continue
 1190 continue
!    call timera(TIME_PUSH)
!------------------
!energy diagnostic
!------------------
      wt = we + wke
      print *,' field, kinetic, total energies = ', we, wke, wt
      !write (6,993) we, wke, wt
!End of do loop
   enddo
!      itime_1 = itime_1 + 1
!      go to 500
! 2000 continue
   
      call timer(timeend )
      print *,'total elapsed time',timeend-timebegin        
!====================================

!* * * end main iteration loop * * *

!====================================
     
      stop
      end

      subroutine store_to_temp(position,velocity,part,npp,npmax,nblok,&
         idimp,npps,bb)
      
      real(kind=8),dimension(npmax,nblok) :: position,velocity 
      real(kind=8) ,intent(OUT) :: bb(2)
      real,dimension(idimp,npmax,nblok) :: part
      integer npp(nblok) 

      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,i) with *npp(i)    :: position,velocity
      !hpf$ align (*,*,i) with *npp(i)    :: part
      integer ,intent(OUT) :: npps

    
      !HPF$ INDEPENDENT,NEW(k,j,t_npps) 
      do k=1,nblok
        do j=1,npp(k)
          position(j,k) = part(1,j,k)
          velocity(j,k) = part(2,j,k)
        enddo
      enddo
      npps = sum(npp)
      bb(1)  = 1
      bb(2)  = npps
      end subroutine 


      subroutine timer(t)

    integer val(8)
    double precision t

    call date_and_time(values=val)
    t = dble(val(8))*1d-3 + dble(val(7)) +          &
        dble(val(6))*60d0 + dble(val(5))*3600d0

    end subroutine timer


      real function elapsed_time()
      integer :: clock_count, clock_rate, clock_max
      integer, save :: last_count
      logical, save :: system_clock_not_yet_called = .true.

      call system_clock(count=clock_count,&                              
                       count_rate=clock_rate, &
                       count_max=clock_max)
      if (system_clock_not_yet_called) then
         system_clock_not_yet_called = .false.
         elapsed_time = 0.0
      else if (clock_count > last_count) then
         elapsed_time = real(clock_count-last_count)/real(clock_rate)
      else
         elapsed_time = real(clock_max-last_count+clock_count)&          
                  /real(clock_rate)
      end if
      last_count = clock_count
      return
     end function elapsed_time

!=========================================================================

! This subroutine determines spatial boundaries for particle
! decomposition, calculates number of grid points in each spatial
! region, and the offset of these grid points from the global address

!=========================================================================
      subroutine dcomp1(edges,nxp,noff,nx,kstrt,nvp,idps,nblok,npp)
      dimension edges(idps,nblok)
      dimension nxp(nblok), noff(nblok),npp(nblok)
    
      !hpf$ distribute *(block)      :: npp
      !hpf$ align (*,i) with *npp(i) :: edges
      !hpf$ align (i) with *npp(i  ) :: noff,nxp
      

      ks = kstrt - 2
      at1 = float(nx)/float(nvp)

      !HPF$ INDEPENDENT,new(k,kb,kr)
      do 10 k = 1, nblok
         kb = k + ks
         edges(1,k) = at1*float(kb)
         noff(k) = edges(1,k) + .5
         edges(2,k) = at1*float(kb + 1)
         kr = edges(2,k) + .5
         nxp(k) = kr - noff(k)
   10 continue
      return
      end
    
!=========================================================================

! This subroutine calculates initial particle co-ordinate
! and velocity, with uniform density and maxwellian velocity with drift.
!=========================================================================
      subroutine pistr1 (part,edges,npp,nps,vtx,vdx,npx,nx,idimp,npmax,&
     nblok,idps)
!ranorm = gaussian random number with zero mean and unit variance
!with spatial decomposition
      
      interface
          function ranorm(d)
            double precision ranorm
          end function
      end interface

      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok), nps(nblok)
      
      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,*,i) with *npp(i) :: part
      !hpf$ align (*,i) with  *npp(i)  :: edges
      !hpf$ align (i) with *npp(i)     :: nps

            
      double precision temp(npx)
!particle distribution
      at1 = float(nx)/float(npx)

      item_cnt = npx/nblok
      do i=1,npx
        temp(i) = ranorm(d)
      enddo

      !HPF$ INDEPENDENT,NEW(j,xt,vxt,k)
      do k=1,nblok
        do j=1,item_cnt
          xt = at1*(float((k-1)*item_cnt+j) - .5)
!maxwellian velocity distribution
         vxt = vtx*temp((k-1)*item_cnt+j)
         npp(k) = npp(k) + 1
         part(1,npp(k),k) = xt
         part(2,npp(k),k) = vxt
       enddo
     enddo

      sum1 = 0.

      !HPF$ INDEPENDENT,NEW(k,j),REDUCTION(sum0,sum1)
      do 40 k = 1, nblok
      sum0 = 0.
      do 30 j = nps(k), npp(k)
         sum0 = sum0 + part(2,j,k)
   30 continue
      sum1 = sum1 + sum0
   40 continue

      sum1 = sum1/float(npx) - vdx

      !HPF$ INDEPENDENT,NEW(k)
      do 60 k = 1, nblok
         do 50 j = nps(k), npp(k)
            part(2,j,k) = part(2,j,k) - sum1
   50    continue
   60 continue
      return

      end
!=========================================================================

! This subroutine calculates particle charge density
! using second-order spline interpolation, with periodi!boundaries.

!=========================================================================
      subroutine pdost1 (part,q,npp,noff,qm,idimp,npmax,nblok,nxpmx)
     dimension part(idimp,npmax,nblok), q(nxpmx,nblok)
     dimension npp(nblok), noff(nblok)


     !hpf$ distribute *(block)        :: npp
     !hpf$ align (*,*,i) with *npp(i) :: part
     !hpf$ align (*,i) with  *npp(i)  :: q
     !hpf$ align (i) with *npp(i)     :: noff       
 
      !HPF$ INDEPENDENT,NEW(k,nnoff)
      do 20 k = 1, nblok
         nnoff = noff(k) - 2
         !HPF$ INDEPENDENT,NEW(j,nn,dx) 
         do 10 j = 1, npp(k)
            nn = part(1,j,k) + .5
            dx = part(1,j,k) - float(nn)
            nn = nn - nnoff
            q(nn,k) = q(nn,k) + qm*(.75 - dx*dx)
            nn = nn + 1
            q(nn,k) = q(nn,k) + .5*qm*(.5 + dx)**2
            nn = nn - 2
            q(nn,k) = q(nn,k) + .5*qm*(.5 - dx)**2
   10    continue
   20 continue
      return
      end

!=========================================================================

!thus subroutine copies data between particle and field partitions.
!if isign = -1, data is copied from particle to field partition.
!if isign = 1, data is copied from field to particle partition.

!=========================================================================
      subroutine cppfp1 (f,fc,isign,scr,kstrt,nvp,nxpmx,nblok,kxp,kblok,&
     idps,npp)
  
      complex, dimension(kxp,kblok) :: fc 
      dimension f(nxpmx,nblok)
      dimension scr(idps,nblok),npp(nblok)
      
      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,i) with  *npp(i)  :: f,fc,scr
     
      ks = kstrt - 2
      moff = nxpmx*nvp
      if (isign) 10, 60, 70
!---------------------------------------
!copy from particle to field partition
!---------------------------------------
      
      !HPF$ INDEPENDENT,NEW(k,j)
   10 do 30 k = 1, kblok
         do 20 j = 1, kxp
            fc(j,k) = cmplx(f(2*j,k),f(2*j+1,k))
   20    continue
   30 continue


      !HPF$ INDEPENDENT,NEW(k,kl,kr)
      do 50 k = 1, kblok
        kl = k + ks
        if (kl.lt.1) then
           scr(1,k) = f(2*kxp+2,kl+nvp)
           scr(2,k) = f(2*kxp+3,kl+nvp)
        else
           scr(1,k) = f(2*kxp+2,kl)
           scr(2,k) = f(2*kxp+3,kl)
        endif
        fc(1,k) = fc(1,k) + cmplx(scr(1,k),scr(2,k))
        kr = k + ks + 2
        if (kr.gt.nvp) then
           scr(1,k) = f(1,kr-nvp)
        else
           scr(1,k) = f(1,kr)
        endif
        fc(kxp,k) = fc(kxp,k) + cmplx(0.,scr(1,k))
  50  continue
       
  60  return
 
!--------------------------------------
!copy from field to particle partition
!--------------------------------------

      !HPF$ INDEPENDENT,NEW(k,j)
   70 do 90 k = 1, kblok
         do 80 j = 1, kxp
            f(2*j,k) = real(fc(j,k))
            f(2*j+1,k) = aimag(fc(j,k))
   80    continue
   90 continue

      !HPF$ INDEPENDENT,NEW(k,kl,kr)
      do 100 k = 1, kblok
         kl = k + ks
         if (kl.lt.1) then
            f(1,k) = aimag(fc(kxp,kl+nvp))
         else
            f(1,k) = aimag(fc(kxp,kl))
         endif
         kr = k + ks + 2
         if (kr.gt.nvp) then
            f(2*kxp+2,k) = real(fc(1,kr-nvp))
            f(2*kxp+3,k) = aimag(fc(1,kr-nvp))
         else
            f(2*kxp+2,k) = real(fc(1,kr))
            f(2*kxp+3,k) = aimag(fc(1,kr))
         endif
 100  continue

      return
      end

!=========================================================================
! This subroutine solves 1d poisson's equation in fourier space for
! force/charge (or convolution of electri!field over particle shape)
! with periodi!boundary conditions.
! The equation used is:
!   fx(k) = -sqrt(-1)*k*g(k)*s(k), where k = 2pi*j/nx, j = fourier mode,
!   g(k) = (affp/k**2)*s(k), and s(k) = exp(-(k*ax)**2/2), except for
!   fx(k=0) = ftx(k=pi) = 0.

! If isign is not equal to 0, force/charge is calculated.
!  aimag(ffc(j,k)) = particle shape factor s for fourier mode jj - 1
!  real(ffc(j,k)) = potential green's function g for fourier mode jj - 1

!=========================================================================
      subroutine ppois1 (q,fx,isign,ffc,ax,affp,wp,nx,kstrt,kxp,kblok,npp)

      complex, dimension(kxp,kblok) :: q, fx, ffc
      dimension npp(kblok)

      !hpf$ distribute *(block)         :: npp
      !hpf$ align (*,i) with  *npp(i)   :: q,fx,ffc
          

      nxh = nx/2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      if (isign.ne.0) go to 30
      if (kstrt.gt.nxh) return
!----------------------------
!prepare form factor array
!----------------------------

      !HPF$ independent,new(k,koff,dkx,at1,j)
      do 20 k = 1, kblok
         koff = kxp*(k + ks) - 1
         do 10 j = 1, kxp
            dkx = dnx*float(j + koff)
            at1 = exp(-.5*(dkx*ax)**2)
            if (dkx.eq.0.) then
               ffc(j,k) = cmplx(0.,1.)
            else
               ffc(j,k) = cmplx(affp*at1/(dkx*dkx),at1)
            endif
   10    continue
   20 continue
      return
!---------------------------------------------
!calculate force/charge and sum field energy
!---------------------------------------------
   30 wp = 0.
      if (kstrt.gt.nxh) go to 60
  
      !HPF$ independent,new(k,koff,j,at1,at2),REDUCTION(sum1,wp)
      do 50 k = 1, kblok
         koff = kxp*(k + ks) - 1
         sum1 = 0.
         do 40 j = 1, kxp
            at1 = real(ffc(j,k))*aimag(ffc(j,k))
            at2 = dnx*float(j + koff)*at1
            fx(j,k) = at2*cmplx(aimag(q(j,k)),-real(q(j,k)))
            sum1 = sum1 + at1*real(q(j,k)*conjg(q(j,k)))
   40    continue
         wp = wp + sum1
   50 continue
   60 continue

      wp = float(nx)*wp
      return
      end

!=========================================================================

! This subroutine updates particle co-ordinate and velocity
! using leap-frog scheme in time and second-order spline interpolation
! in space, with periodi!boundary conditions.
! equations used are:
!   v(t+dt/2) = v(t-dt/2) + (q/m)*fx(x(t))*dt, where q/m is charge/mass,
!   x(t+dt) = x(t) + v(t+dt/2)*dt
! fx(x(t)) is approximated by interpolation from the nearest grid points

!=========================================================================

 subroutine ppush1(part,fx,npp,noff,qtm,dt,ek,idimp,npmax,nblok,&
     nxpmx)

      dimension part(idimp,npmax,nblok), fx(nxpmx,nblok)
      dimension npp(nblok), noff(nblok)

      !hpf$ distribute *(block)        :: npp
      !hpf$ align (*,*,i) with *npp(i) :: part
      !hpf$ align (*,i) with *npp(i)   :: fx
      !hpf$ align (i) with *npp(i)     :: noff


      sum1 = 0.
      !HPF$ independent,new(k,nnoff)
      do 20 k = 1, nblok
      nnoff = noff(k) - 2
      !HPF$ independent,new(j,nn,dx,xt),REDUCTION(sum1)

      do 10 j = 1, npp(k)
!----------------------------
!find interpolation weights
!----------------------------
         nn = part(1,j,k) + .5
         dx = part(1,j,k) - float(nn)
         nn = nn - nnoff
!----------------------------
!find acceleration
!----------------------------
         dx = (.75 - dx*dx)*fx(nn,k) + .5*(fx(nn+1,k)*(.5 + dx)**2 &
             + fx(nn-1,k)*(.5 - dx)**2)
!----------------------------
!new velocity
!----------------------------
         dx = part(2,j,k) + qtm*dx
!----------------------------
!average kineti!energy
!----------------------------
         sum1 = sum1 + (part(2,j,k) + dx)**2
         part(2,j,k) = dx
!----------------------------
!new position
!----------------------------
         part(1,j,k) = part(1,j,k) + dx*dt
   10 continue
   20 continue

!normalize kineti!energy
      ek = ek + .125*sum1
      return
      end




!=========================================================================
      subroutine pmove1 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,&
     jsr,jsl,jss,nx,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
 
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)
    
      !hpf$ distribute *(block)  :: npp 
      !hpf$ align (*,*,i) with *npp(i) :: part,sbufr,sbufl,rbufr,rbufl
      !hpf$ align (*,i) with *npp(i)   :: edges,jsl,jsr,jss,ihole


      anx = float(nx)
      ks = kstrt - 2
      iter = 4
!----------------------------------------------------------------
!debugging section: count total number of particles before move
!----------------------------------------------------------------
      npr = 0

      !HPF$ INDEPENDENT, NEW(k),REDUCTION(npr)
      do 5 k = 1, nblok
         npr = npr + npp(k)
    5 continue

!--------------------------
!buffer outgoing particles
!--------------------------

      !HPF$ INDEPENDENT, NEW(k,kb)
   10 do 30 k = 1, nblok
         kb = k + ks
         jsl(1,k) = 0
         jsr(1,k) = 0
         jss(2,k) = 0
         !HPF$ INDEPENDENT, NEW(j,xt)
         do 20 j = 1, npp(k)
            xt = part(1,j,k)

            if (xt.ge.edges(2,k)) then
               if (jsr(1,k).lt.nbmax) then
                  jsr(1,k) = jsr(1,k) + 1
                  if ((kb+1).eq.nvp) xt = xt - anx
                  sbufr(1,jsr(1,k),k) = xt
                  sbufr(2,jsr(1,k),k) = part(2,j,k)
                  ihole(jsl(1,k)+jsr(1,k),k) = j
               else
                  jss(2,k) = 1
               endif

            elseif (xt.lt.edges(1,k)) then
               if (jsl(1,k).lt.nbmax) then
                  jsl(1,k) = jsl(1,k) + 1
                  if (kb.eq.0) xt = xt + anx
                  sbufl(1,jsl(1,k),k) = xt
                  sbufl(2,jsl(1,k),k) = part(2,j,k)
                  ihole(jsl(1,k)+jsr(1,k),k) = j
               else
                  jss(2,k) = 1
            endif
         endif
   20    continue
         jss(1,k) = jsl(1,k) + jsr(1,k)
   30 continue
!--------------------------------
!check for full buffer condition
!--------------------------------
      ibflg = 0

      !HPF$ INDEPENDENT, NEW(k),REDUCTION(ibflg)
      do 40 k = 1, nblok
         ibflg = ibflg + jss(2,k)
   40 continue

!-------------------------
!copy particle buffers
!-------------------------
   50 iter = iter + 4
!------------------------------------------------------------------------
! Get the number of particles which will come from the left neighbor
! and store it in jsl array.Also Shift send buffers which will go to
! the right
! neighbor ,to to right trough third dimension

!------------------------------------------------------------------------

      JSL(2,:) = CSHIFT(JSR(1,:),-1)
      RBUFL = CSHIFT(SBUFR,SHIFT=-1,DIM=3)
     
!------------------------------------------------------------------------
! Get the number of particles which will come from the right neigbor
! and store it in jsr array.Also Shift send buffers which will go to
! the left  neighbor ,to left  trough third dimension
!------------------------------------------------------------------------
     JSR(2,:) = CSHIFT(JSL(1,:),+1)
     RBUFR = CSHIFT(SBUFL,SHIFT=+1,DIM=3)


!------------------------------------------- 
!check if particles must be passed further
!------------------------------------------- 
      nps = 0
   
      !HPF$ INDEPENDENT,NEW(k),REDUCTION(nps)
      do 110 k = 1, nblok
      jsl(1,k) = 0
      !HPF$ INDEPENDENT,NEW(j)  
      do 90 j = 1, jsr(2,k)
         if (rbufr(1,j,k).lt.edges(1,k)) jsl(1,k) = 1
   90 continue
      jsr(1,k) = 0
      !HPF$ INDEPENDENT,NEW(j)  
      do 100 j = 1, jsl(2,k)
      if (rbufl(1,j,k).ge.edges(2,k)) jsr(1,k) = 1
  100 continue
      nps = nps + (jsl(1,k) + jsr(1,k))
  110 continue

      if (nps.eq.0) go to 150
!-------------------------------------------
!remove particles which do not belong here
!-------------------------------------------
! HPF$ INDEPENDENT,NEW(k,j,kb,xt) 

      !HPF$ INDEPENDENT,NEW(k,kb)
      do 140 k = 1, nblok
      kb = k + ks
      jsl(1,k) = 0
      jss(2,k) = 0
      !HPF$ INDEPENDENT,NEW(j,xt)
      do 120 j = 1, jsr(2,k)
      xt = rbufr(1,j,k)
      if (xt.lt.edges(1,k)) then
         jsl(1,k) = jsl(1,k) + 1
         if (kb.eq.0) xt = xt + anx
         sbufl(1,jsl(1,k),k) = xt
         sbufl(2,jsl(1,k),k) = rbufr(2,j,k)
      else
         jss(2,k) = jss(2,k) + 1
         rbufr(1,jss(2,k),k) = xt
         rbufr(2,jss(2,k),k) = rbufr(2,j,k)
      endif
  120 continue
      jsr(2,k) = jss(2,k)

      jsr(1,k) = 0
      jss(2,k) = 0
      !HPF$ INDEPENDENT,NEW(j,xt)
      do 130 j = 1, jsl(2,k)
      xt = rbufl(1,j,k)
      if (xt.ge.edges(2,k)) then
         jsr(1,k) = jsr(1,k) + 1
         if ((kb+1).eq.nvp) xt = xt - anx
         sbufr(1,jsr(1,k),k) = xt
         sbufr(2,jsr(1,k),k) = rbufl(2,j,k)
      else
         jss(2,k) = jss(2,k) + 1
         rbufl(1,jss(2,k),k) = xt
         rbufl(2,jss(2,k),k) = rbufl(2,j,k)
      endif
  130 continue
      jsl(2,k) = jss(2,k)
  140 continue
!-------------------------------------------
!check if move would overflow particle array
!-------------------------------------------
  150 ierr = 0

      !HPF$ INDEPENDENT,NEW(k),REDUCTION(ierr)
      do 160 k = 1, nblok
      if ((npp(k)+jsl(2,k)+jsr(2,k)-jss(1,k)).gt.npmax) then
         jss(2,k) = 1
      else
         jss(2,k) = 0
      endif
      ierr = ierr + jss(2,k)
  160 continue

      if (ierr.gt.0) then
         write (6,*) 'particle overflow error, ierr =', ierr
         return
      endif

      !HPF$ INDEPENDENT,NEW(k)
      do 200 k = 1, nblok
!------------------------------------------------------
!distribute incoming particles from buffers
!------------------------------------------------------
      jss(2,k) = min0(jss(1,k),jsl(2,k))
      !HPF$ INDEPENDENT,NEW(j) 
      do 170 j = 1, jss(2,k)
      part(1,ihole(j,k),k) = rbufl(1,j,k)
      part(2,ihole(j,k),k) = rbufl(2,j,k)
  170 continue
      if (jss(1,k).gt.jsl(2,k)) then
         jss(2,k) = min0(jss(1,k)-jsl(2,k),jsr(2,k))
      else
         jss(2,k) = jsl(2,k) - jss(1,k)
      endif
      !HPF$ INDEPENDENT,NEW(j)
      do 180 j = 1, jss(2,k)

      if (jss(1,k).gt.jsl(2,k)) then
         part(1,ihole(j+jsl(2,k),k),k) = rbufr(1,j,k)
         part(2,ihole(j+jsl(2,k),k),k) = rbufr(2,j,k)
      else
         part(1,j+npp(k),k) = rbufl(1,j+jss(1,k),k)
         part(2,j+npp(k),k) = rbufl(2,j+jss(1,k),k)
      endif
  180 continue
      if (jss(1,k).le.jsl(2,k)) then
         npp(k) = npp(k) + (jsl(2,k) - jss(1,k))
         jss(1,k) = jsl(2,k)
      endif
      jss(2,k) = jss(1,k) - (jsl(2,k) + jsr(2,k))
      if (jss(2,k).gt.0) then
         jss(1,k) = (jsl(2,k) + jsr(2,k))
         jsr(2,k) = jss(2,k)
      else
         jss(1,k) = jss(1,k) - jsl(2,k)
         jsr(2,k) = -jss(2,k)
      endif
      !HPF$ INDEPENDENT,NEW(j,j1,j2)
      do 190 j = 1, jsr(2,k)
!------------------------------------------------------
!holes left over
!fill up remaining holes in particle array with particles from bottom
!------------------------------------------------------
      if (jss(2,k).gt.0) then
         j1 = npp(k) - j + 1
         j2 = jss(1,k) + jss(2,k) - j + 1
         if (j1.gt.ihole(j2,k)) then
!------------------------------------------------------
!move particle only if it is below current hole
!------------------------------------------------------
            part(1,ihole(j2,k),k) = part(1,j1,k)
            part(2,ihole(j2,k),k) = part(2,j1,k)
         endif
      else
!------------------------------------------------------
!no more holes
!distribute remaining particles from right into bottom
!------------------------------------------------------
         part(1,j+npp(k),k) = rbufr(1,j+jss(1,k),k)
         part(2,j+npp(k),k) = rbufr(2,j+jss(1,k),k)
      endif
  190 continue
      if (jss(2,k).gt.0) then
         npp(k) = npp(k) - jsr(2,k)
      else
         npp(k) = npp(k) + jsr(2,k)
      endif
      jss(1,k) = 0
  200 continue

!------------------------------------------------------
!check if any particles have to be passed further
!------------------------------------------------------
      if (nps.gt.0) go to 50
!check if buffer overflowed and more particles remain to be checked
      if (ibflg.gt.0) go to 10
!debugging section: count total number of particles after move
      nps = 0

      !HPF$ INDEPENDENT,NEW(k),REDUCTION(nps)
      do 205 k = 1, nblok
      nps = nps + npp(k)
  205 continue

      if (npr.ne.nps) write (6,*) 'number error, np-old,np-new=',npr,nps
      return
      end

!=========================================================================

! This subroutine performs a one dimensional real to complex fast
! fourier transform and its inverse, using complex arithmetic,
! for data which is distributed in blocks
!   f = input and output data
!   g = scratch array
!   indx = power of 2 which determines length of transform, nx = 2**indx

!   if isign = 0, the fft tables are prepared
!   if isign = -1, an inverse fourier transform is performed

!   f(n,m) = (1/nx)*sum(f(j,k)*exp(-sqrt(-1)*2pi*nn*jj/nx))
!   if isign = 1, a forward fourier transform is performed
!   f(j,k) = sum(f(n,m)*exp(sqrt(-1)*2pi*nn*jj/nx))
!     where nn = n + kxp*(m - 1) and jj = j + kxp*(k - 1)
!written by viktor k. decyk, ucla
!=========================================================================
      subroutine pfft1r(f,g,isign,mixup,sct,indx,kstrt,kxp,kblok,npp)

      complex, dimension(kxp,kblok) ::  f, g, sct
      dimension mixup(kxp,kblok),npp(kblok)
            
      !hpf$ distribute *(block)        :: npp 
      !hpf$ align (*,i) with *npp(i)   :: f,g,mixup,sct 
      complex  t1
          
      interface
         subroutine bitrv1(f,g,mixup,isign,indx,kstrt,kxp,kblok,npp)
           complex, dimension(kxp,kblok) :: f, g
           dimension  mixup(kxp,kblok),npp(kblok)

           !hpf$ distribute *(block)        :: npp 
           !hpf$ align (*,i) with *npp(i)   :: f,g,mixup

         end subroutine
      end interface 


      nx = 2**indx
      nxh = nx/2
      if (kstrt.gt.nxh) return
      indx1 = indx - 1
      kbs = nxh/kxp
      kxp2 = kxp + 2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      if (isign) 60, 10, 220
!---------------------------
!prepare fft tables
!bit-reverse index table:
!mixup(j,k) = 1 + reversed bits of (j + kxp*(k + kstrt - 2) - 1)
!---------------------------
   10 call bitrv1(f,g,mixup,isign,indx1,kstrt,kxp,kblok,npp)
!---------------------------
!local sine/cosine table for the angles 2*n*pi/kxp
!---------------------------

      !HPF$ INDEPENDENT,NEW(k,j,arg)
      do 50 k = 1, kblok
      do 40 j = 1, kxp
         arg = dnx*float((nxh*(j - 1))/kxp)
         sct(j,k) = cmplx(cos(arg),-sin(arg))
   40 continue
   50 continue
      return
!---------------------------
!inverse fourier transform
!bit-reverse array elements to temporary
!---------------------------
   60 call bitrv1(f,g,mixup,isign,indx1,kstrt,kxp,kblok,npp)

      do 160 l = 1, indx1
      nxs = 2**(l - 1)
      kxs = nxs/kxp
!---------------------------
!local calculation
!---------------------------
      if (kxs.eq.0) then

         !HPF$ INDEPENDENT,NEW(k,j,lb)
         do 110 k = 1, kblok
         do 90 j = 1, kxp
            lb = (j - 1)/nxs
            if (lb.eq.(2*(lb/2))) then
               f(j,k) = g(j+nxs,k)
            else
               f(j,k) = g(j-nxs,k)
            endif
   90    continue
  110    continue  
!---------------------------
!perform reduction
!---------------------------
         km = kxp/nxs
         !HPF$ INDEPENDENT,NEW(k,j,kb,lb,t1) 
         do 111 k = 1, kblok
         do 101 j = 1, kxp
            kb = j - 1
            lb = kb/nxs
            t1 = sct(1+km*(kb-nxs*lb),k)
            if (lb.eq.(2*(lb/2))) then
               g(j,k) = g(j,k) + t1*f(j,k)
            else
               g(j,k) = f(j,k) - t1*g(j,k)
            endif
  101  continue
  111  continue
      else
!---------------------------
!copy data
!---------------------------

         !HPF$ INDEPENDENT,NEW(k,j,kb,lb)
         do 130 k = 1, kblok
            kb = k + ks
            lb = kb/kxs
            kb = kb + 1
        do 120 j = 1, kxp
        if (lb.eq.(2*(lb/2))) then
           f(j,k) = g(j,kb+kxs)
        else
           f(j,k) = g(j,kb-kxs)
        endif
 120    continue

  130    continue
!------------------
!perform reduction
!------------------
         km = nxh/nxs
         dns = dnx*float(km)

         !HPF$ INDEPENDENT,NEW(k,j,kb,lb,arg,t1)
         do 150 k = 1, kblok
            kb = k + ks
            lb = kb/kxs
            kb = kxp*(kb - kxs*lb) - 1
            lb = lb - 2*(lb/2)
            do 140 j = 1, kxp
               arg = dns*float(j + kb)
               t1 = cmplx(cos(arg),-sin(arg))
               if (lb.eq.0) then
                  g(j,k) = g(j,k) + t1*f(j,k)
               else
                  g(j,k) = f(j,k) - t1*g(j,k)
               endif
  140       continue
  150    continue
      endif
  160 continue
!unscramble coefficients and normalize result

      !HPF$ INDEPENDENT,NEW(k,j,kb,kk)
      do 180 k = 1, kblok
         kb = k + ks
         kk = kbs - kb
      do 170 j = 2, kxp
         f(j,k) = g(j,kk)
 170  continue
      if (kb.gt.0) f(1,k) = g(1,kk+1)

  180  continue
!------------------
!perform reduction
!------------------
         ani = .5/float(nx)
       
         !HPF$ INDEPENDENT,NEW(k,j,kb,koff,jj,arg)
         do 210 k = 1, kblok
            kb = k + ks
            koff = kxp*kb - 1
            do 190 j = 2, kxp
               jj = kxp2 - j
               arg = dnx*float(j + koff)
               g(j,k) = ani*((g(j,k) + conjg(f(jj,k))) &
                 - (g(j,k) &
                 - conjg(f(jj,k)))*cmplx(sin(arg),cos(arg)))
  190       continue
            if (kb.eq.0) then
               g(1,k) = 2.*ani*cmplx(real(g(1,k)) + &
                        aimag(g(1,k)),real(g(1,k))&
                        - aimag(g(1,k)))
            else
               arg = dnx*float(1 + koff)
               g(1,k) = ani*((g(1,k) + conjg(f(1,k))) - &
                        (g(1,k) - conjg(f(1,k)))* &
                        cmplx(sin(arg),cos(arg)))
            endif
!copy result to destination
            do 200 j = 1, kxp
               f(j,k) = g(j,k)
  200       continue
  210 continue
      return
!------------------
!forward fourier transform
!scramble coefficients
!------------------

      !HPF$ INDEPENDENT,NEW(k,j,kb,kk)
  220 do 240 k = 1, kblok
         kb = k + ks
         kk = kbs - kb
      do 230 j = 2, kxp
         g(j,k) = f(j,kk)
  230 continue
      if (kb.gt.0) g(1,k) = f(1,kk+1)

  240 continue
!------------------
!perform reduction
!------------------

         !HPF$ INDEPENDENT,NEW(k,j,kb,koff,jj,arg)
      do 270 k = 1, kblok
         kb = k + ks
         koff = kxp*kb - 1
         do 250 j = 2, kxp
            jj = kxp2 - j
            arg = dnx*float(j + koff)
            f(j,k) = (f(j,k) + conjg(g(jj,k))) - &
                    (f(j,k) - conjg(g(jj,k)))*cmplx(sin(arg),-cos(arg))
  250    continue
         if (kb.eq.0) then
            f(1,k) = cmplx(real(f(1,k)) + aimag(f(1,k)),real(f(1,k)) - &
                     aimag(f(1,k)))
         else
            arg = dnx*float(1 + koff)
            f(1,k) = (f(1,k) + conjg(g(1,k))) - &
                    (f(1,k) - conjg(g(1,k)))*cmplx(sin(arg),-cos(arg))
         endif
!copy result to temporary
         do 260 j = 1, kxp
            g(j,k) = f(j,k)
  260    continue
  270 continue
!----------------------------
!bit-reverse array elements
!----------------------------
      call bitrv1(g,f,mixup,isign,indx1,kstrt,kxp,kblok,npp)
        
      do 370 l = 1, indx1
         nxs = 2**(l - 1)
         kxs = nxs/kxp
!-------------------------
!local calculation
!-------------------------
         if (kxs.eq.0) then
!copy local data
              
            !HPF$ INDEPENDENT,NEW(k,j,lb)
            do 320 k = 1, kblok
               do 300 j = 1, kxp
                  lb = (j - 1)/nxs
                  if (lb.eq.(2*(lb/2))) then
                     g(j,k) = f(j+nxs,k)
                  else
                     g(j,k) = f(j-nxs,k)
                  endif
  300          continue
  320       continue
!perform reduction
               km = kxp/nxs

               !HPF$ INDEPENDENT,NEW(k,j,lb,kb,t1)
               do 321 k = 1, kblok
               do 311 j = 1, kxp
                 kb = j - 1
                 lb = kb/nxs
                 t1 = conjg(sct(1+km*(kb-nxs*lb),k))
                 if (lb.eq.(2*(lb/2))) then
                    f(j,k) = f(j,k) + t1*g(j,k)
                 else
                    f(j,k) = g(j,k) - t1*f(j,k)
                 endif
  311         continue
  321      continue
         else

            !HPF$ INDEPENDENT,NEW(k,j,lb,kb)
            do 340 k = 1, kblok
               kb = k + ks
               lb = kb/kxs
               kb = kb + 1
               lb = lb - 2*(lb/2)
        do 330 j = 1, kxp
           if (lb.eq.0) then
              g(j,k) = f(j,kb+kxs)
           else
              g(j,k) = f(j,kb-kxs)
           endif
 330    continue

  340    continue
!------------------
!perform reduction
!------------------
         km = nxh/nxs
         dns = dnx*float(km)

         !HPF$ INDEPENDENT,NEW(k,j,lb,kb,arg,t1)
         do 360 k = 1, kblok
         kb = k + ks
         lb = kb/kxs
         kb = kxp*(kb - kxs*lb) - 1
         lb = lb - 2*(lb/2)
         do 350 j = 1, kxp
         arg = dns*float(j + kb)
         t1 = cmplx(cos(arg),sin(arg))
         if (lb.eq.0) then
            f(j,k) = f(j,k) + t1*g(j,k)
         else
            f(j,k) = g(j,k) - t1*f(j,k)
         endif
  350    continue
  360    continue
      endif
  370 continue
      return
      end
!=========================================================================
!this subroutine performs bit-reversal of distributed complex data,
!that is, it calculates  
!   g(j,k) = f(jj,kk), where
!   kk = (mixup(j,k)-1)/kxp+1 and jj = mixup(j,k)-kxp*(kk-1), and
!   mixup(j,k) = 1 + reversed bits of (j + kxp*(k + kstrt - 2) - 1)
!   f = complex input array
!   g = complex output array
!   mixup = array of bit reversed addresses
!   if isign = 0, bit-reverse array is prepared
!   if isign = 1, bit-reversal is performed
!written by viktor k. decyk, ucla
!=========================================================================
      subroutine bitrv1_XXX(f,g,mixup,isign,indx,kstrt,kxp,kblok,npp)
     
      complex, dimension(kxp,kblok) ::  f, g
      dimension  mixup(kxp,kblok),npp(kblok)
     
      !hpf$ distribute *(block)        :: npp 
      !hpf$ align (*,i) with *npp(i)   :: f,g,mixup

      nx = 2**indx
      if (kstrt.gt.nx) return
      ks = kstrt - 2
      if (isign.ne.0) go to 40
!---------------------------------
!prepare bit-reverse index table
!---------------------------------

         !HPF$ INDEPENDENT,NEW(k,j,koff,lb,ll,jb,it) 
      do 30 k = 1, kblok
         koff = kxp*(k + ks) - 1
         do 20 j = 1, kxp
            lb = j + koff
            ll = 0
            do 10 l = 1, indx
               jb = lb/2
               it = lb - 2*jb
               lb = jb
               ll = 2*ll + it
   10       continue
            mixup(j,k) = ll + 1
   20    continue
   30 continue
      return
!---------------------------------
!bit-reverse array elements
!---------------------------------

      !HPF$ INDEPENDENT,NEW(k,j,ll,kk,jj)
  40  do 60 k = 1, kblok
      do 50 j = 1, kxp
        ll = mixup(j,k)
        kk = (ll - 1)/kxp + 1
        jj = ll - kxp*(kk - 1)
        g(j,k) = f(jj,kk)
  50  continue
  60  continue

      return
      end
!=========================================================================
!this program calculates a random number y from a gaussian distribution
!with zero mean and unit variance, according to the method of
!mueller and box:
!   y(k) = (-2*ln(x(k)))**1/2*sin(2*pi*x(k+1))
!   y(k+1) = (-2*ln(x(k)))**1/2*cos(2*pi*x(k+1)),
!where x is a random number uniformly distributed on (0,1).
!written for the ibm by viktor k. decyk, ucla
!=========================================================================
      function ranorm_XX(d)
      integer r1,r2,r4,r5
      double precision ranorm,h1l,h1u,h2l,r0,r3,asc,bsc,temp
      save iflg,r1,r2,r4,r5,h1l,h1u,h2l,r0
      data r1,r2,r4,r5 /885098780,1824280461,1396483093,55318673/
      data h1l,h1u,h2l /65531.0d0,32767.0d0,65525.0d0/
      data iflg,r0 /0,0.0d0/
      if (iflg.eq.0) go to 10
      ranorm = r0
      r0 = 0.0d0
      iflg = 0
      return
   10 isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r1 - (r1/isc)*isc
      r3 = h1l*dfloat(r1) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r2/isc
      isc = r2 - i1*isc
      r0 = h1l*dfloat(r2) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r2 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r1 = r3 - dfloat(isc)*bsc
      temp = dsqrt(-2.0d0*dlog((dfloat(r1) + dfloat(r2)*asc)*asc))
      isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r4 - (r4/isc)*isc
      r3 = h2l*dfloat(r4) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r5/isc
      isc = r5 - i1*isc
      r0 = h2l*dfloat(r5) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r5 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r4 = r3 - dfloat(isc)*bsc
      r0 = 6.28318530717959d0*((dfloat(r4) + dfloat(r5)*asc)*asc)
      ranorm = temp*dsin(r0)
      r0 = temp*dcos(r0)
      iflg = 1
      return
      end



