#include <sys/types.h>
#include <rpc/types.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpc/xdr.h> 
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>


extern int   errno;
extern char  *sys_errlist[];
extern int  t_errno;
extern char *t_errlist[];
int max_push_id=0;

#define HOST   (pghpf_myprocnum_() == 0)  
#define ME     (pghpf_myprocnum_())
#define SERVER_CONNECT_RETRY_DELAY 1
#define QLEN            5
#define MAX_NAME_LENGTH 256
#define MAX_DIMS        7
#define MAX_VARS        256
/*#define NOFILE          20 */ /* There are 20 clients we can service */

#define T_CHAR          0
#define T_INT           1
#define T_FLOAT         2
#define T_LONG          3
#define T_DOUBLE        4
#define T_BOOL          5 
#define T_BYTE          6
#define T_SHORT         7
#define T_RECORD        8
#define REGBUFSIZE      5012   
#define UNCONNECTED     0
#define CONNECTED       1
/* debugger commands */
#define REGDATAREQ      1 /* request the registrated data */
#define CONNECT         2 /* make conenction to HPF server */
#define CONTINUE        1
#define STOP_COMM       0
#define MAXVARLEN       256

int     msock;                  /* master server socket         */
fd_set  rfds;                   /* read file descriptor set     */
fd_set  afds;                   /* active file descriptor set   */
int     alen;                   /* from-address length          */
int     fd, nfds;
char    portnum[64]="3335";

int  SOCK_RECV;

/* Use the registrated variable "sock_recv_bits " as temporary to
   copy the select events to other processors */

typedef struct push_data{
  int  type;
  char name[MAXVARLEN];
  int  ndims;  /* 0 for simple variables,ndims for array vars */ 
  int  dim_info[MAX_DIMS] ;
  int  sockfd; /* socket file descriptor */
  int  port;
  int  sizeof_data; /* total size of push_data in bytes */
  long buf; /* includes actual data whatever it is */
  int  elsize; /* sizeof of each element */
  int  connected;
  int  nelems ; /* total element count */
  int  tind; /* index to type_list,returns the pointer to function */
  int  javabuflen;
  unsigned long  client_addr;
  int scope;
  
  long data_desc;
  int  isDistributed;
  int  localIndex;/* If I am distributed,this points to my local data */
  /* this will be used to transfer from parallel dat ato local data */
}DATA;
  
typedef struct send_buf{   
  char *sbuf;
  int   len;
  XDR   xdrs;
}SBUF;


DATA  stack[MAX_VARS]; 

void err(char *) ;
void configure_client_(int *,char *,int *);
int server(int ,int ,int );

int find_var(int scope,char *name)
{
  int i,found;
  
  for(i=0,found=-1;i< max_push_id;i++){
    if ((stack[i].scope==scope)&&
	(!strcmp(stack[i].name,name))){
       found = i;
       break;
    }
  }
  
  return found;
}

void register_var_(int *var_type, int *var_id, long var,char *name, 
int *var_ndims, int *dim_info, int *scope, int *isDist, ...)
{
  va_list ap;
  int item_len,nelems,i;
  int type = *var_type,ndims=*var_ndims;
  float *ptr;
  int id,myid,found;

  printf("I will register -%s- \n",name);
  
  /* look for whether we registered this variable with name,"name", and
     scope ,"scope" */
  /* this is just debugging tool */
  if (find_var(*scope,name) >= 0){
    printf("We registered... before ...this var=%s scope=%d\n",name,
       *scope );
    return;
  }

  id =  max_push_id++; 
  *var_id = id;

  /* It is clear that we registered the local temporary created for any
     distributed variable one step after in HPF program */
  if (*isDist)  
    stack[id].localIndex  = id+1;
  else
    stack[id].localIndex = -1;
  
  stack[id].type = type;
  stack[id].ndims = ndims;
  strcpy(stack[id].name, name);
  
   /* calculate size of component length in bytes */
   /* transform bool,byte and short vars to int var */

   stack[id].buf = var;
   switch(type){
   case T_CHAR  :  item_len = 1;break;
   case T_INT   :  item_len = 4;break;
   case T_FLOAT :  item_len = 4;break;
   case T_LONG  :  item_len = 8;break;
   case T_DOUBLE:  item_len = 8;break;
   /* convert bool,byte and short vars into integer cars */
   case T_BOOL  :  item_len = 1;break;
   case T_BYTE  :  item_len = 1;break;
   case T_SHORT :  item_len = 2;break;
   }

   stack[id].elsize = item_len;
    /* if variable is array */
   for(i=0;i<ndims;i++)
     stack[id].dim_info[i] = dim_info[i];
   
   stack[i].sockfd = -1;
   stack[i].port   = 0;

   /*calculate total length of variable in bytes:calculate # of elemens 
   first.The the total sizeof */

   for(i=0,nelems = 1;i<ndims;i++)
     nelems *= (stack[id].dim_info[i])    ;
   stack[id].nelems = nelems;

   stack[id].sizeof_data = nelems*(stack[id].elsize);
   stack[id].connected = UNCONNECTED; 
   stack[id].javabuflen = stack[id].sizeof_data;

   /* XDR encode the char/bool/short/byte data as integer.Therefore
      XDR encoded buffer length should be made multiple of 4 */ 
   if ((type == T_BOOL)||(type == T_CHAR)||(type == T_BYTE))
      stack[id].javabuflen *= 4;
   else if (type == T_SHORT)
      stack[id].javabuflen *= 2;
   else if (type == T_RECORD)
      stack[id].javabuflen = nelems*(stack[id].elsize);

   stack[id].scope = (*scope);
   
   stack[id].isDistributed = (*isDist);
   
   /* read data descriptor*/
   va_start(ap,isDist);
   va_arg(ap, long); 
   va_arg(ap, long); 
   /* read data_desc */
   stack[id].data_desc = va_arg(ap, long); 
   va_end(ap);

   if (!strcmp(stack[id].name,"sock_recv_bits"))
      SOCK_RECV = id;

   printf("variable =%s with scope =%d id= %d was registered totalsize=%d javalen = %d \n",name,*scope,id,stack[id].sizeof_data,stack[id].javabuflen);
}

XDR_data(id,XDR_flag,xdrs)
int id ;
int XDR_flag;
XDR  *xdrs;
{

  int datalen,elsize,type,ndims;
  int nelems;
  void *dptr;
  int i,tind,j;
  bool_t ret;
  float *fp;
  

  elsize  = stack[id].elsize;
  type    = stack[id].type;
  ndims   = stack[id].ndims;

  type = stack[id].type;
  nelems = stack[id].nelems;

  switch(type){
   case T_CHAR : 
               dptr = (char *)(stack[id].buf);
               if (ndims > 0) /* Character string */
                 ret = xdr_vector(xdrs,(char *)dptr,nelems,elsize,xdr_char);
               else /* just only one character . sent it as int var.*/
                 ret = xdr_char(xdrs,dptr);
               break;

  case T_INT  : case T_BOOL : case T_SHORT:case T_BYTE:
               dptr = (int *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_int);
               else 
                 ret = xdr_int(xdrs,dptr);
               break;
   case T_FLOAT: 
               dptr = (float *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_float);
               else
                 ret = xdr_float(xdrs,dptr); 
               break;
   
   case T_LONG : 
               dptr = (long int *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_long);
               else
                 ret = xdr_long(xdrs,dptr); 
               break;

   case T_DOUBLE:  
               dptr = (double *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_double);
               else
                 ret = xdr_double(xdrs,dptr); 
               break;
 
  }
  
   if (ret == FALSE){
      printf("There is error while encoding send data  with type=%d\n",type);
      exit(0);  
  }
}

/* id for stack array . XDR_flag : XDR_DECODE or XDR_ENCODE */
encode_send_data(id,sendbuf,XDR_flag)
int id ;
SBUF *sendbuf;
int XDR_flag;
{
  XDR xdrs;
  char *buf;
  int datalen;
  int ret;
  float *fp;
  int i,j;
  
  
  datalen = stack[id].javabuflen;
  buf = (char *)malloc(datalen); 
  
  xdrmem_create(&xdrs,buf,datalen,XDR_flag);
  XDR_data(id,XDR_flag,&xdrs);

  sendbuf->sbuf = buf;
  sendbuf->len  = xdr_getpos(&xdrs);
  sendbuf->xdrs = xdrs;

}

send_data_(int id,int sockfd)
{
  SBUF sendbuf;
  int  len;
  char *ptr;
  int nleft,nwritten;
  float *fp;
  int i,j;
  

  encode_send_data(id,&sendbuf,XDR_ENCODE);
  len = sendbuf.len;
  ptr = (char *)(sendbuf.sbuf);
  nleft = len;

  while (nleft > 0) {
    nwritten = write(sockfd, ptr, nleft);
    if (nwritten<=0)
      err("send_buffer: nwritten <= 0");
    nleft -= nwritten;
    ptr += nwritten;
  }
  xdr_destroy(&(sendbuf.xdrs));
  free(sendbuf.sbuf);
  printf("I send data for id=%d sett len =%d\n",id,len);

}  


decode_recv_data(id,recvbuf,datalen,XDR_flag)
int id ;/* id for stack array . XDR_flag : XDR_DECODE or XDR_ENCODE */
char  *recvbuf;
int datalen; /* total data length received */
int XDR_flag;
{
  XDR xdrs;
  char *buf;

  buf = (char *)malloc(datalen);
  xdrmem_create(&xdrs,buf,datalen,XDR_flag);
  bcopy(recvbuf,buf,datalen);
  XDR_data(id,XDR_flag,&xdrs);
  
  xdr_destroy(&xdrs);
  free(buf);

}


recv_data_(int id,int sockfd)
{
  int  len;
  XDR xdrs;
  char *buf;
  int datalen,type,ndims;
  
  datalen = stack[id].javabuflen; 
  type    = stack[id].type;
  ndims   = stack[id].ndims;

  buf = (char *)malloc(datalen);
  len = read_raw_data(sockfd,buf,datalen);
  decode_recv_data(id,buf,len,XDR_DECODE);

  printf("%d bytes read for var =%d original = %d\n",len,id,datalen);
}  


int read_raw_data(fd, buf, nbytes)
register int    fd;
register char   *buf;
register int    nbytes;
{
    int nleft, nread;
    char *ptr = buf;
    
    nleft = nbytes;
    while (nleft > 0) {
        nread = read(fd, ptr, nleft);
        if (nread < 0)
            return(nread);      /* error, return < 0 */
        else if (nread == 0)
            break;          /* EOF */

        nleft -= nread;
        ptr   += nread;
    }
    printf("%d bytes read for var =%d original = %d left =%d\n",ptr-buf,
         fd,nbytes,nleft);

    return(nbytes - nleft);     /* return >= 0 */

}

send_raw_data(int sockfd,char *buf,int len)
{
  char *ptr;
  int nleft,nwritten;

  ptr   = buf;
  nleft = len;

  while (nleft > 0) {
    nwritten = write(sockfd, ptr, nleft);
    if (nwritten<=0)
      err("send_buffer: nwritten <= 0");
    nleft -= nwritten;
    ptr += nwritten;
  }
  printf("I send raw data sent len =%d\n",len);
}



send_registration_info_(int tag,char *variableName,int type,int ndims
		       ,int nelems,int *dims,int sockfd)
{
   char buffer[REGBUFSIZE];
   XDR xdrs;
   int len;
   

   /* Send the registration info about "variableName":
      tag+variableName+type+ndims+dims */
   /* Be aware that xdr_string will put the length of variableName
      into the beginning of translated version */
   xdrmem_create(&xdrs,buffer,REGBUFSIZE,XDR_ENCODE);

   xdr_int(&xdrs,&tag);
   xdr_string(&xdrs,&variableName,strlen(variableName));
   xdr_int(&xdrs,&type);
   xdr_int(&xdrs,&ndims);
   xdr_int(&xdrs,&nelems);
   xdr_vector(&xdrs,(char *)dims,ndims,sizeof(int),xdr_int);

   len = xdr_getpos(&xdrs);
   send_raw_data(sockfd,buffer,len);
   xdr_destroy(&xdrs);

}

sendrecv_string(char *buf,int sockfd,int XDR_FLAG)
{
   char *buffer;
   XDR xdrs;
   int len;

   len = strlen(buf);
   buffer = (char *)malloc(len);
   xdrmem_create(&xdrs,buffer,len,XDR_FLAG);
   
   xdr_string(&xdrs,&buf,len);
   len = xdr_getpos(&xdrs);
   send_raw_data(sockfd,buffer,len);
   xdr_destroy(&xdrs);
   free(buffer);
}


send_registered_data(int scope,char *variableName,int sockfd)
{ 
  int i,id;
  int j,event;
    

  if (HOST)
  {
     id = find_var(scope,variableName);
     if (id < 0){
        printf("I could'nt find the variable = %s with scope=%d in send_regis                 tered_data\n",variableName,scope);
        event = 1;
     }
     /* If the variable is local, then */
     else if (!(stack[id].isDistributed )){
        send_data_(id,sockfd);
        printf("I send the local variable =%s with scope=%d\n",
           variableName,scope);
        event = 2;
	
     }
  }
  /* distribute the event processor_0 found to others */
  distribute_oneitem_(&event); 
  if (event == 1)
    printf("Send Error Mesg to Java Applet \n");
  else if (event == 2)
    return;
      
  /* Send the distributed data. First copy from parallel data to serial one */
  copy_from_parallel_to_serial(id);
  
  if (HOST ){
  /* set the message tag to zero */
    send_registration_info_(0, variableName, stack[id].type, stack[id].ndims
	  ,stack[id].nelems, stack[id].dim_info, sockfd); 
    send_data_(stack[id].localIndex,sockfd); 
  }
  
}

copy_from_parallel_to_serial(id)
{
   int li = stack[id].localIndex;/* localIndex = id-1 */

   printf("li = %d type = %d ndims=%d ME = %d\n",stack[id].localIndex,
	  stack[id].type, stack[id].ndims,pghpf_myprocnum_());
   
   switch(stack[id].type){
   case T_INT:
     switch(stack[id].ndims)
     {
       case 1: 
        copy1dint_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
        break;
       case 2:
        copy2dint_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
        break;
       case 3:
        copy3dint_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
        break;
     }
     break;
     
   case T_FLOAT: 
     switch(stack[id].ndims)
     {
       case 1:
         copy1dfloat_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 2:
         copy2dfloat_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 3:
         copy3dfloat_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
     }
     break;

   case T_LONG:
     switch(stack[id].ndims)
     {
       case 1:
         copy1dlong_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 2:
         copy2dlong_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 3:
         copy3dlong_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
      }
      break;

   case T_DOUBLE: 
     switch(stack[id].ndims)
     {
       case 1:
         copy1ddouble_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 2:
         copy2ddouble_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 3:
         copy3ddouble_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
     }
     break;

   case T_BOOL: 
     switch(stack[id].ndims)
     {
       case 1:
         copy1dboolean_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 2:
         copy2dboolean_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 3:
         copy3dboolean_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
     }
     break;

   case T_BYTE:  
    switch(stack[id].ndims)
     {
       case 1:
         copy1dbyte_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 2:
         copy2dbyte_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 3:
         copy3dbyte_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
     }
     break;  

   case T_SHORT:  
     switch(stack[id].ndims)
     {
       case 1:
         copy1dshort_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 2:
         copy2dshort_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
       case 3:
         copy3dshort_(stack[id].buf,stack[li].buf,
           stack[id].data_desc,stack[li].data_desc);
         break;
     }
    break;

   }
   
}


init_hpf_server_()
{
  if (HOST)
     msock = passivesock(portnum, "tcp",5);
  distribute_oneitem_(&msock);

  nfds = getdtablesize();
  FD_ZERO(&afds);
  FD_SET(msock, &afds);

  printf("MSOCK = %d\n",msock);
  
}


void hpf_server_(int *CurLineNumber,int *scope)
{
  struct sockaddr_in fsin;        /* the from address of a client */
  fd_set  rfds,wfds;              /* read file descriptor set     */
  int     alen;                   /* from-address length          */
  int     fd;
  int     sresult;
  int     iter = 0,nready;
  struct timeval tval={0,0};
  
    
  while (1) { 
    bcopy((char *)&afds, (char *)&rfds, sizeof(rfds));
    if (HOST){

        if (nready=select(nfds, &rfds,(fd_set *)0, (fd_set *)0,&tval)<0)
 	    err("select: error \n");  

        if (FD_ISSET(msock, &rfds)) {
            int     ssock;

            alen = sizeof(fsin);
            ssock = accept(msock, (struct sockaddr *)&fsin, &alen);
            if (ssock < 0)
	      err("accept error\n");
            FD_SET(ssock, &afds);
         }
     }

    distribute_sock_event(&rfds);
    for (fd=0; fd<nfds; ++fd)
      if (fd != msock && FD_ISSET(fd, &rfds))
    	 sresult = server(fd,*CurLineNumber,*scope);     
    
  }   
}


int
echo(fd)
int     fd;
{
        char    buf[REGBUFSIZE];
        int     cc;

        printf("ENTERED echo ME = %d\n",ME);
	if (!HOST)
          return;
	
        cc = read(fd, buf, sizeof buf);
        if (cc < 0)
                err("echo read: \n");
        if (cc && write(fd, buf, cc) < 0)
                err("echo write: \n");
        return cc;
}


distribute_sock_event(fd_set *rfds)
{
    long *fp;
    int i,len = howmany(FD_SETSIZE, NFDBITS);

    /* distribute the select events P_0 found to others */          
    fp = (long *)(stack[SOCK_RECV].buf);

    for(i=0;i<len;i++) fp[i] = rfds->fds_bits[i];
    distribute_item_(stack[SOCK_RECV].buf,stack[SOCK_RECV].data_desc);
    for(i=0;i<len;i++) rfds->fds_bits[i] = fp[i];
}

int server(int slavesockfd,int CurLineNumber,int scope)
{
   int command,lineNumber,len;
   char varName[256];
   
   printf("ENTERED into server ME = %d\n",ME);
   if (HOST )
      read_cmnd(slavesockfd,&command,&lineNumber,varName);

   printf("command = %d ln = %d varname=%s ME = %d\n",
	  command,lineNumber,varName,ME);
  /*  send_registered_data(scope,varName,slavesockfd);*/
      
}


read_cmnd(int sockfd,int *command,int *lineNumber,char *variableName)
{
   char buffer[REGBUFSIZE]="",cmndLineBuf[8]="";
   XDR xdrs;
   int strLen=-1,totalLen=-1;

   printf("before read_cmnd..... \n");

        
   read_raw_data(sockfd,cmndLineBuf,2*sizeof(int));
   xdrmem_create(&xdrs,cmndLineBuf,2*sizeof(int),XDR_DECODE); 
   xdr_int(&xdrs,command);
   xdr_int(&xdrs,lineNumber);
   
   read_raw_data(sockfd,buffer,sizeof(int));
   xdrmem_create(&xdrs,buffer,sizeof(int),XDR_DECODE); 
   xdr_int(&xdrs,&strLen);
   totalLen = strLen+4-strLen%4;

   read_raw_data(sockfd,&buffer[4],totalLen);
   xdrmem_create(&xdrs,buffer,REGBUFSIZE,XDR_DECODE);
   xdr_string(&xdrs,&variableName,strLen);    
   
   xdr_destroy(&xdrs);
   
   printf("after sl = %d read_cmnd.....strlen=%d command=%d ln=%d var=%s \n",
      totalLen, strLen,*command,*lineNumber,variableName);
   
}


void err(char *string) {
  perror(string);
  printf("Error no :\n",errno);
  exit(0);
}
















