#include <stdio.h>
#include "compilerApi.h"
#include <malloc.h>
#include <string.h>

#define TIMER_CLEAR    0
#define TIMER_START    1
#define TIMER_STOP     2
#define TIMER_ELAPSED  3
#define REGISTER       4
#define HPF_SERVER     5

#define SLOCAL       1
#define DISTRIBUTED 2

#define BEFORE  0
#define AFTER   1
#define MAX_FUNC_NUM 300
#define MAX_PAIR     300

enum FLAG {off, on};
FLAG in_main_program;
FLAG in_procedure;
int tabnumber=0;
void doFullType(SgType *t);
void doSymbAttribs(SgSymbol *s);
void doRoutineHeader(SgStatement *s);
void doSymb(SgSymbol *s);
void register_var();
void timer_start(),timer_clear(),timer_stop(),timer_elapsed();
void configure();
void insert_register_and_config_params(SgStatement *);
int give_type_symbol();
SgStatement *find_first_exec_stmt(SgStatement *);
void hpf_server();
void find_exec_next_stmt(SgStatement *,int *,int *);
int find_outermost_scope(SgStatement *);
void create_one_func(SgStatement *);
SgVariableSymb *DeclareTemporaryLike(SgSymbol * ,int );
void insert_decls(SgStatement *);
void insert_assign_stmt(SgAssignStmt *,SgStatement *,SgStatement *,int);
void build_function(int ,SgStatement *);
int isInLoop(SgStatement *);
//SgExpression * SgFuncHedrStmt::my_AddArg(char *, SgType &);


SgStatement *first;

typedef struct festmts
{
  char *fname;
  int  feline;
  SgStatement *s;
}FUNCS;
typedef struct Dist_ser_pair
{
  SgVariableSymb *symb;
  SgSymbol       *val;
  int            scope;/*line  number of control parent */
  int            emit ;/* If there are more than one assignment to one
 distributed variable in one parallel loop,we use this field */
  
}PAIRS;
int nextp=0;
PAIRS pairs[MAX_PAIR];


FUNCS funcs[MAX_FUNC_NUM];
int nextf = 0;


// Insert a C function call statement
void 
InsertFCallNode(int myline,SgStatement *s,int wfunction,int how)
{

  SgSymbol *symb ;
           
  switch (wfunction) {
    case TIMER_CLEAR: symb = new SgVariableSymb("timer_clear");
                        break;
    case TIMER_START: symb = new SgVariableSymb("timer_start");
                        break;
    case TIMER_STOP:  symb = new SgVariableSymb("timer_stop");
                        break;
    case TIMER_ELAPSED:symb = new SgVariableSymb("timer_elapsed"); 
                         break;
    case REGISTER : symb = new SgVariableSymb("REGISTER");
      break;
    case HPF_SERVER: symb = new SgVariableSymb("hpf_server");
      break;
    
 }

 
  SgCallStmt *func;
  SgValueExp *ev;
  SgStatement *lns;
  SgCallStmt *cs;
  int lexnextline,callnextline;
  
  func = new SgCallStmt(*symb);
  /* add myline parameter */  
  ev = new SgValueExp(myline);
  func->addArg(*ev);
    
  /* add scope of this statement */
  int hline = find_outermost_scope(s);
  ev = new SgValueExp(hline);
  func->addArg(*ev);

  switch (how) {
    case BEFORE:
         s->insertStmtBefore(*func,*(s->controlParent()));
         break;
    case AFTER:
         s->insertStmtAfter(*func,*(s->controlParent()));
         break;
  }
  
}

void action(SgStatement * s)
{
  SgDeclarationStatement *decl;
  SgProgHedrStmt *spg;
  SgCallStmt *cs;
  SgForStmt *fs;
  int flag = 0;
  SgAssignStmt *ass ;
  SgForStmt *sfo ; 
  
  printf("id:%4d   variant:%4d linenum:%4d\n", s->id(), s->variant(),
    s->lineNumber());
       
  if (cs=isSgCallStmt(s)){
      if ((!strcmp(cs->name()->identifier(),"register_var"))||
          (!strcmp(cs->name()->identifier(),"configure"))||
          (!strcmp(cs->name()->identifier(),"hpf_server")))
        return;
  }
  
  SgStatement *cp = s->controlParent();
  if ((isSgExecutableStatement(s))&&
      (s->variant() != CONTROL_END)&&
      (s->variant() != MODULE_STMT)&&
      ((!cp)||(!(cp->lineNumber()==s->lineNumber()))) )
            InsertFCallNode(s->lineNumber(),s , HPF_SERVER,BEFORE);

  // put server function call before end of function unit
    
}

void transform_registration(SgStatement * s) 
{
  int i=0;

  int org_numberOfChildren1=s->numberOfChildrenList1();
   int org_numberOfChildren2= s->numberOfChildrenList2();
  int org_numberOfChildren = org_numberOfChildren1 + org_numberOfChildren2;
   SgStatement * * children = new SgStatement * [org_numberOfChildren];
   SgStatement *st;
   SgStatement *line;
   
   for (i=0; i<org_numberOfChildren1; i++)
	children[i]=s->childList1(i);

  for (; i<org_numberOfChildren; i++)
     children[i]=s->childList2(i-org_numberOfChildren1);

   for (i=0; i<org_numberOfChildren; i++){
     st =  children[i];
     if ((st -> variant() == FORALL_CONSTRUCT_STMT)||
         (st -> variant() == INTERFACE_STMT)||
         (st -> variant() == FORALL_STMT))
     {
         action(st);
	 continue;
     }
     	 /* pass on INDEPENDENT DO LOOP */
     else if  (st -> variant() == INDEPENDENT_STMT)
     {
           action(st);
	   i++;
	   continue;
     }
     else{
         action(st);
         insert_register_and_config_params(st);
         transform_registration(st);
       }
     
   }
       
   delete [] children;
   return;
   
}

reset_emit()
{
  
  for (int i= 0;i<nextp;i++)
    pairs[i].emit = 0;
}

void insert_assign_stmt(SgAssignStmt *cass,SgStatement *after,
      SgStatement *cp,int flag)
{
    SgVariableSymb *tsymb;
    SgSymbol *symb;
    SgArrayRefExp *lhs,*rhs;
    SgAssignStmt *ass;
    SgArrayRefExp *tar;
    int i;

    SgExpression *st = cass->lhs();
    symb = st->symbol();


    if ((symb->attributes() & ALIGN_BIT)||
	(symb->attributes() & DISTRIBUTE_BIT )){

      printf("I  GOT %s as parallel stmtm line = %d after %d\n", 
           symb->identifier(),cass->lineNumber(),after->lineNumber());
      for(i=0;i<nextp;i++)
        if ((!strcmp(symb->identifier(),(pairs[i].val)->identifier()))&&
            (symb->scope()->lineNumber() == pairs[i].scope))
          break;
      if (i == nextp){
	printf("I COULD'NT find the name = %s\n",symb->identifier());
	return;
      }
      if (pairs[i].emit){
	printf("%s was emitted before \n",symb->identifier());
	return;
      }

      pairs[i].emit = 1;
      if (i < nextp)
	tsymb = pairs[i].symb;

      if (flag){
        lhs = new SgArrayRefExp(*(&tsymb->copy()));
        rhs = new SgArrayRefExp(*(&symb->copy()));
      }
      else{

        tar = ( SgArrayRefExp *)st;
        SgSymbol *ts = (SgSymbol *)tsymb; 
        if (tar->subscripts())
           lhs = new SgArrayRefExp(*(&ts->copy()),
               *(&tar->subscripts()->copy()));
        else
           lhs = new SgArrayRefExp(*(&ts->copy()));
        rhs = tar;
      }		       

      ass = new SgAssignStmt(*lhs,*rhs);
      after->insertStmtAfter(*ass,*cp);
    
    }

}


int find_outermost_scope(SgStatement *s)
{
   SgProgHedrStmt *    prog,*curs;    
   SgProcHedrStmt *    subr;  
   SgProsHedrStmt *    pros;
   SgFuncHedrStmt *    func;  
   SgStatement *cp;


   cp = s;
   curs = NULL;
   do{
      if(subr = isSgProcHedrStmt (cp))
        curs = (SgProgHedrStmt *)subr;
      else if (pros = isSgProsHedrStmt(cp))
        curs = (SgProgHedrStmt *)pros;
      else if(func = isSgFuncHedrStmt(cp))
        curs = (SgProgHedrStmt *)func;
      else if(prog = isSgProgHedrStmt(cp))
        curs = prog;
      if (curs == NULL)
	cp = cp->controlParent();
   }while(!curs);
   return (curs->lineNumber());
   
}

//find whether or not current stmt s is in one loop ,if so return 1 
// otherwise return 0
int isInLoop(SgStatement *s)
{

   SgStatement *cp,*curs;
   SgForStmt      *fstmt;
   SgDoWhileStmt  *dwstmt;

   cp   = s;
   curs = NULL;
   do{
     if ((fstmt = isSgForStmt(cp))||(dwstmt = isSgDoWhileStmt(cp)))
       curs =  cp;
     else if(cp != NULL)
   	cp = cp->controlParent();
   }while((cp)&&(!curs));
   if (curs )
     return 1;
   else 
     return 0;
         
}



void find_exec_next_stmt(SgStatement *s,int *lexnext,int *callnext)
{
  SgStatement *cp;
  SgCallStmt  *cs;
  int i;
  
  if (s->lexNext()->variant() != CONTROL_END){
    if (s->lineNumber() == s->lexNext()->lineNumber()){
       cp = s;
       while((cp->nextInChildList())&& 
          (cp->nextInChildList()->variant() == CONTROL_END))
       cp = cp->controlParent();
       if (cp->nextInChildList())
         *lexnext  = cp->nextInChildList()->lineNumber();
       else
         *lexnext  = -3;     
    }
    
    else
      *lexnext  = s->lexNext()->lineNumber();
  }
      
  else{
    cp = s;
    while((cp->nextInChildList())&& 
         (cp->nextInChildList()->variant() == CONTROL_END))
      cp = cp->controlParent();
    if (cp->nextInChildList())
      *lexnext  = cp->nextInChildList()->lineNumber();
    else
      *lexnext  = -2;
    
  }
  
  if (cs = isSgCallStmt(s)){
    for(i=0;i<nextf;i++)
      if(!strcmp(funcs[i].fname,cs->name()->identifier()))
        break;
    if (i == nextf) 
      *callnext = -1;
    else
      *callnext = funcs[i].feline;
  }
  else
    *callnext  = -2;
  
}


SgStatement *find_first_exec_stmt(SgStatement *s)
{
  int i;
  
   int nlchilds = s->numberOfChildrenList1();
   
   for(i=0;i<nlchilds;i++)
     if ((isSgExecutableStatement(s->childList1(i)))&&
        (s->childList1(i)->variant() != CONTROL_END))
	 return (s->childList1(i));
   
   return NULL;
   
}

int give_type_symbol(SgType *vtype)
{
  int variant_type = vtype->variant();
  int var_type;

  switch(variant_type){
           
     case T_CHAR: var_type = 0; break;
     case T_INT: var_type = 1; break;
     case T_FLOAT: var_type = 2; break;
     case T_DOUBLE : var_type = 4; break;
     case T_BOOL: var_type = 5; break;
     case T_ARRAY: var_type = give_type_symbol(vtype->baseType());break;
     case T_KIND:
	int base_type = 
            vtype->thetype->entry.kind_decl.base_type->variant;
        int kind_val = (vtype->thetype->entry.kind_decl.kindexp)->entry.                   list_s.item->entry.literal.value->entry.ival; 
	
        switch(base_type){
	case T_INT:
	  switch(kind_val){
	  case 1:var_type = 1;break; /* Byte */
	  case 2:var_type = 7;break; /* Short */
	  case 4:var_type = 1;break; /* Integer */
	  case 8:var_type = 3;break; /* Long */
	  }
	  	  
	  break;
	case T_FLOAT:  
	  switch(kind_val){
	  case 4:var_type = 2;break; /* Float */
	  case 8:var_type = 4;break; /* Double */
	  }
	  
	  break;
	}
		
	break;
  }

  return (var_type);
 
}


int symb_ndims(SgSymbol *symb)

{
  int vtype =  symb->type()->variant();
  SgArrayType *atype;
  int ndims;
    
  switch(vtype){
  case T_ARRAY: if (atype = isSgArrayType(symb->type()))
                   ndims = atype->dimension();
  break;
  
  default:ndims = 0;break;
  }
  return ndims;
  
}

  
void insert_register_and_config_params(SgStatement *s)
{

    SgProgHedrStmt *    prog,*curs;    
    SgProcHedrStmt *    subr;  
    SgProsHedrStmt *    pros;
    SgFuncHedrStmt *    func;     
    SgSymbol *symb;
    SgVariableSymb *fsymb,*shape_symb;
    SgStatement    *    festmt;
    int var_type;
    int ndims,scope,var_id;
    char *name,indexName[256] ;
    SgValueExp *sve;
    SgCallStmt *scstmt;
    SgVarRefExp *e1;
    SgVariableSymb *ptsymb;
    SgFunctionCallExp *fce;
    int distFlag;
    SgExpression *exp;
    SgExprListExp *elist;
    SgStatement *decls;
    int nextTemp=0;
    
        
    curs = NULL;
    if(subr = isSgProcHedrStmt (s))
       curs = (SgProgHedrStmt *)subr;
    else if (pros = isSgProsHedrStmt(s))
      curs = (SgProgHedrStmt *)pros;
    else if(func = isSgFuncHedrStmt(s))
      curs = (SgProgHedrStmt *)func;
    else if(prog = isSgProgHedrStmt(s))
      curs = prog;
    
    if (curs == NULL) return;
    
   	printf("symbol:%s", (curs->name()).identifier());
        printf(" which is a subroutine.\n");
        /* find first executable stmt in current scope */
        festmt = find_first_exec_stmt(curs);
        if (festmt == NULL)
          return;

        /* add the first executable stmt of this func./subr. to funcs
           record array */
        funcs[nextf].feline = festmt->lineNumber();
	funcs[nextf].fname   = (curs->name()).identifier();
        funcs[nextf++].s = curs;
       		
        // new addition
        // get the expression list of variables delcared
        for(decls=curs->lexNext();decls!=festmt;decls = decls->lexNext())
	{
           if (!isSgVarDeclStmt(decls))
             continue;

	   SgVarDeclStmt *svar = (SgVarDeclStmt *)decls;
           int ns = svar->numberOfSymbols();
           for(int i=0;i<ns;i++)
	   {
              if (svar->symbol(i)->variant() != VARIABLE_NAME) 
                continue;
	      symb = svar->symbol(i);
	          	   
              printf("I am registring : %s variant = %d\n",
                   symb->identifier(),symb->variant());

              /* Dont register variables other than variable names */
              if (symb->variant() != VARIABLE_NAME)
                continue;
                          
              fsymb = new SgVariableSymb("register_var");
	      scstmt = new SgCallStmt(*fsymb);
	      
              /* add var_type parameters to register_var subroutine */
	      var_type = give_type_symbol(symb->type());
              sve = new SgValueExp(var_type);
	      scstmt->addArg(*sve);

              /* add variable id = parameter var_id */	
              sprintf(indexName, "ind_%d", nextTemp++);
              ptsymb = new SgVariableSymb(indexName);
	      e1 = new SgVarRefExp(*ptsymb);
	      scstmt->addArg(*e1);
              // make declaration this new created temporary variable
	      // in this function/subroutine
              ptsymb->declareTheSymbol(*curs);
	      

              /* add variable itself = var parameter */
              e1 = new SgVarRefExp(*symb);
	      scstmt->addArg(*e1);
             
              /* add variable literal name = name parameter */
              char *ts = (char *)malloc(strlen(symb->identifier())+3);
	      strcpy(&ts[1],symb->identifier());
	      ts[0] = '\'';
	      ts[strlen(symb->identifier())+1] = '\'';
              ts[strlen(symb->identifier())+2] = '\0';
              	      
	      sve = new SgValueExp(ts);
	      scstmt->addArg(*sve);

              /* add var_ndims parameter */
	      ndims = symb_ndims(symb);
              sve = new SgValueExp(ndims);
	      scstmt->addArg(*sve);

              /*prepare shape(A) call fucntion for dim_info parameter*/
              /* Here we are trying to calculate the shape of var */
              shape_symb = new SgVariableSymb("shape");
	      fce   = new SgFunctionCallExp(*shape_symb );
	      sve = new SgValueExp(symb->identifier());
              fce -> addArg(*sve);
              /* Add dim_info parameter  */
              scstmt->addArg(*fce);

              /* add scope of this symbol */
             
              int scope_num = curs->lineNumber();
	      sve = new SgValueExp(scope_num);
	      scstmt->addArg(*sve);

              /* Add distributed flag:is distributed or not */      
              if ((symb->attributes() & ALIGN_BIT)||
                    (symb->attributes() & DISTRIBUTE_BIT ))
		  distFlag = 1;
	      else
		  distFlag = 0;
              SgValueExp *dflag = new SgValueExp(distFlag);       
              scstmt->addArg(*dflag);
	      
              /* finally add register call stmt */
              festmt->insertStmtBefore(*scstmt,*(festmt->controlParent()));

	   }
	}
	
        // put server function call before end of function unit
        SgStatement *lastStmt = curs->lastExecutable()->lexNext();
	InsertFCallNode(lastStmt->lineNumber(),lastStmt , HPF_SERVER,BEFORE);

	
}


SgVariableSymb *DeclareTemporaryLike(SgSymbol * x,int sd)
{
       SgStatement  *old_decl_stmt;
       SgVariableSymb *tmp;
       SgVarDeclStmt *decl_stmt;
       SgExpression *exp;
       char *name;
       SgExprListExp *elist;
       SgType *old_type;
       SgVarDeclStmt *decl;
       SgProgHedrStmt *ispg;
              
       // now find the declaration statement for x and get
       // the correct initializing expression.
       printf("BEG I am declaraing temp = %s\n",x->identifier());
       
       old_decl_stmt = x ->declaredInStmt();
       // get the expression list of variables delcared
       exp = old_decl_stmt->expr(0);
       // look for the reference to our symbol x.
       while(elist = isSgExprListExp(exp)){
       	 if((elist->value())->IsSymbolInExpression(*x) )
               break;       
            exp = elist->next();
            }    
       if(exp == NULL) return NULL;

       // now create a new expression list and 
       // and a new declaration statement
       elist = new SgExprListExp(exp->copy());
       if (decl = isSgVarDeclStmt(old_decl_stmt))
         old_type = decl->type();
       decl_stmt = new SgVarDeclStmt(*elist,*old_type);

       // create a new variable of the same class
       // generate the name as "_Ti" where i is counter.
       name = (char *)malloc(256);
       if (sd == SLOCAL)     
	 sprintf(name, "LOCAL_%s", x->identifier());
       else
	 sprintf(name, "DISTR_%s", x->identifier());
       tmp = new SgVariableSymb(name, (x->type())->copy(), *(x->scope()));
        
       //Test if new declaration can be added
       SgArrayRefExp *arefexp = new  SgArrayRefExp(*(&tmp->copy())); 
       //prepare shape(A) call fucntion for dim_info parameter
       //Here we are trying to calculate the shape of var 
       for(int i=1;i<=symb_ndims(x);i++){
            SgVariableSymb *shape_symb = new SgVariableSymb("size");
	    SgFunctionCallExp *fce   = new SgFunctionCallExp(*shape_symb );
	    SgValueExp *sve = new SgValueExp(x->identifier());
            fce -> addArg(*sve);
            sve = new SgValueExp(i);
            fce -> addArg(*sve); 
            arefexp->addSubscript(*fce);
	}
	SgVarDeclStmt *ndecl = new SgVarDeclStmt(*arefexp,
               *(x->type()->baseType()));
	
	old_decl_stmt->insertStmtBefore(*ndecl,
                *(old_decl_stmt->controlParent()));

        return tmp;

    }

void insert_decls(SgStatement *first)
{
   SgStatement *prev;
   SgSymbol *symb;
   SgVarDeclStmt *decl;
   SgProgHedrStmt *p;
         
   for(prev = first;prev;prev = prev->lexNext())
     if  ((prev->variant() == COMBINED_DIRECTIVE_DECL)||
               (prev->variant() == ALIGN_DECL)||
	  (prev->variant() == DISTRIBUTE_DECL)){
       
         decl = (SgVarDeclStmt *)prev;
       	 for(int i=0;i<decl->numberOfSymbols();i++){
            pairs[nextp].symb    = 
                     DeclareTemporaryLike(decl->symbol(i),SLOCAL);
            pairs[nextp].val     = decl->symbol(i);
            pairs[nextp++].scope = decl->controlParent()->lineNumber();
	 }
	 
     }
   
}


int main(int argc, char **argv)
{

  SgFile * file;
  SgProject * project = new SgProject("test", argc, argv);

  for (int i=0; i< project->numberOfFiles(); i++) {

    file = &(project->file(i));
            
    SgStatement * s = file->firstStatement();
    first = s;

    insert_decls(first);
    printf("AFTER INSERT_DECLS ...\n");
   
    transform_registration(s);
    printf("AFTER TRANSFORM_REGISTRATION..\n");
    
    // save a copy of the input file, if you like.
    file->saveDepFile("debug.dep");

  };

  return 0;
}







