/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <stdio.h>
#include "compilerApi.h"

// This program demonstrates the functions that access the
// contents of the parsed program stored in the .dep files.
// It is very simple.  It traverses the program file in lexical
// order and examines each statement and prints out the
// information stored in the file about the statement.

// THIS PROGRAM WORKS, BUT IT IS NOT FINISHED!!!

// each object in the .dep file has an integer "variant" to
// describe its type.  There is a standard defined name for
// each and that may be retrieved from the array "tag" below.
#define MAXTAGS 1000
static char *tag[MAXTAGS];
void initVariantNames(){
   for(int i = 0; i < MAXTAGS; i++) tag[i] = NULL;
}
void printVariantName(int i){
     if((i >= 0 && i < MAXTAGS) && tag[i]) printf("%s", tag[i]);
     else printf("not a known node variant");
}
  
// the unparse operation regenerates source code.
// The function below prints the entire project to 
// stdout.  note:  the current unparser will only print
// to stdout.  in the near future we will have a version
// that will print to a string buffer.

void
ProjectUnparse(SgProject *project)
{
  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;
    
    f = &(project->file(i));
    f->unparsestdout();
  }
}

// some forward declarations.

void doExpr(SgExpression *e);
void doSymb(SgSymbol *s);
void doType(SgType *t);
void doRoutineHeader(SgStatement *s);
void doIfStmts(SgStatement *s);
void doLoopStmts(SgStatement *s);
void doIOStmts(SgStatement *s);
void doGoToStmt(SgStatement *s);
void doVarDecl(SgStatement *s);
void doFullType(SgType *t);
void doSymbAttribs(SgSymbol *s);
void doPortDecl(SgStatement *s);  /* Fortran M */
void doChannel(SgStatement *s);  /* Fortran M */
void doMerger(SgStatement *s);  /* Fortran M */
void doMoveport(SgStatement *s);  /* Fortran M */
void doSend(SgStatement *s);  /* Fortran M */
void doReceive(SgStatement *s);  /* Fortran M */
void doEndchannel(SgStatement *s);  /* Fortran M */
void doProbe(SgStatement *s);  /* Fortran M */
void doPortTypeExpr(SgExpression *e);  /* Fortran M */
void doPort(SgExpression *e);  /* Fortran M */
void doStore(SgExpression *e);  /* Fortran M */
void doLabel(SgExpression *e);  /* Fortran M */
void doDataImplDo(SgDataImpliedDoExp *e); /*Fortran M */
void doDataElt(SgExpression *e); /* Fortran M */

// this function traverses the list of statements in a file.
// for each statement it prints out all the information available
// in the .dep data base.

void classifyStatements(SgFile *f){
  SgStatement *line;
  SgStatement *s;
  int i,j;

  // grab the first statement in the file.
  s = f->firstStatement();

  // follow the statements in lexical order.
  // an alternative is to follow them in scope order
  // using a recursive traversal, but this is simpler
  // for what we want to do here.

  for (line = s; line; line = line->lexNext()){
   printf("---------------------------------------------------------------------\n");
         // print the line to the standard out.
         // note: if the statement controls other blocks
         // nested within, it will print the entire block
         // of "control children".
         line->unparsestdout();

         // print out the variant tag, i.e. the type of statement.
         printf("STMT[%d] at line %d has variant:", 
                                      line->id(), line->lineNumber());
         printVariantName(line->variant());
         printf("\n");

         // is this a labeled statement? if so print the label.
         if(line->hasLabel()){
		SgLabel *lab = line->label();
                printf("has label[%d]\n", lab->id());
		}

         // are there comments that precede this statement?
         // if so print them out.
	 if(line->comments())
		printf("comments befor this line:\n%s", line->comments());
	
         // for each type of statement call the appropriate special case
         // function.	  
         switch(line->variant()){
               case GLOBAL:
                     printf("The head-of-fole global node. not really a statement\n");
                     break;
               case PROG_HEDR:
               case PROC_HEDR:
               case PROS_HEDR:
               case FUNC_HEDR:
                    printf(" a new routine.\n");
                    doRoutineHeader(line);
                    break;
               case BASIC_BLOCK	:
                    printf(" a beginning of a control block for C programs \n");
                    break;
               case CONTROL_END:
                    printf(" end of a control block for stmt %d\n",
                             (line->controlParent())->id());
                    break;
               case IF_NODE:
               case ELSEIF_NODE:
               case ARITHIF_NODE:
               case LOGIF_NODE:
                     doIfStmts(line);
                     break;
               case LOOP_NODE:
               case FOR_NODE:
               case FORALL_NODE:
               case WHILE_NODE:
               case PROCESS_DO_STAT:
                     doLoopStmts(line);
                     break;
               case EXIT_NODE:
                    // note the use of the special cast function
                    // we are casting to a subtype but this is o.k.
                    // with this function.
		    {SgExitStmt *   exst =  isSgExitStmt(line);
		     printf("an exit statement. exit from:\n");
		     doSymb(exst->constructName());
                     }
		     break;
               case ASSIGN_STAT:
		     { SgAssignStmt *ass =  isSgAssignStmt(line);
		     printf("an assignment\n");
                     // print out the two expressions on each side of =
		     printf("left hand side:\n");
			doExpr(ass->lhs());
		     printf("\nright hand side:\n");
			doExpr(ass->rhs());
                     printf("\n");
		     }
		     break;
               case PROS_STAT: /* Fortran M */
                    { int i;
                      SgProsCallStmt *call = isSgProsCallStmt(line);
                     // a "process call f(...) statement.  extract the passed
                     // arguement expressions and print them.
                     printf("a process call. passed args are\n");
                     for(i = 0; i < call->numberOfArgs(); i++){
                        printf("argument %d:\n", i);
                        doExpr(call->arg(i));
                        printf("\n");
                        }
                      }
                     break;
               case PROS_STAT_LCTN: /* Fortran M */
                    { int i;
                      SgProcessorsRefExp *pr;
                      SgProsCallLctn *call = isSgProsCallLctn(line);
                     // a "process call f(...) statement.  extract the passed
                     // arguement expressions and print them.
                     printf("a process call. passed args are\n");
                     for(i = 0; i < call->numberOfArgs(); i++){
                        printf("argument %d:\n", i);
                        doExpr(call->arg(i));
                        printf("\n");
                        }
                     printf("executed on processor location:\n");
                     pr = (SgProcessorsRefExp *) call->location()->lhs();
                     for(i = 0 ; i < pr->numberOfSubscripts() ; i++){
                        printf("dimension %d:\n", i+1);
                        doExpr(pr->subscript(i));
                        printf("\n");
                        }
                     }
                     break;
               case PROS_STAT_SUBM: /* Fortran M */
                    { int i;
                      SgProcessorsRefExp *pr;
                      SgProsCallSubm *call = isSgProsCallSubm(line);
                     // a "process call f(...) statement.  extract the passed
                     // arguement expressions and print them.
                     printf("a process call. passed args are\n");
                     for(i = 0; i < call->numberOfArgs(); i++){
                        printf("argument %d:\n", i);
                        doExpr(call->arg(i));
                        printf("\n");
                        }
                     printf("executed on sub-machine:\n");
                     pr = (SgProcessorsRefExp *) call->submachine()->lhs();
                     for(i = 0 ; i < pr->numberOfSubscripts() ; i++){
                        printf("dimension %d:\n", i+1);
                        doExpr(pr->subscript(i));
                        printf("\n");
                        }
                     }
                     break;
               case PROC_STAT:
		    { int i;
                      SgCallStmt *call = isSgCallStmt(line);
                     // a "call f(...) statement.  extract the passed
                     // arguement expressions and print them.
                     printf("a subroutine call. passed args are\n");
		     for(i = 0; i < call->numberOfArgs(); i++){
			printf("argument %d:\n", i);
			doExpr(call->arg(i));
		        printf("\n");
			}
                      }
                     break;
               case WHERE_NODE:
               case ALLDO_NODE:
               case IDENTIFY:
               case FORMAT_STAT:
               case STOP_STAT:      break;
               case RETURN_STAT:    break;
               case GOTO_NODE:
               case ASSGOTO_NODE:
               case COMGOTO_NODE:
			doGoToStmt(line);
                        break;
               case VAR_DECL:
               case INTENT_STMT:        
               case OPTIONAL_STMT:
               case PUBLIC_STMT:
               case PRIVATE_STMT:
               case DIM_STAT:
               case PROCESSORS_STAT:  /* Fortran M */
               case ALLOCATABLE_STMT:
               case POINTER_STMT:
               case TARGET_STMT:
               case MODULE_PROC_STMT:
               case EXTERN_STAT:
               case COMM_STAT:
               case PROS_COMM:	/* Fortran M */
               case NAMELIST_STAT:
               case EQUI_STAT:
               case PARAM_DECL:
               case IMPL_DECL:
               case SAVE_DECL:
               case DATA_DECL:
			doVarDecl(line);
                        break;
               case PAUSE_NODE:     break;
               case STOP_NODE:      break;
               case ASSLAB_STAT:    break;
               case COMMENT_STAT:    break;
               case CONT_STAT:              
                    printf(" continue statement: end of cntl for stmt %d\n",
                             (line->controlParent())->id());
		    break;
               case ENTRY_STAT:
               case STMTFN_STAT:
               case BLOCK_DATA:
               case INTRIN_STAT:
               case ATTR_DECL:	/* attribute declaration */
                     printf("another fortran statement\n");
		     break;
                // the statements below are all valid, and will be
                // treated in example when I have more time.
               case READ_STAT:
               case WRITE_STAT:
               case PRINT_STAT:
               case BACKSPACE_STAT:
               case REWIND_STAT	:
               case ENDFILE_STAT:
               case INQUIRE_STAT:
               case OPEN_STAT:
               case CLOSE_STAT:
               case OTHERIO_STAT:
			doIOStmts(line);
                        break;
               case INCLUDE_STAT:
               case ALLOCATE_STMT:
               case NULLIFY_STMT:
               case DEALLOCATE_STMT:
               case SEQUENCE_STMT:
               case CYCLE_STMT:
               case EXIT_STMT:
               case CONTAINS_STMT:
               case WHERE_BLOCK_STMT:
               case MODULE_STMT:
               case USE_STMT:
               case INTERFACE_STMT:
               case OVERLOADED_ASSIGN_STAT:
               case POINTER_ASSIGN_STAT:
               case OVERLOADED_PROC_STAT:
                        printf("yet another fortran 90 statement.\n");
                        break;
               case PROCESSES_STAT:  /* Fortran M */
               case PROCESSES_END:  /* Fortran M */
                        printf("another fortran M statement.\n");
                        break;
               case DECOMPOSITION_STMT:  /* Old Fortran D stuff */
               case ALIGN_STMT:
               case DISTRIBUTE_STMT:
               case REDUCE_STMT:
                         printf("Old Fortran D. not treated here..needs update\n");
                         break;
               case CDOALL_NODE:  /* added for Cedar Fortran */
               case SDOALL_NODE:
               case DOACROSS_NODE:
               case CDOACROSS_NODE:
                        printf("Cedar Fortran.  who cares anymore. \n");
                        break;
               case PARDO_NODE:     /* Following added for PCF Fortran */
               case PARSECTIONS_NODE:
               case SECTION_NODE:
               case GUARDS_NODE:
               case LOCK_NODE:
               case UNLOCK_NODE:
               case CRITSECTION_NODE:
               case POST_NODE:
               case WAIT_NODE:
               case CLEAR_NODE:
               case POSTSEQ_NODE:
               case WAITSEQ_NODE:
               case SETSEQ_NODE:
               case ASSIGN_NODE:
               case RELEASE_NODE:
               case PRIVATE_NODE:
               case SCOMMON_NODE:
               case PARREGION_NODE:
               case PDO_NODE:
               case PSECTIONS_NODE:
               case SINGLEPROCESS_NODE:
               case SKIPPASTEOF_NODE:
                   printf("Old pcf fortran. needs update.  who cares anymore. \n");
		   break;
               case INPORT_DECL:  /* Fortran M */
               case OUTPORT_DECL:  /* Fortran M */
                   doPortDecl(line);
                   break;
               case CHANNEL_STAT:  /* Fortran M */
                   doChannel(line);
                   break;
               case MERGER_STAT:  /* Fortran M */
                   doMerger(line);
                   break;
               case MOVE_PORT:  /* Fortran M */
                   doMoveport(line);
                   break;
               case SEND_STAT:  /* Fortran M */
                   doSend(line);
                   break;
               case RECEIVE_STAT:  /* Fortran M */
                   doReceive(line);
                   break;
               case ENDCHANNEL_STAT:  /* Fortran M */
                   doEndchannel(line);
                   break;
               case PROBE_STAT:  /* Fortran M */
                   doProbe(line);
                   break;
               default:
                   printf("unknown statement. not treated here..I am tired.\n");
                   break;
         }
   printf("---------------------------------------------------------------------\n");
    }
}

// print out the name and formal parameters for this subroutine,
// function or main program header.

void doRoutineHeader(SgStatement *s){
SgProgHedrStmt *    prog;    
SgProcHedrStmt *    subr;  
SgProsHedrStmt *    pros;
SgFuncHedrStmt *    func;     
int i;
SgSymbol *sy;

   if(subr = isSgProcHedrStmt (s)){
	printf("symbol:%s", (subr->name()).identifier());
        printf(" which is a suroutine.\n");
        printf("parameter list:\n");
	for(i = 0; i < subr->numberOfParameters(); i++){
		sy = subr->parameter(i);
		printf("parameter(%d):", i); doSymb(sy);
                printf(" of type: "); doFullType(sy->type());
		doSymbAttribs(sy); 
		printf("\n"); 
		}
	}
   else if(pros = isSgProsHedrStmt (s)){
        printf("symbol:%s", (pros->name()).identifier());
        printf(" which is a process.\n");
        printf("parameter list:\n");
        for(i = 0; i < pros->numberOfParameters(); i++){
                sy = pros->parameter(i);
                printf("parameter(%d):", i); doSymb(sy);
                printf(" of type: "); doFullType(sy->type());
                doSymbAttribs(sy);
                printf("\n");
                }
        }
   else if(func = isSgFuncHedrStmt (s)){
	printf("symbol:%s", (func->name()).identifier());
        printf("which is a function of type ");
	doFullType((func->name()).type());
        printf("\n");
        printf("parameter list:\n");
	for(i = 0; i < func->numberOfParameters(); i++){
		sy = func->parameter(i);
		printf("parameter(%d):", i); doSymb(sy); 
                printf(" of type: "); doFullType(sy->type()); 
		doSymbAttribs(sy); 
		printf("\n"); 
		}
	}
}

// this routine handles the different types of "if" statements.
// not finished.
void doIfStmts(SgStatement *s){
SgLogIfStmt *    logif;   
SgIfStmt *       ifst;        
SgArithIfStmt *  arithif;  

   if (logif = isSgLogIfStmt (s)){
	printf("logical if: condition EXPR[%d] ", logif->conditional()->id());
	(logif->conditional())->unparsestdout();
	printf(" :if true do STMT[%d]\n", (logif->body())->id());
	}
   else if(ifst = isSgIfStmt (s)){}
   else if(arithif = isSgArithIfStmt (s)){}
}

// this routine handles the different types of loop statements.
// only does do loops now.  more later.

void doLoopStmts(SgStatement *s){
SgForStmt *  doloop;
SgProcessDoStmt *prosdo;    
SgWhileStmt *  whileloop;          
SgExpression *e;
   if (doloop = isSgForStmt(s)){
	printf("a DO LOOP: control var:");
	doSymb(doloop->symbol());
	printf("\n lower bound = ");
	if(e = doloop->start()) e->unparsestdout();
	printf("  , upper bound = ");
	if(e = doloop->end()) e->unparsestdout();
	if(e = doloop->step()){
		printf(" , step = ");
	 	e->unparsestdout();
		}
	printf(" \n");
	if(doloop->endOfLoop()){
		printf("end of loop label[%d]\n",(doloop->endOfLoop())->id());
		}
	}
   else if (prosdo = isSgProcessDoStmt(s)){  /* Fortran M */
        printf("a process do. control var:");
        doSymb(prosdo->symbol());
        printf("\n lower bound = ");
        if(e = prosdo->start()) e->unparsestdout();
        printf("  , upper bound = ");
        if(e = prosdo->end()) e->unparsestdout();
        if(e = prosdo->step()){
                printf(" , step = ");
                e->unparsestdout();
                }
        printf(" \n");
        if(prosdo->endOfLoop()){
                printf("end of loop label[%d]\n",(prosdo->endOfLoop())->id());
                }
        }
   else if(whileloop = isSgWhileStmt (s)){}
   else printf("not finished\n");
}

// print out info about I/O statements.
// not done.
         
void doIOStmts(SgStatement *s){
SgIOStmt *   iost; 
SgInputOutputStmt * inoust;   
SgIOControlStmt *  iocntl;  
   if (iost = isSgIOStmt(s)){}
   else if(inoust = isSgInputOutputStmt (s)){}
   else if(iocntl = isSgIOControlStmt(s)){};
}

// print out the information about the different types of
// goto statements.  
// not done.  only does standard goto's.

void doGoToStmt(SgStatement *s){
SgGotoStmt *   gotostmt;  
SgAssignedGotoStmt * assgoto;  
SgComputedGotoStmt * compgoto; 
  if(gotostmt = isSgGotoStmt(s)){
	printf("garden variety goto: branch to stmt with label [ %d ]\n",
		gotostmt->branchLabel()->id());

	}
  else if(assgoto = isSgAssignedGotoStmt(s)){}
  else if(compgoto = isSgComputedGotoStmt(s)){}
}

// this does variable declarations.
// there are lots of different forms that a
// declaration can take.  this just does the standard
// types now.  not done yet.

void doVarDecl(SgStatement *s){
SgVarDeclStmt * var;        
SgIntentStmt * intent;
SgVarListDeclStmt *  varlist;   
SgStructureDeclStmt * strucdecl;   
SgNestedVarListDeclStmt*  declst; 
SgParameterStmt *    parm;    
SgImplicitStmt *     implic;   
int i;
   if (var = isSgVarDeclStmt(s)){
	 int i;
         printf("Variables declared here are:\n");
	 for(i = 0; i < var->numberOfSymbols(); i++){
		SgSymbol *s;
                s =  var->symbol(i);
                doSymb(s); printf(" of type: ");
		if(s->type()) doFullType(s->type());
		printf("\n");
		}
         } 
   else if (intent = isSgIntentStmt (s)){
         int i;
         printf("Variable with attribute ");
         doExpr(intent->attribute());
         printf(" are: \n");
         for(i = 0; i < intent->numberOfArgs(); i++){
           printf("Variable %d:\n", i);
           doExpr(intent->arg(i));
           printf("\n");
           }
         }
   else if (varlist = isSgVarListDeclStmt (s)){}
   /* else if (strucdecl = isSgStructureDeclStmt (s)){} */
   else if ( declst = isSgNestedVarListDeclStmt (s)){}
   /* else if ( parm = isSgParameterStmt (s)){} */
   else if ( parm = isSgParameterStmt (s)){
     int i;
     for (i=0; i<parm.numberOfConstants(); ++i) {
       printf("PARAMETER %d VALUE: ", i);
       doExpr(parm.value(i));
       printf("\n");
     } 
   }
   else if ( implic =  isSgImplicitStmt (s))
     for (i = 0; i< implic->numberOfImplicitTypes(); i++) {
       SgType *impl_t = implic->implicitType(i);
 
       printf("type(%d) = %d ", i, impl_t->variant());
       printVariantName(impl_t->variant());
       printf("\n");
   }
   else printf("not here!\n");
}


void doSymb(SgSymbol *s){
    printf("SYMB[%d]:%s",s->id(),s->identifier());
}

void doSymbAttribs(SgSymbol *s){
	int i;
	if(s->attributes()){
	    i = s->attributes();
	    printf(" attributes =( ");
              if(i & ALLOCATABLE_BIT ){ printf(" allocatable,"); }
              if(i & DIMENSION_BIT  ){ printf(" dim,"); }
              if(i & EXTERNAL_BIT ){ printf(" extern,"); }
              if(i & IN_BIT  ){ printf(" in,"); }
              if(i & INOUT_BIT  ){ printf(" inout,"); }
              if(i & INTRINSIC_BIT  ){ printf(" intrinsic,"); }
              if(i & OPTIONAL_BIT  ){ printf(" optional,"); }
              if(i & OUT_BIT  ){ printf(" out,"); }
              if(i & PARAMETER_BIT  ){ printf(" parameter,"); }
              if(i & POINTER_BIT ){ printf(" pointer,"); }
              if(i & PRIVATE_BIT ){ printf(" private,"); }
              if(i & PUBLIC_BIT  ){ printf(" public,"); }
              if(i & SAVE_BIT  ){ printf(" save,"); }
              if(i & SEQUENCE_BIT  ){ printf(" seq,"); }
              if(i & RECURSIVE_BIT  ){ printf(" recur,"); }
              if(i & TARGET_BIT  ){ printf(" target,"); }
	    printf(")");
	}
}
		
void doPortTypeExpr(SgExpression *e)  /* Fortran M */
{
  SgPortTypeExp *exp = isSgPortTypeExp(e);
  SgType *t = exp->type();
  int length = exp->numberOfRef();
  SgPortTypeExp *ref;
  int i;

  if (exp) {
    if (t)
      if (exp->ref()) {
        doFullType(t);
        if (t->length()) {
          printf(" Length: ");
          doExpr(t->length());
        }
        doExpr(exp->ref());
        return;
      } else {
        doFullType(t);
        if (t->length()) {
          printf(" Length: ");
          doExpr(t->length());
        }
        return;
      }
    else {
      switch(NODE_CODE(exp->thellnd))
      {
        case INPORT_TYPE_OP:
          printf("recursive port type declaration: inport");
          break;
        case OUTPORT_TYPE_OP:
          printf("recursive port type declaration: outport");
          break;
        default:
          printf("strange stuff: %d\n", NODE_CODE(exp->thellnd));
      }
      printf("\n         ");
      ref = (SgPortTypeExp *) exp->ref();
      for (i=0; i<length; i++) {
        printf("type %d:  ", i+1);
        doPortTypeExpr(ref);
        printf("\n         ");
        ref = ref->next();
      }
      return;
    }
  }
}

void doPortDecl(SgStatement *s)  /* Fortran M */
{ 
  SgPortTypeExp *porttypes;
  int length;
  int i;

  switch(BIF_CODE(s->thebif))
    {
    case INPORT_DECL:
      SgInportStmt *inport = isSgInportStmt(s);
      porttypes = inport->porttypes();
      length = inport->numberOfPortTypes();
      printf("an inport declaration.\n");
      printf("data types are: \n");
      for(i = 0; i < length; i++) {
        printf("type %d:  ", i+1);
        doPortTypeExpr(porttypes);
        porttypes = porttypes->next();
        printf("\n");
        };
      printf("declared names are:\n");
      for(i = 0; i < inport->numberOfNames(); i++){
        printf("name %d:  ", i+1);
        doExpr(inport->name(i));
        printf("\n");
        }
      break;
    case OUTPORT_DECL:
      SgOutportStmt *outport = isSgOutportStmt(s);
      porttypes = outport->porttypes();
      length = outport->numberOfPortTypes();
      printf("an outport declaration.\n");
      printf("data types are:\n");
      for(i = 0; i < length; i++) {
        printf("type %d:  ", i+1);
        doPortTypeExpr(porttypes);
        porttypes = porttypes->next();
        printf("\n");
        };
      printf("declared names are:\n");
      for(i = 0; i < outport->numberOfNames(); i++){
        printf("name %d:  ", i+1);
        doExpr(outport->name(i));
        printf("\n");
        }
      break;
    }
}

void doDataElt(SgExpression *e)
{
  SgDataEltExp *d = isSgDataEltExp(e);
  SgSymbol *name = d->name();
  SgDataSubsExp *subs;
  SgDataRangeExp *range;
  
  if (name == NULL)
    doDataImplDo(isSgDataImpliedDoExp(d->dataimplieddo()));
  else {
    subs = isSgDataSubsExp(d->datasubs());
    range = isSgDataRangeExp(d->datarange());
    printf("      name: %s\n", name->identifier());
    if (subs != NULL) {
      printf("        subscripts: ");
      doExpr(subs->iconexprlist());
      printf("\n");
    }
    if (range != NULL) {
      printf("        range: ");
      doExpr(range->iconexpr1());
      printf("  ");
      doExpr(range->iconexpr2());
      printf("\n");
    }
  }
}

void doDataImplDo(SgDataImpliedDoExp *e)
{
  SgSymbol *name = e->iname();
  int i, length = e->numberOfDataelt();

  printf("\n  DataImpliedDoList:");
  printf("\n    DataList:\n");
  for (i=0; i<length; i++)
    doDataElt(e->dataelt(i));
  printf("      index: %s\n", name->identifier());
  printf("        init:");
  if (e->init() != NULL)
    doExpr(isSgIconExprExp(e->init())->expr());
  printf("\n");
  printf("        limit:");
  if (e->limit() != NULL)
    doExpr(isSgIconExprExp(e->limit())->expr());
  printf("\n");
  printf("        increment:");
  if (e->increment() != NULL)
    doExpr(isSgIconExprExp(e->increment())->expr());
  else
    printf("1");
  printf("\n");
}

void doPort(SgExpression *e)  /* Fortran M */
{
  SgDataImpliedDoExp *dataimpldo;

  switch(NODE_CODE(e->thellnd))
    {
    case INPORT_NAME:
      SgInportExp *in = isSgInportExp(e);
      printf("inport: ");
      doExpr(in->exp());
      break;
    case OUTPORT_NAME:
      SgOutportExp *out = isSgOutportExp(e);
      printf("outport: ");
      if (dataimpldo = isSgDataImpliedDoExp(out->exp()))
        doDataImplDo(dataimpldo);
      else
        doExpr(out->exp());
      break;
    case FROMPORT_NAME:
      SgFromportExp *from = isSgFromportExp(e);
      printf("fromport: ");
      doExpr(from->exp());
      break;
    case TOPORT_NAME:
      SgToportExp *to = isSgToportExp(e);
      printf("toport: ");
      doExpr(to->exp());
      break;
    default:
      printf("strange stuff: %d\n", NODE_CODE(e->thellnd));
    }
}

void doStore(SgExpression *e)  /* Fortran M */
{
  switch(NODE_CODE(e->thellnd))
    {
    case IOSTAT_STORE:
      SgIO_statStoreExp *io = isSgIO_statStoreExp(e);
      printf("io_state_store: ");
      doExpr(io->exp());
      break;
    case EMPTY_STORE:
      SgEmptyStoreExp *empty = isSgEmptyStoreExp(e);
      printf("empty_store: ");
      doExpr(empty->exp());
      break;
    default:
      printf("strange stuff: %d\n", NODE_CODE(e->thellnd));
    }
}

void doLabel(SgExpression *e)  /* Fortran M */
{
  switch(NODE_CODE(e->thellnd))
    {
    case ERR_LABEL:
      SgErrLabelExp *err = isSgErrLabelExp(e);
      printf("err_label: ");
      doExpr(err->exp());
      break;
    case END_LABEL:
      SgEndLabelExp *end = isSgEndLabelExp(e);
      printf("end_label: ");
      doExpr(end->exp());
      break;
    default:
      printf("strange stuff: %d\n", NODE_CODE(e->thellnd));
    }
}

void doChannel(SgStatement *s)  /* Fortran M */
{
  SgChannelStmt *channel = isSgChannelStmt(s);
 
  if (channel) {
    printf("a channel statement.\n");
    printf("the control list:\n");
    doPort(channel->outport());
    printf("\n");
    doPort(channel->inport());
    printf("\n");
    if (channel->ioStore()) {
      doStore(channel->ioStore());
      printf("\n");
    }
    if (channel->errLabel()) {
      doLabel(channel->errLabel());
      printf("\n");
    }
  }
}

void doMerger(SgStatement *s)  /* Fortran M */
{
  int i, length;
  SgMergerStmt *merger = isSgMergerStmt(s);
 
  if (merger) {
    printf("a merger statement.\n");
    printf("the control list:\n");
    length = merger->numberOfOutports();
    for (i=0; i<length; i++) {
      doPort(merger->outport(i));
      printf("\n");
    }
    doPort(merger->inport());
    printf("\n");
    if (merger->ioStore()) {
      doStore(merger->ioStore());
      printf("\n");
    }
    if (merger->errLabel()) {
      doLabel(merger->errLabel());
      printf("\n");
    }
  }
}

void doMoveport(SgStatement *s)  /* Fortran M */
{
  SgMoveportStmt *moveport = isSgMoveportStmt(s);
 
  if (moveport) {
    printf("a moveport statement.\n");
    printf("the control list:\n");
    doPort(moveport->fromport());
    printf("\n");
    doPort(moveport->toport());
    printf("\n");
    if (moveport->ioStore()) {
      doStore(moveport->ioStore());
      printf("\n");
    }
    if (moveport->errLabel()) {
      doLabel(moveport->errLabel());
      printf("\n");
    }
  }
}

void doSend(SgStatement *s)  /* Fortran M */
{
  int i, length;
  SgSendStmt *send = isSgSendStmt(s);

  if (send) {
    printf("a send statement.\n");
    length = send->numberOfOutports();
    printf("the control list:\n");
    for(i = 0; i < length; i++) {
      doPort(send->outport(i));
      printf("\n");
    }
    if (send->ioStore()) {
      doStore(send->ioStore());
      printf("\n");
    }
    if (send->errLabel()) {
      doLabel(send->errLabel());
      printf("\n");
    }
    printf("the arguments:\n");
    for(i = 0; i < send->numberOfArguments(); i++){
        printf("argument %d:  ", i+1);
        doExpr(send->argument(i));
        printf(" Type: ");
        printVariantName(((send->argument(i))->type())->variant());
        printf("\n");
        }
    printf("\n");
  }
}

void doReceive(SgStatement *s)  /* Fortran M */
{
  int i, length;
  SgReceiveStmt *receive = isSgReceiveStmt(s);
 
  if (receive) {
    printf("a receive statement.\n");
    length = receive->numberOfInports();
    printf("the control list:\n");
    for(i = 0; i < length; i++) {
      doPort(receive->inport(i));
      printf("\n");
    }
    if (receive->ioStore()) {
      doStore(receive->ioStore());
      printf("\n");
    }
    if (receive->errLabel()) {
      doLabel(receive->errLabel());
      printf("\n");
    }
    if (receive->endLabel()) {
      doLabel(receive->endLabel());
      printf("\n");
    }
    printf("the arguments:\n");
    for(i = 0; i < receive->numberOfArguments(); i++){
        printf("argument %d:  ", i+1);
        doExpr(receive->argument(i));
        printf(" Type: ");
        printVariantName(((receive->argument(i))->type())->variant());
        printf("\n");
        }
    printf("\n");
  }
}

void doEndchannel(SgStatement *s)  /* Fortran M */
{
  int i, length;
  SgEndchannelStmt *endchannel = isSgEndchannelStmt(s);
 
  if (endchannel) {
    printf("a endchannel statement.\n");
    length = endchannel->numberOfOutports();
    printf("the control list:\n");
    for(i = 0; i < length; i++) {
      doPort(endchannel->outport(i));
      printf("\n");
    }
    if (endchannel->ioStore()) {
      doStore(endchannel->ioStore());
      printf("\n");
    }
    if (endchannel->errLabel()) {
      doLabel(endchannel->errLabel());
      printf("\n");
    }
  }
}

void doProbe(SgStatement *s)  /* Fortran M */
{
  int i, length;
  SgProbeStmt *probe = isSgProbeStmt(s);
 
  if (probe) {
    printf("a probe statement.\n");
    length = probe->numberOfInports();
    printf("the control list:\n");
    for(i = 0; i < length; i++) {
      doPort(probe->inport(i));
      printf("\n");
    }
    if (probe->ioStore()) {
      doStore(probe->ioStore());
      printf("\n");
    }
    if (probe->errLabel()) {
      doLabel(probe->errLabel());
      printf("\n");
    }
    if (probe->emptyStore()) {
      doStore(probe->emptyStore());
      printf("\n");
    }
  }
}



// given a type, traverse it and generate full infomration
void doFullType(SgType *t){
   SgArrayType *arrayt;
   SgPointerType *ptr;
   SgReferenceType *ref; // only for C++
   SgDerivedType *deriv;
   SgDescriptType *descr; // only for C and C++
   SgDerivedCollectionType *col; // only for pC++
   SgType *tmp;
   int i,n;

   if(arrayt = isSgArrayType(t)){
        printf("dimension("); 
	n = arrayt->dimension();
	for(i = 0; i < n; i++){
		(arrayt->sizeInDim(i))->unparsestdout();
		if(i < n-1) printf(", ");
		}
	printf(") ");
	}
   else{
	switch(t->variant()){
           case T_INT:     printf("integer "); break;
           case T_FLOAT:   printf("real "); break;
           case T_DOUBLE:  printf("double precision "); break;
           case T_CHAR:    printf("character "); break;
           case T_BOOL:    printf("boolean "); break;
           case T_STRING:  printf("string "); break;
           case T_RECORD:  printf("record.. more later "); break;
           case T_UNKNOWN: printf("unknown "); break;
           case T_COMPLEX: printf("complex "); break;
           case LOCAL:
           case INPUT: 
           case OUTPUT:
           case IO: 
               printf("do these later\n");
 	       break;
	   }
   }
   if(t->hasBaseType()){
        tmp = t->baseType();
        if (tmp != NULL) {
	  printf("of "); 
	  doFullType(tmp);
          }
	}
} 

// help functions for classifyExpression
    
void doExpr(SgExpression *e){
    printf("EXPR[%d]:",e->id());
    e->unparsestdout();
}

int lhsId(SgExpression *e){
    if(e->lhs()) return (e->lhs())->id();
    else return -1;
}
int rhsId(SgExpression *e){
    if(e->rhs()) return (e->rhs())->id();
    else return -1;
}
int symbId(SgExpression *e){
    if(e->symbol()) return (e->symbol())->id();
    else return -1;
}
char * symbName(SgExpression *e){
    if(e->symbol()) return (e->symbol())->identifier();
    else return "";
}
int typeId(SgExpression *e){
    if(e->type()) return (e->type())->id();
    else return -1;
}


void classifyExpressions(SgFile *f){
    SgExpression *e;
    for(e = f->firstExpression(); e; e = e->nextInExprTable()){
       SgValueExp *valexp = isSgValueExp(e);
       printf("EXPR[%3d] ", e->id());
       switch(e->variant()){    
               case INT_VAL:
                    printf("integer value:%d ", valexp->intValue());
		    break;
               case FLOAT_VAL:
                    printf("real value:%f ", valexp->floatValue());
		    break;
               case DOUBLE_VAL:
                    printf("double value:%f ", valexp->doubleValue());
		    break;
               case BOOL_VAL:
                    printf("bool value:%i ", valexp->intValue());
		    break;
               case CHAR_VAL:
                    printf("char value:%c ", valexp->charValue());
		    break;
               case STRING_VAL:
                    printf("string value:%s ", valexp->stringValue());
		    break;
               case COMPLEX_VAL:
                     printf("complex: real part id=%d im part id=%d ",
                                       lhsId(e),rhsId(e));
		     break;
               case KEYWORD_VAL:   /* Strings to be printed with quotes */
		     break;
               case CONST_REF:
		     printf("constant ref: symb[%d]:%s type[%d] ", 
                                 symbId(e), symbName(e), typeId(e));
                     break;
               case VAR_REF:
		     printf("variable ref: symb[%d]:%s type[%d] ", 
			          symbId(e), symbName(e), typeId(e));
		     break;
               case ARRAY_REF:
               case PROCESSORS_REF:  /* Fortran M */
		     printf("array ref: symb[%d]:%s type[%d] subscript list id (%d) ", 
			          symbId(e), symbName(e), typeId(e), lhsId(e));
		     break;
               case RECORD_REF:   
               case LABEL_REF:
			break;
               case VAR_LIST:
               case EXPR_LIST:
               case RANGE_LIST:
                       printf("this item [%d]  next list [%d] ", lhsId(e), rhsId(e));
			break;
               case CASE_CHOICE:
               case DEF_CHOICE:
               case VARIANT_CHOICE:
			break;
               case DDOT:
               case RANGE_OP:
               case UPPER_OP:
               case LOWER_OP:
               case EQ_OP:
               case LT_OP:
               case GT_OP:
               case NOTEQL_OP:
               case LTEQL_OP:
               case GTEQL_OP:
               case ADD_OP:
               case SUBT_OP:
               case OR_OP:
               case MULT_OP:
               case DIV_OP:
               case MOD_OP:
               case AND_OP:
               case EXP_OP:
               case ARRAY_MULT:
               case CONCAT_OP:	/* cancatenation of strings */
               case XOR_OP:	/* .XOR. in fortran */
               case EQV_OP:	/* .EQV. in fortran */
               case NEQV_OP:	/* .NEQV. in fortran */
                       printf("left oprnd[%d] right oprnd[%d] ", lhsId(e), rhsId(e));
                      break;
               case MINUS_OP:	/* unary operations */
               case NOT_OP:
                       printf("operand   [%d]                 ", lhsId(e));
		      break;
               case STAR_RANGE:	/* operations with no operands 360.. */
		      break;
               case PROC_CALL:
               case FUNC_CALL:
               case CONSTRUCTOR_REF:
               case ACCESS_REF:
               case CONS:
               case ACCESS:
               case IOACCESS:
               case CONTROL_LIST:
               case SEQ:
               case SPEC_PAIR:
               case COMM_LIST:
               case STMT_STR:
               case EQUI_LIST:
               case IMPL_TYPE:
               case STMTFN_DECL:
               case BIT_COMPLEMENT_OP:
               case EXPR_IF:
               case EXPR_IF_BODY:
               case FUNCTION_REF:
               case UNARY_ADD_OP:
               case SIZE_OP:
               case INTEGER_DIV_OP:
               case SUB_OP:
               case NAMELIST_LIST:
               case ORDERED_OP:     /* Following added for PCF FORTRAN */
               case EXTEND_OP:
               case MAXPARALLEL_OP:
               case SAMETYPE_OP:
               case TYPE_REF:     /* Added for FORTRAN 90 */
               case STRUCTURE_CONSTRUCTOR:
               case ARRAY_CONSTRUCTOR:
               case SECTION_REF:
               case VECTOR_SUBSCRIPT:
               case SECTION_OPERANDS:
               case KEYWORD_ARG:
               case OVERLOADED_CALL:
               case INTERFACE_REF:
               case RENAME_NODE:
               case TYPE_NODE:
               case PAREN_OP:
               case PARAMETER_OP:
               case PUBLIC_OP:
               case PRIVATE_OP:
               case ALLOCATABLE_OP:
               case DIMENSION_OP:
               case EXTERNAL_OP:
               case IN_OP:
               case OUT_OP:
               case INOUT_OP:
               case INTRINSIC_OP:
               case POINTER_OP:
               case OPTIONAL_OP:
               case SAVE_OP:
               case TARGET_OP:
               case ONLY_NODE:
               case LEN_OP:
               case TYPE_OP:
               default:
                   break;
        }
    printf(" variant: ");
    printVariantName(e->variant());
    printf(" expr is: "); 
    e->unparsestdout();
    printf("\n");
    }
}

void classifySymbols(SgFile *f){
    SgSymbol *s;
    SgVariableSymb *vs;
    SgConstantSymb *cs;
    SgFunctionSymb *fs;
    SgLabelVarSymb *lvs;
    SgExternalSymb *exts;
    SgConstructSymb *cnsts;
    SgInterfaceSymb *ifs;
    SgModuleSymb *mods;
    for(s = f->firstSymbol(); s; s = s->next()){
	printf("SYMB[%d]: %s = %s  scope=STMT[%d] type:",
		s->id(), tag[s->variant()], s->identifier(), 
                (s->scope())?(s->scope())->id(): -1);
	if(s->type()) doFullType(s->type());
	doSymbAttribs(s); 
        printf("\n");
	switch(s->variant()){
            case CONST_NAME:
            case VARIABLE_NAME:
            case PROGRAM_NAME:
            case PROCEDURE_NAME:
            case LABEL_VAR:	/* dest of assigned goto stmt */
            case FUNCTION_NAME:
            case LABEL_NAME:	/* new added for VPC */
            case ROUTINE_NAME:	/*added for external statement*/
            case CONSTRUCT_NAME:
            case INTERFACE_NAME:
            case MODULE_NAME:
                       printf(" need work\n");
		       break;
            default:
		       printf("forgot one\n");
		       break;
	    }
       }
		
    }


/*int main(int argc, char **argv){

  SgProject project("x.proj");
  SgFile file("x.f");
  int c;

  initVariantNames();

  switch (file.languageType()) {
  case CSrc:
         printf("this is a c program.  this example is for fortran \n");
         return -1;
  case ForSrc:
         classifyStatements(&file);
         printf("**************************************************\n");
         printf("**** Expression Table ****************************\n");
         printf("**************************************************\n");
         classifyExpressions(&file);
         printf("**************************************************\n");
         printf("**** Symbol  Table *******************************\n");
         printf("**************************************************\n");
         classifySymbols(&file);
  }
  return 0;
} */
