#include <stdio.h>
#include "compilerApi.h"
#include <malloc.h>
#include <string.h>
#define  MAXUSEDVARS 256
// Find the enclosing function/subroutine unit

// decl for lexical analyser
#define VARID 509
char *input_file;
FILE *inf_ptr;
#undef YYLMAX
#define YYLMAX 8192
extern char yytext[YYLMAX];
extern int yylineno;



insertList(char *list[],int *len,char *key);
findUsedVars(char *inputFile,char *usedVars[],int *nVars);

SgStatement  *find_scope(SgStatement *s)
{
   SgProgHedrStmt *    prog,*curs;    
   SgProcHedrStmt *    subr;  
   SgProsHedrStmt *    pros;
   SgFuncHedrStmt *    func;  
   SgStatement *cp;
   

   cp = s;
   curs = NULL;
   do{
      if(subr = isSgProcHedrStmt (cp))
        curs = (SgProgHedrStmt *)subr;
      else if (pros = isSgProsHedrStmt(cp))
        curs = (SgProgHedrStmt *)pros;
      else if(func = isSgFuncHedrStmt(cp))
        curs = (SgProgHedrStmt *)func;
      else if(prog = isSgProgHedrStmt(cp))
        curs = prog;
      if (curs)
         break;
      cp = cp->controlParent();
   }while(!curs);
   return (curs);
   
}       
    
void generateNewFunction(char *hpfStmtList[],SgStatement *curStmt,
  int nLines)
{
  char *usedVars[MAXUSEDVARS];
  SgFunctionSymb *fsymb;
  SgStatement    *prev,*s,*ptfunc;
  int nVars;
    
  // Store the user's HPF stsmts into temporary File
  FILE *fTemp = fopen("userStmts","w");
  for(int i=0;i<nLines;i++)
    fprintf(fTemp,"%s",hpfStmtList[i]);
  fclose(fTemp);
  
  //Find the variables used in the HPF stmts user gave in intercative mode
  findUsedVars("userStmts",usedVars,&nVars);
  
  // get the scope and name of enclosing program unit
  SgStatement *origProgUnit = find_scope(curStmt);
  char *origProgName = (((SgProgHedrStmt *)origProgUnit)->name()).identifier();
   
  // Create new function
  char *newSubName          = (char *)malloc(256);
  sprintf(newSubName, "NewGen_%s", origProgName);
  fsymb   = new SgFunctionSymb(PROCEDURE_NAME,newSubName,
			       TYNULL,*(curStmt->controlParent()));
  
  ptfunc  = new SgProcHedrStmt(*fsymb);

  // add all the declaration stmts of original function
  prev = ptfunc;
  for(int i=0;i<origProgUnit->numberOfChildrenList1()-1;i++){
    SgStatement *chld = origProgUnit->childList1(i);
    if ((chld->variant() == VAR_DECL)||
        (chld->variant() == COMBINED_DIRECTIVE_DECL)||
        (chld->variant() == ALIGN_DECL))
    {
     
       s = &origProgUnit->childList1(i)->copy();
       prev->insertStmtAfter(*s,*ptfunc);
       prev = s;
   }
  }
  
    
  // Add used variables as parameters to new function
  for(int i=0;i<nVars;i++){
    SgVariableSymb *ns = new SgVariableSymb(usedVars[i]);
    fsymb->insertParameter(i,*ns);
  }
      
  // add the user's HPF stmts as comments before "subroutine End" stmt
  SgStatement *endStmt = ptfunc->childList1(ptfunc->numberOfChildrenList1()-1);
  for(int i=0;i<nLines;i++)
     endStmt ->addComment(hpfStmtList[i]);

  // Unparse new created function to standard output 
  ptfunc->unparsestdout(); 

}

findUsedVars(char *inputFile,char *usedVars[],int *nVars)
{

  int cnt=0,token;
  
  input_file = (char *)malloc(strlen(inputFile)+1);
  strcpy(input_file,inputFile);
 
  if (!(inf_ptr = fopen(input_file,"r"))){
    printf("I can not open input file : %s\n",input_file);
    exit(0);
  }
  
  while((token=yylex())){
    if (token == VARID)
      insertList(usedVars,&cnt,yytext);
    printf("text = %s line = %d token = %d\n",yytext,yylineno,token);
  }
  fclose(inf_ptr);
  *nVars = cnt;

  printf("I found %d used Varibales \n",cnt);
}

insertList(char *list[],int *len,char *key)
{
  int found = 0;
  
  for(int i=0;i<(*len);i++)
    if (!strcmp(list[i],key)){
      found = 1;
      break;
    }
  if (!found){
    list[(*len)] = (char *)malloc(strlen(key)+1);
    strcpy(list[(*len)],key);
    (*len)++;
  }
}

