#include <sys/types.h>
#include <rpc/types.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpc/xdr.h> 
#include "fortran_xdr.c"

extern int   errno;
extern char  *sys_errlist[];
extern int  t_errno;
extern char *t_errlist[];
u_short portbase = 0;
int max_push_id=0;
int max_type_id=3;
  
#define SERVER_CONNECT_RETRY_DELAY 1
#define QLEN            5
#define MAX_NAME_LENGTH 256
#define MAX_DIMS        7
#define MAX_VARS        32

#define T_CHAR          0
#define T_INT           1
#define T_FLOAT         2
#define T_LONG          3
#define T_DOUBLE        4
#define T_BOOL          5 
#define T_BYTE          6
#define T_SHORT         7
#define T_RECORD        8

#define UNCONNECTED     0
#define CONNECTED       1
#define MAX_OPEN_PORTS  32
#define MAX_ELEMS       999999999
#define MAX_XDR_BUF     999999999
#define MAX_TYPE_NAME   256
#define MAX_TYPE_NUMBER 256
#define CONTINUE        1
#define STOP_COMM       0

typedef struct push_data{
  int  type;
  char *name;
  int  ndims;  /* 0 for simple variables,ndims for array vars */ 
  int  dim_info[MAX_DIMS] ;
  int  sockfd; /* socket file descriptor */
  int  port;
  int  sizeof_data; /* total size of push_data in bytes */
  void *buf; /* includes actual data whatever it is */
  int  elsize; /* sizeof of each element */
  int  connected;
  int  nelems ; /* total element count */
  int  tind; /* index to type_list,returns the pointer to function */
  int  javabuflen;
  int  term_flag; /* terminate or continue sending/receiving var */
}DATA;
  
typedef struct send_buf{   
  char *sbuf;
  int   len;
  XDR   xdrs;
}SBUF;

typedef struct struct_type_list{
   /* name of XDR function */
   char   name[MAX_TYPE_NAME];
   bool_t  (*XDR_func)(register XDR *xdrs,void *obj ); 
   /* XDR function generated by XDR */
   int xdrlen;
   int reclen;
}TYPE;


DATA  stack[MAX_VARS]; 
TYPE  type_list[3]={{"XXX",xdr_XXX,96,60},{"one",xdr_one,28,28},
      {"x",xdr_x,96,60}};
/* Later, we have to construct the above table automatically. The first field
of each item is name of structure,second the corresponding XDR function for 
that type and last one is the total length XDR will generate for each data
item with the type of "name" . Last field is sizeof(record) */



void err(char *) ;
int server(int);


int find_tind(fname)
char *fname;
{
   int i,found=0;

   for(i=0;i<max_type_id;i++){
     if (!strcmp(fname,type_list[i].name) ){
       found = 1;
       break;
     }     
   }
     
   if (found) return i;
   else {
    printf("I could not found the fname = %s  len=%d\n",fname,strlen(fname));
    exit(0);
   }
 }


register_var_(var_type,var_id,var,name,var_ndims,dim_info)
int *var_type,*var_ndims,*dim_info,*var_id;
char *name;
void *var;
{
   int id,item_len,nelems,i;
   int type = *var_type,ndims=*var_ndims;
   float *ptr;

   id = max_push_id++;
   *var_id = id;

   stack[id].type = type;
   stack[id].ndims = ndims;
   
   printf("registering var with id =%d ndims = %d\n",id,ndims);
   /* calculate size of component length in bytes */
   /* transform bool,byte and short vars to int var */
   switch(type){
   case T_CHAR  :  item_len = 1;
                   stack[id].buf = (char  *)var;
                   break;
   case T_INT   :  item_len = 4;
                   stack[id].buf = (int *)var; break;
   case T_FLOAT :  item_len = 4;
                   stack[id].buf = (float *)var; break;
   case T_LONG  :  item_len = 8;
                   stack[id].buf = (long int *)var; break;
   case T_DOUBLE:  item_len = 8;
                   stack[id].buf = (double *)var; break;
 
   /* convert bool,byte and short vars into integer cars */
   case T_BOOL  :  item_len = 1;
                   stack[id].buf = (int *)var; break;    
   case T_BYTE  :  item_len = 1;
                   stack[id].buf = (int *)var; break;
   case T_SHORT :  item_len = 2;
                   stack[id].buf = (int *)var; break;
   case T_RECORD : stack[id].buf = (char *)var;
                   stack[id].tind = find_tind(name);
                   item_len       = type_list[stack[id].tind].xdrlen;
                   printf("item_len_record = %d sizeof len = %d\n",
                       item_len,sizeof((x *)var));
                   break;
   }

   stack[id].elsize = item_len;
    /* if variable is array */
   for(i=0;i<ndims;i++)
     stack[id].dim_info[i] = dim_info[i];
   
   stack[i].sockfd = -1;
   stack[i].port   = 0;

   /*calculate total length of variable in bytes:calculate # of elemens 
   first.The the total sizeof */

   for(i=0,nelems = 1;i<ndims;i++)
     nelems *= (stack[id].dim_info[i])    ;
   stack[id].nelems = nelems;

   if (type == T_RECORD) 
      stack[id].sizeof_data = nelems*(type_list[stack[id].tind].reclen);
   else
      stack[id].sizeof_data = nelems*(stack[id].elsize);

   stack[id].name = name;
   stack[id].connected = UNCONNECTED; 
   stack[id].javabuflen = stack[id].sizeof_data;

   /* XDR encode the char/bool/short/byte data as integer.Therefore
      XDR encoded buffer length should be made multiple of 4 */ 
   if ((type == T_BOOL)||(type == T_CHAR)||(type == T_BYTE))
      stack[id].javabuflen *= 4;
   else if (type == T_SHORT)
      stack[id].javabuflen *= 2;
   else if (type == T_RECORD)
      stack[id].javabuflen = nelems*(stack[id].elsize);

   stack[i].term_flag = CONTINUE;

   printf("%d was registered totalsize=%d javalen = %d\n",id,
          stack[id].sizeof_data,stack[id].javabuflen);

}

set_term_flag_(var_id)
int *var_id;
{
   int id = *var_id;

   stack[id].term_flag = STOP_COMM;
}




XDR_data(id,XDR_flag,xdrs)
int id ;
int XDR_flag;
XDR  *xdrs;
{

  int datalen,elsize,type,ndims;
  int nelems;
  void *dptr;
  int i,tind;
  bool_t ret;

  elsize  = stack[id].elsize;
  type    = stack[id].type;
  ndims   = stack[id].ndims;

  type = stack[id].type;
  nelems = stack[id].nelems;

  switch(type){
   case T_CHAR : 
               dptr = (char *)(stack[id].buf);
               if (ndims > 0) /* Character string */
                 ret = xdr_vector(xdrs,(char *)dptr,nelems,elsize,xdr_char);
               else /* just only one character . sent it as int var.*/
                 ret = xdr_char(xdrs,dptr);
               break;

  case T_INT  : case T_BOOL : case T_SHORT:case T_BYTE:
               dptr = (int *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_int);
               else 
                 ret = xdr_int(xdrs,dptr);
               break;
   case T_FLOAT: 
               dptr = (float *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_float);
               else
                 ret = xdr_float(xdrs,dptr); 
               break;

   case T_LONG : 
               dptr = (long int *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_long);
               else
                 ret = xdr_long(xdrs,dptr); 
               break;

   case T_DOUBLE:  
               dptr = (double *)(stack[id].buf);
               if (ndims > 0) /* array */
                 ret = xdr_vector(xdrs,dptr,nelems,elsize,xdr_double);
               else
                 ret = xdr_double(xdrs,dptr); 
               break;
  case T_RECORD:
               dptr = (char *)(stack[id].buf);
               tind = stack[id].tind;
               ret = (*type_list[tind].XDR_func)(xdrs,(char *)dptr);  
               break;
  }
   
  /* append termination flag */
  ret = xdr_int(xdrs,&(stack[id].term_flag)); 

  if (ret == FALSE){
      printf("There is error while encoding send data  with type=%d\n",type);
      exit(0);  
  }
}

/* id for stack array . XDR_flag : XDR_DECODE or XDR_ENCODE */
encode_send_data(id,sendbuf,XDR_flag)
int id ;
SBUF *sendbuf;
int XDR_flag;
{
  XDR xdrs;
  char *buf;
  int datalen;

  datalen = stack[id].javabuflen;
  buf = (char *)malloc(datalen);

  xdrmem_create(&xdrs,buf,datalen,XDR_flag);
  XDR_data(id,XDR_flag,&xdrs);

  sendbuf->sbuf = buf;
  sendbuf->len  = xdr_getpos(&xdrs);
  sendbuf->xdrs = xdrs;
}

send_data_(var_id)
int *var_id;
{
  SBUF sendbuf;
  int  len;
  char *ptr;
  int nleft,nwritten;
  int id = *var_id;

  encode_send_data(id,&sendbuf,XDR_ENCODE);
  len = sendbuf.len;
  ptr = (char *)(sendbuf.sbuf);
  nleft = len;

  while (nleft > 0) {
    nwritten = write(stack[id].sockfd, ptr, nleft);
    if (nwritten<=0)
      err("send_buffer: nwritten <= 0");
    nleft -= nwritten;
    ptr += nwritten;
  }
  xdr_destroy(&(sendbuf.xdrs));
  free(sendbuf.sbuf);
  printf("I send data for id=%d\n",id);

}  

decode_recv_data(id,recvbuf,datalen,XDR_flag)
int id ;/* id for stack array . XDR_flag : XDR_DECODE or XDR_ENCODE */
char  *recvbuf;
int datalen; /* total data length received */
int XDR_flag;
{
  XDR xdrs;
  char *buf;

  buf = (char *)malloc(datalen);
  xdrmem_create(&xdrs,buf,datalen,XDR_flag);
  bcopy(recvbuf,buf,datalen);
  XDR_data(id,XDR_flag,&xdrs);
  
  xdr_destroy(&xdrs);
  free(buf);

}


recv_data_(var_id)
int *var_id;
{
  int  len;
  int id = *var_id;
  XDR xdrs;
  char *buf;
  int datalen,type,ndims;

  datalen = stack[id].javabuflen; 
  type    = stack[id].type;
  ndims   = stack[id].ndims;

  buf = (char *)malloc(datalen);
  printf("%d size of data will be read for %d\n",datalen,id);
  len = readn(stack[id].sockfd,buf,datalen);
  printf("%d bytes read for var =%d\n",len,id);
  decode_recv_data(id,buf,len,XDR_DECODE);
}  



int
readn(fd, ptr, nbytes)
register int    fd;
register char   *ptr;
register int    nbytes;
{
    int nleft, nread;

    nleft = nbytes;
    while (nleft > 0) {
        nread = read(fd, ptr, nleft);
        if (nread < 0)
            return(nread);      /* error, return < 0 */
        else if (nread == 0)
            break;          /* EOF */

        nleft -= nread;
        ptr   += nread;
    }
    return(nbytes - nleft);     /* return >= 0 */
}




configure_(var_id,var_port)
int *var_id;
int  *var_port;
{
  int id = *var_id,port = *var_port;
  int sockfd,i,flag=0;

  for(i=0;i<max_push_id;i++)
    if ((stack[i].port == port)&&(stack[i].connected == CONNECTED)){ 
       flag = 1;
       break;
     } 

  if (!flag)  
    sockfd = server(port);
  else
    sockfd = stack[i].sockfd;
  
  stack[id].connected = CONNECTED;
  stack[id].port      = port;
  stack[id].sockfd    = sockfd;

  printf("after configure for %d\n",id);
}

unregister_(var_id)
int *var_id;
{
   int i,port,sockfd,id=*var_id;
   int unconnected = 0;

   port     = stack[id].port;
   sockfd   = stack[id].sockfd;

   for(i=0;i<max_push_id;i++)
    if ((stack[i].port == port)&&(stack[i].connected == UNCONNECTED)){ 
       unconnected = 1;
       break;
    }   
  
   if (unconnected)
     printf("port = %d is already unconnected\n",port);
   else{
     close(sockfd);
     stack[id].connected == UNCONNECTED;
   }
}
  

int  server(int  portnum) {

  int  i, sockfd, newsockfd;
  int clilen, one=1;
  struct sockaddr_in cli_addr, serv_addr;

  if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    err("server: can't open stream socket");
  
  if ( setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
		   (char *) &one, sizeof(one)) == -1) {
    err("server: setsockopt SO_REUSEADDR failed");
  }

  memset((char *)&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons((u_short)portnum);    
  if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
    err("server: can't bind local address");
  
  /* wait for socket connection from client */

  fflush(stdout);
  listen(sockfd, 5);
  
  clilen = sizeof(cli_addr);
  newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
  if (newsockfd < 0) {
    close(sockfd);
    err("server: socket accept error");
  }
  close(sockfd);
  printf("server: data_client connected on port %d...\n", portnum);
  return(newsockfd);

}

void err(char *string) {
  perror(string);
  exit(0);
}

int bind_server(char *serv_host_name, int serv_tcp_port) {
  int sockfd;
  char serv_host_addr[64];
  struct sockaddr_in serv_addr;
  struct sockaddr_in tmp_sin;
  struct hostent *hp;

  if ((hp = gethostbyname(serv_host_name)) == NULL) {
    printf("daqv_bind_server: unknown host \"%s\".\n", serv_host_name);
    return -1;
  }
  memcpy(&tmp_sin.sin_addr, hp->h_addr, hp->h_length);
  strcpy(serv_host_addr, inet_ntoa(tmp_sin.sin_addr));

  printf("daqv_bind_server: server info - name=%s, addr=%s, port=%d\n",
	 hp->h_name, serv_host_addr, serv_tcp_port);

  do {
    printf("daqv_bind_server: trying to connect...\n");
    memset((char *) &serv_addr,0,sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(serv_host_addr);
    serv_addr.sin_port = htons(serv_tcp_port);
    if((sockfd = socket(AF_INET,SOCK_STREAM,0))<0) {
      printf("daqv_bind_server: can't open stream socket\n");
      return -1;
    }
    sleep(SERVER_CONNECT_RETRY_DELAY);
  } while (connect(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr))<0);

  return sockfd;
}

configure_client_(var_id,char serv_host_name,var_port)
int *var_id;
int  *var_port;
char *serv_host_name;
{
  
  int id = *var_id,port = *var_port;
  int sockfd,i,flag=0;

  for(i=0;i<max_push_id;i++)
    if ((stack[i].port == port)&&(stack[i].connected == CONNECTED)){ 
       flag = 1;
       break;
     } 

  if (!flag)  
    sockfd = bind_server(serv_host_name,port);
  else
    sockfd = stack[i].sockfd;
    
  stack[id].connected = CONNECTED;
  stack[id].port      = port;
  stack[id].sockfd    = sockfd;

  printf("after configure for %d\n",id);
}
