#include "compilerApi.h"
#include "deftypes.h"

#define VARID   509
#define	QLEN		   5	/* maximum connection queue length	*/
#define	BUFSIZE		20480
#define VARLEN            256
#define OPEN_PROJECT    1
#define READ_PARSE_TREE 2
#define CLOSE_CONN      3
#define EXIT            4
#define INSTRUMENT      5


/* Decls for yylex() */
/********************/
char *input_file;
FILE *inf_ptr;
#undef YYLMAX
#define YYLMAX 8192
extern char yytext[YYLMAX];
extern int yylineno;
extern "C" int yylex();

/********************/
/* End of decls for yylex() */

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif  
extern int  errno;
extern char *sys_errlist[];

struct Internalfunctable{
  SgStatement *startAddr;
  char *fname;
}curFuncTable[256];
char DefinedVars[VARLEN][VARLEN];
int numOfDefinedVars;

//Some declarations
FILES *files;  
int nfiles;
SgFile *curSgFile;
static int  numOfFuncs;

SgStatement *find_first_exec_stmt(SgStatement *);
int isDecl(SgStatement *chld);
insertList(char *list[],int *len,char *key);
void readParseTree(char *projectFileName,int *nfiles,FILES **files);
void send_instrumented_file(int sockfd);  
int passivesock(char *service, char *protocol, int qlen )  ;
int errexit(char *format, ...);
extern "C" void bcopy(const void *s1, void *s2, size_t n);
extern "C" int getdtablesize(void);
extern "C" void bzero(void *s, size_t n);
void send_String(int sockfd,char *variableName);
void recv_String(int sockfd,char *variableName);
void send_int(int sockfd,int command);
void recv_int(int sockfd,int *command);
void send_raw_data(int sockfd,char *buf,int len);
int  recv_raw_data(int sockfd,char *buf,int len);
int  insertList(char *list[],int *len,char *key);
void findUsedVars(char *hpfStmtList,char *funcName,
		  char *usedVars[],int *nVars);
SgSymbol *IsItInScope(SgStatement * scope, char * identifier);


void openNewProject(char *depFileName)
{
   int j,fcnt=0;
   char *myargs[3]={NULL,NULL,(char *)0};
   myargs[1] =  new char[256];//(char *)malloc(256);

   strcpy(myargs[1],depFileName);
   SgProject *project = new SgProject("",2,myargs);
   curSgFile = &(project->file(0));

   for(j=0;j<curSgFile->numberOfFunctions();j++){
     SgStatement *sub = curSgFile->functions(j);
     if (sub->controlParent()->variant() !=  INTERFACE_STMT){
	 curFuncTable[fcnt].startAddr = sub;
	 SgSymbol *subsym  =  (curFuncTable[fcnt].startAddr)->symbol();
	 curFuncTable[fcnt].fname     = subsym->identifier();
	 fcnt++;
     }
   }
   numOfFuncs = fcnt;
   numOfDefinedVars = 0;
   
} 
  

SgSymbol *IsItInScope(SgStatement * scope, char * identifier)
{
  SgSymbol * symb = scope->symbol();
  int found = 0;
   
  // find the first symbol in next scope;
  // suppose that at least one symbol in this scope;
  // else it may need a long run;
 
  while ((symb->next()!=NULL)) {
    symb=symb->next();
    if (symb->scope()==scope)
      break;
  }
 
  for( ; symb->nextInScope(); symb=symb->nextInScope()) {
    if(strcasecmp(symb->identifier(), identifier)==0) {
      found = 1;
      break;
    };
  }
 
  if (found)
    return symb;
  else
    return NULL;
  
}


int findex(char *name)
{
   int found = -1;  

   for(int j=0;j<numOfFuncs;j++){
     if (!strcmp(name,curFuncTable[j].fname )){
         found = j;
	 break;
     }
   }  
  
   if ((found<0)||(found==numOfFuncs)){
      printf("I could not find file : %s\n",name);
      exit(0);
   }
   return found;

}

int inUsedList(char *list[],int len,SgSymbol  *key)
{
  int found = 0;
  
  for(int i=0;i<len;i++)
    if (!strcmp(list[i],key->identifier())){
      found = 1;
      break;
    }
  return found;

}

   
int  generateNewFunction(char *hpfStmtList, char *calledFunc,
		  int  *in_nVars,char *usedVars[],int nextModule)
{
  SgFunctionSymb *fsymb;
  SgStatement    *prev,*s,*ptfunc;
  int cnt,interface,ret,i,nVars ;
  SgType *type;
  FILE *fTemp; 
  char cmnd[256],depfile[VARLEN],filename[VARLEN],*tempBuf;
  char command[5096];
  
  nVars = *in_nVars;
  // get the scope and name of enclosing program unit
  int index = findex(calledFunc);
  SgStatement *origProgUnit = curFuncTable[index].startAddr;
  char *origProgName = (((SgProgHedrStmt *)origProgUnit)->name()).identifier();

  // Add also unused original parameters of main function in which
  // we are requesting interpreter statements
  SgProgHedrStmt *hedr   = (SgProgHedrStmt *)origProgUnit;
  SgFunctionSymb *paramsSymb = (SgFunctionSymb *)(&hedr->name());
  int nparams = paramsSymb->numberOfParameters();
  for(i=0;i<nparams;i++){
     SgSymbol *nsymb = &paramsSymb->parameter(i)->copy();
     if (!inUsedList(usedVars,nVars,nsymb))
	insertList(usedVars,&nVars,nsymb->identifier());
  }
  *in_nVars = nVars;

  // Create first new function
  char *newSubName          = new char[512];//(char *)malloc(512);
  sprintf(newSubName, "newfunction_%d", nextModule);
  type = new SgType(T_VOID);
  fsymb   = new SgFunctionSymb(PROCEDURE_NAME,newSubName,
			       *type,*(origProgUnit->controlParent()));
  ptfunc  = new SgProcHedrStmt(*fsymb);

  // Add used variables as parameters to new function
  for(i=0,cnt=0;i<nVars;i++){
     printf("%s is appended as parameter \n",usedVars[i]);
     SgVariableSymb *ns = new SgVariableSymb(usedVars[i]);
     fsymb->insertParameter(cnt++,*ns);
  }

  // add all the declaration stmts of original function
  int totdecls=0;
  prev = ptfunc;
  int numOfChildren = origProgUnit->numberOfChildrenList1()-1;
  for(i=0;i<numOfChildren;i++){
    SgStatement *chld = origProgUnit->childList1(i);
    if (isDecl(chld))
     {
        s = &chld->copy();
        prev->insertStmtAfter(*s,*ptfunc);
        prev = s;
	totdecls++;
     }
  }    
  
  //add the user's HPF stmts as comments before "subroutine End" stmt
  SgStatement *endStmt =ptfunc->childList1(ptfunc->numberOfChildrenList1()-1);
  endStmt ->addComment(hpfStmtList); 
  tempBuf = ptfunc->unparseToBuf();
  printf("UNPARSE of FIRST FUNCTION = \n %s\n",tempBuf);

  ///////////////////////////////////////////////////////
  /////////////////////////////////////////////////////
  /* Set separaretor stmt id which actually is 
     the number of total declarations in newly generated subroutine */
  interface = totdecls;

  /* Write the first uninstrumented subroutine to file "foos.f" */
  strcpy(filename,"foos.f");
  strcpy(depfile,"foos.dep"); 
  fTemp = fopen(filename,"w");
  fprintf(fTemp,"%s", tempBuf);
  fclose(fTemp);

  /* make dep. file */
  sprintf(cmnd,"hpf2dep %s > sageServer_err 2>&1",filename);
  ret = system(cmnd);
  if (ret != 0)
    return COMPILE_ERROR ;
  
    
  /* Build command:  instru. foos.dep nextModule interface */
  strcpy(command,"instrumentInterpreter  ");
  strcat(command,depfile);

  /* Add nextModule parameter */
  strcpy(cmnd,"");
  tempBuf = lltostr(nextModule,cmnd);
  strcat(command," ");
  strcat(command,tempBuf);
    
  /* Add interface parameter */
  strcpy(cmnd,"");
  tempBuf = lltostr(interface,cmnd);
  strcat(command," ");
  strcat(command,tempBuf);

  printf("command = %s\n",command);
   
  /* Try not to produce module but only function */
   if ((ret = system(command)) < 0)
      printf("system error : %d\n",ret);
  
  ret = 0;
  
  if (ret != 0)
    return COMPILE_ERROR ;
  else
    return 0;
}

void clear_path(char *projectFileName)
{
  char *p;
  int len;
    
  p = strrchr(projectFileName,'/');
  if (p){
    len = strlen(p+1);
    strcpy(projectFileName,p+1);
    projectFileName[len] = 0;
  }
    
}
  

/***********************************************/
/***********************************************/
/* Sage HPF Front End Server */
/***********************************************/
/***********************************************/
sageServer(int sockfd)
{
  char *usedVars[VARLEN],calledFunc[VARLEN],hpfStmtList[BUFSIZE];
  
  int  nVars=0,f,nfiles,fnfuncs,fsize,command,errcode;
  char depFileName[VARLEN],projectFileName[VARLEN];
  static int opened = 0;
  int nextModule,ret;
     

  printf("Entered sageServer sockfd =%d\n",sockfd);
  
  recv_int(sockfd,&command);
  switch(command){
   case OPEN_PROJECT: 
     recv_String (sockfd,depFileName);
     clear_path(depFileName);
     openNewProject(depFileName);
     break;  
   case  READ_PARSE_TREE:
     recv_String (sockfd,projectFileName);
     clear_path(projectFileName);
     
     if (!opened){
       readParseTree(projectFileName,&nfiles,&files);
       printf("after reading projectFile : %s\n",projectFileName);
       opened = 1;
     }
     printf("SageServer : after readParseTree nfiles=%d \n",nfiles);
     
     send_int(sockfd,nfiles);
     for(f = 0;f<nfiles;f++){
       /* later change the following by using XDR library  */
       send_raw_data(sockfd,(char *)(&files[f]),sizeof(FILES));
       fsize   = files[f].fileSize;
       fnfuncs = files[f].nfuncs; 
       send_int(sockfd,fsize);
       send_int(sockfd,fnfuncs );
       /* Later also change the following two lines by using XDR library*/
       send_raw_data(sockfd,(char *)(files[f].functab),sizeof(FTAB)*fnfuncs);
       send_raw_data(sockfd,(char *)(files[f].lineInfo),sizeof(SINFO)*fsize);
     }
     break;
   case INSTRUMENT: 
     recv_int(sockfd,&nextModule);
     recv_String (sockfd,calledFunc);
     recv_String (sockfd,hpfStmtList);

     /* find the used variables in user HPF stmts */
     findUsedVars(hpfStmtList,calledFunc,usedVars,&nVars);
     /* Generate module including function */
     ret = generateNewFunction(hpfStmtList,calledFunc,&nVars,usedVars,
			     nextModule);

     /* send the error code */
     send_int(sockfd,ret);
     /* Send parameter list */
     send_int(sockfd,nVars);
     for(f=0;f<nVars;f++){
        send_String(sockfd,usedVars[f]);
        free(usedVars[f]);  /* free space for params */ 
     }
     /* send the generated module */
     send_instrumented_file(sockfd);

     break;
   case CLOSE_CONN:  
     return 1;
     break;
   case EXIT: 
     return 2;
     break;
  }
  printf("return SageServer \n");
  
  return 0;

  
}

void store_DefinedVars(char *line)
{
  int n = numOfDefinedVars;
  char *p;
  
  p = strtok(line,"\n\r\t ");
  if (p != NULL)
    strcpy(DefinedVars[n++],p);
  else
    return;
  while((p = strtok(NULL,"\n\r\t ")))
    strcpy(DefinedVars[n++],p);
  numOfDefinedVars = n;
  
}

int isItDefinedVar(char *varname)
{
   int n = numOfDefinedVars,k;
   for(k=0;k<n;k++)
     if (!strcmp(DefinedVars[k],varname))
       return 1;
   return 0;
     
}



void send_instrumented_file(int sockfd)
{
  FILE *fp;
  char line[BUFSIZE];
  char buf[BUFSIZE];
     
  strcpy(buf,"");
  fp = fopen("foos.f","r"); 
  
  fgets(line,VARLEN,fp);
  /* If there is new generated variables at interpreter mode,
     store these for later use */
  if (!strstr(line,"MODULE")){
    store_DefinedVars(line);
    fgets(line,VARLEN,fp);
  }
  do{
    strcat(buf,line); 
    fgets(line,VARLEN,fp);
  }while(!feof(fp));
  fclose(fp);
  send_String(sockfd,buf);
  
}




/*------------------------------------------------------------------------
 * main - Concurrent TCP server for ECHO service
 *------------------------------------------------------------------------
 */

int main(int argc, char *argv[])
{
	char	*service = "3334";	/* service name or port number	*/
	char    *basedir;
	struct sockaddr_in fsin;	/* the from address of a client	*/
	int	msock;			/* master server socket		*/
	fd_set	rfds;			/* read file descriptor set	*/
	fd_set	afds;			/* active file descriptor set	*/
	int	alen;			/* from-address length		*/
	int	fd, nfds;
	int ret,i;
	struct timeval tval={0,0};
		
	switch (argc) {
	case	1:
		break;
	case	2:
		service = argv[1];
		break;
        case    3:
	        service = argv[1];
		basedir = argv[2];
		
		break; 
    	        
	default:
		errexit("usage: TCPmechod [port]\n"); 
	}

        printf("Begin enf SageServer \n");
        chdir(basedir);
	msock = passivesock(service, "tcp", QLEN);

	nfds = getdtablesize();
	FD_ZERO(&afds);
	FD_SET(msock, &afds);

	while (1) {
	    bcopy((char *)&afds, (char *)&rfds, sizeof(rfds));

	    if (select(nfds, &rfds, (fd_set *)0, (fd_set *)0,&tval) < 0)
		errexit("select: %s\n", sys_errlist[errno]);
	    if (FD_ISSET(msock, &rfds)) {
		int	ssock;

		alen = sizeof(fsin);
		ssock = accept(msock, (struct sockaddr *)&fsin,	&alen);
		if (ssock < 0)
		   errexit("accept: %s\n", sys_errlist[errno]);
		FD_SET(ssock, &afds);

		printf("I am connected to port : %d\n",ssock);
		
	     }
   	     for (fd=0; fd<nfds; ++fd)
	       if (fd != msock && FD_ISSET(fd, &rfds)){
		   ret = sageServer(fd);
                   /* Close connection */				   
		   if (ret == 1) {
		      (void) close(fd);
		      FD_CLR(fd, &afds);
		   }
		   /* Stop server */
		   else if (ret == 2){
		     for (fd=0; fd<nfds; ++fd)
		       if (fd != msock && FD_ISSET(fd, &rfds)){
			 (void) close(fd);
			 FD_CLR(fd, &afds);
			 printf("I am closing fd : %d\n",fd);
			 
		       }
		     (void) close(msock);
		     FD_CLR(msock, &afds);
		     printf("I am closing msock : %d\n",msock);
		     exit(0);
		   } /*els if (ret == 2) */
		   
	       } /* if (fd != msock.. */

	}/* while */
}/* main */



/*------------------------------------------------------------------------
 * passivesock - allocate & bind a server socket using TCP or UDP
 *------------------------------------------------------------------------
 */

int passivesock(char *service, char *protocol, int qlen )
/* char    *service;  */     /* service associated with the desired port     */
/* char    *protocol;  */    /* name of protocol to use ("tcp" or "udp")     */
/* int     qlen;      */     /* maximum length of the server request queue   */
{
        struct servent  *pse;   /* pointer to service information entry */
        struct protoent *ppe;   /* pointer to protocol information entry*/
        struct sockaddr_in sin; /* an Internet endpoint address         */
        int     s, type;        /* socket descriptor and socket type    */
	int ret;
	u_short  portbase = 0;  

        bzero((char *)&sin, sizeof(sin));
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = INADDR_ANY;

    /* Map service name to port number */
        if ( pse = getservbyname(service, protocol) )
                sin.sin_port = htons(ntohs((u_short)pse->s_port)
                        + portbase);
        else if ( (sin.sin_port = htons((u_short)atoi(service))) == 0 )
                errexit("can't get \"%s\" service entry\n", service);

    /* Map protocol name to protocol number */
        if ( (ppe = getprotobyname(protocol)) == 0)
                errexit("can't get \"%s\" protocol entry\n", protocol);

    /* Use protocol to choose a socket type */
        if (strcmp(protocol, "udp") == 0)
                type = SOCK_DGRAM;
        else
                type = SOCK_STREAM;

    /* Allocate a socket */
        s = socket(PF_INET, type, ppe->p_proto);
        if (s < 0)
                errexit("can't create socket: %s\n", sys_errlist[errno]);

    /* Bind the socket */
        if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
                errexit("can't bind to %s port: %s\n", service,
                        sys_errlist[errno]);
        if (type == SOCK_STREAM && listen(s, qlen) < 0)
                errexit("can't listen on %s port: %s\n", service,
                        sys_errlist[errno]);
	
        return s;
}

int errexit(char *format, ...)

{
    va_list args;
 
    va_start(args,format);
    _doprnt(format, args, stderr);
    va_end(args);
    exit(1);
}   

int recv_raw_data(int fd, char *buf, int nbytes)
{
    int nleft, nread;
    char *ptr = buf;
    sigset_t oset;
    
    /* Mask three signals */
    /* mask_three_signals(&oset); */

    nleft = nbytes;
    while (nleft > 0) {
      /* nread = read(fd, ptr, nleft);*/
        nread = recv(fd, ptr, nleft,0);
        if (nread < 0)
            return(nread);      /* error, return < 0 */
        else if (nread == 0)
            break;          /* EOF */

        nleft -= nread;
        ptr   += nread;
    }
    printf("%d bytes read for var =%d original = %d left =%d\n",ptr-buf,
	       fd,nbytes,nleft);
    
    /* Unmask three signals */
    /* unmask_three_signals(&oset); */
    return(nbytes - nleft);     /* return >= 0 */

}

void send_raw_data(int sockfd,char *buf,int len)
{
  char *ptr;
  int nleft,nwritten;
  sigset_t oset;
 
  /* Mask three signals */
  /* mask_three_signals(&oset); */

  ptr   = buf;
  nleft = len;
  while (nleft > 0) {
    /*nwritten = write(sockfd, ptr, nleft); */
    nwritten = send(sockfd, ptr, nleft,0);
    if (nwritten<=0)
      errexit("send_buffer: nwritten <= 0",sys_errlist[errno]);
    nleft -= nwritten;
    ptr += nwritten;
  }

  /* Unmask three signals */
  /* unmask_three_signals(&oset); */
  
  printf("I send raw data sent len =%d\n",len);
}

void recv_int(int sockfd,int *command)
{
  char cmnd[4];
  XDR xdrs;
 
  recv_raw_data(sockfd,cmnd,sizeof(int));
  xdrmem_create(&xdrs,cmnd,sizeof(int),XDR_DECODE); 
  xdr_int(&xdrs,command);

  xdr_destroy(&xdrs);
  printf("I received integer : %d\n",*command);
 
}

void send_int(int sockfd,int command)
{
  char cmnd[4];
  XDR xdrs;
 
  xdrmem_create(&xdrs,cmnd,sizeof(int),XDR_ENCODE); 
  xdr_int(&xdrs,&command);
  send_raw_data(sockfd,cmnd,sizeof(int));

  xdr_destroy(&xdrs);
  printf("After send INTEGER : %d\n",command);
}

void recv_String (int sockfd,char *variableName)
{
   bool_t ret;

   char buffer[REGBUFSIZE],*tbuf,intbuf[4];
   int strLen,totalLen;
   XDR xdrs;   
   int i;

   tbuf = (char *)malloc(REGBUFSIZE);
   
   recv_raw_data(sockfd,buffer,sizeof(int));
   strncpy(intbuf,buffer,4);
   xdrmem_create(&xdrs,buffer,sizeof(int),XDR_DECODE); 
   if ((&xdrs) == NULL)
     printf("1:xdrmem_create error in recv_string \n");
   ret=xdr_int(&xdrs,&strLen);
   if (ret == FALSE)
     printf("xdr_int error in recv_string \n");
   xdr_destroy(&xdrs);

   totalLen = strLen;
   if ((strLen%4) != 0)
     totalLen = strLen+4-strLen%4;
  
   recv_raw_data(sockfd,&buffer[sizeof(int)],totalLen);
   /* strncpy(buffer,intbuf,4); */
   xdrmem_create(&xdrs,buffer,REGBUFSIZE,XDR_DECODE);
   if ((&xdrs) == NULL)
     printf("2:xdrmem_create error in recv_string \n");
   ret = xdr_string(&xdrs,&tbuf,totalLen);
   if (ret == FALSE)
     printf("xdr_string error in recv_string \n");

   tbuf[strLen] = 0; /* Append terminating character  */ 
   strcpy(variableName,tbuf);
   /* variableName[strLen] = 0; */ /* Append terminating character  */ 
   if (DEBUG) 
      printf("I received string : %s strlen = %d totlen=%d\n",
         variableName,strLen,totalLen);
      
   xdr_destroy(&xdrs);
}

void send_String (int sockfd,char *variableName)
{
   char *buffer,*ts;
   int totalLen,ret;
   XDR xdrs;
       
   ts = (char *)malloc(strlen(variableName)+1);
   buffer = (char *)malloc(REGBUFSIZE);
   strcpy(ts,variableName);
   
   xdrmem_create(&xdrs,buffer,REGBUFSIZE,XDR_ENCODE); 
   /* ret = xdr_string(&xdrs,&variableName,strlen(variableName));  */
   ret = xdr_string(&xdrs,&ts,strlen(ts));

   totalLen = xdr_getpos(&xdrs);
   send_raw_data(sockfd,buffer,totalLen);
   
   xdr_destroy(&xdrs);
   free(ts);
   free(buffer);
   
   if (DEBUG)
      printf("I  send string = %s strlen =%d\n",
          variableName,strlen(variableName));
}


void findUsedVars(char *hpfStmtList,char *funcName,
		  char *usedVars[],int *nVars)
{

  int cnt,token,fi,symb_atts,sln,fln;
  char input_file[VARLEN];
  SgStatement *scope,*fexecstmt;
  SgSymbol *symb;
  
  fi = findex(funcName);
  scope = curFuncTable[fi].startAddr;
  fexecstmt = find_first_exec_stmt(scope);
  fln = fexecstmt->lineNumber();
    
  strcpy(input_file,"tmpFile");
  inf_ptr = fopen(input_file,"w+");
  if (inf_ptr == NULL){
    printf("I can not open input file : %s\n",input_file);
    exit(0);
  }
  fprintf(inf_ptr,"%s",hpfStmtList);
  fseek(inf_ptr,0L,SEEK_SET);

  cnt = 0;
  while((token=yylex())){
    if (token == VARID){
      symb = IsItInScope(scope, yytext);
      /* Don't insert constant parameter (PARAM(...))  and vars defined 
	 at interpreter mode as argument */
      
      if ((symb)||(!isItDefinedVar(yytext))){
	sln = symb->declaredInStmt()->lineNumber();
	symb_atts = symb->attributes();
	if ((sln < fln)&&(!(symb_atts & PARAMETER_BIT)))
	  insertList(usedVars,&cnt,yytext);
      }
      
    }
  }
  
  fclose(inf_ptr);  
  *nVars = cnt;
}

int  insertList(char *list[],int *len,char *key)
{
  int i,found = 0;

  for(i=0;i<(*len);i++)
    if (!strcmp(list[i],key)){
      found = 1;
      break;
    }
  if (!found){
    list[(*len)] = (char *)malloc(strlen(key)+1);
    strcpy(list[(*len)],key);
    (*len)++;
  }
}

  
